from typing import Any

import distrax
import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import PyTree

from tensorflow_probability.substrates import jax as tfp

eps = 2 * jnp.finfo(jax.numpy.float32).eps


class EqxDistribution(eqx.Module):
    args: PyTree[Any]
    kwargs: PyTree[Any]

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def dist(self):
        return self

    def sample_and_log_prob(self, seed, sample_shape=()):
        dist = self.dist()
        sample = dist.sample(seed=seed, sample_shape=sample_shape)
        log_prob = dist.log_prob(sample)

        return sample, log_prob

    def log_prob(self, x):
        return self.dist().log_prob(x)

    def feature_log_prob(self, x):
        return self.dist()._distribution.log_prob(x)

    def sample(self, seed, sample_shape=()):
        return self.dist().sample(seed=seed, sample_shape=sample_shape)

    def kl_divergence(self, other_dist):
        return self.dist().kl_divergence(other_dist.dist())

    def mean(self):
        return self.dist().mean()

    def std(self):
        return self.dist().stddev()


class SafeTanh(distrax.Tanh):
    def inverse_and_log_det(self, y):
        """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
        y = jax.lax.clamp(min=-1 + eps, x=y, max=1 - eps)
        x = jnp.arctanh(y)
        return x, -self.forward_log_det_jacobian(x)


class TanhMultivariateNormalDiag(EqxDistribution):
    args: PyTree[Any]
    kwargs: PyTree[Any]
    dist_type = "tanh_normal"

    def dist(self):
        return distrax.Independent(
            distrax.Transformed(distrax.Normal(*self.args, **self.kwargs), SafeTanh()), 1
        )

    def mean(self):
        return jax.nn.tanh(self.kwargs["loc"])

    def std(self):
        return self.kwargs["scale"]

    def kl_penalty(self):
        kwargs = dict(
            loc=jnp.zeros_like(self.kwargs["loc"]), scale=jnp.ones_like(self.kwargs["scale"] * 0.5)
        )
        return self.dist().kl_divergence(
            distrax.Independent(distrax.Transformed(distrax.Normal(**kwargs), SafeTanh()), 1)
        )

    @property
    def params(self):
        return {
            "mean": self.kwargs["loc"],
            "std": self.kwargs["scale"],
            "dist_type": self.dist_type,
        }

    def entropy(self):
        # the empirical maximima entropy tanhNormal is (0, 0.8742) in which p(0.5) = 0.5
        kwargs = dict(
            loc=jnp.zeros_like(self.kwargs["loc"]), scale=jnp.ones_like(self.kwargs["scale"] * 0.8742)
        )
        return self.dist().kl_divergence(
            distrax.Independent(distrax.Transformed(distrax.Normal(**kwargs), SafeTanh()), 1)
        )


class MultivariateNormalDiag(EqxDistribution):
    args: PyTree[Any]
    kwargs: PyTree[Any]
    dist_type = "normal"

    def dist(self):
        return distrax.Independent(distrax.Normal(*self.args, **self.kwargs), 1)

    def kl_penalty(self):
        kwargs = dict(
            loc=jnp.zeros_like(self.kwargs["loc"]), scale=jnp.ones_like(self.kwargs["scale"])
        )
        return self.dist().kl_divergence(distrax.Independent(distrax.Normal(**kwargs), 1))

    def mean(self):
        return self.kwargs["loc"]

    def std(self):
        return self.kwargs["scale"]

    def var(self):
        return jnp.square(self.kwargs["scale"])

    @property
    def params(self):
        return {
            "mean": self.kwargs["loc"],
            "std": self.kwargs["scale"],
            "dist_type": self.dist_type,
        }


class MultivariateBernoulliDiag(EqxDistribution):
    args: PyTree[Any]
    kwargs: PyTree[Any]
    dist_type = "bernoulli"

    def dist(self):
        return distrax.Independent(distrax.Bernoulli(*self.args, **self.kwargs), 1)

    @property
    def params(self):
        return {
            "logits": self.kwargs["logits"],
            "dist_type": self.dist_type,
        }

    def mode(self):
        return self.dist().mode()


def multivariate_normal_diag(mean, std):
    return MultivariateNormalDiag(loc=mean, scale=std)


def tanh_multivariate_normal_diag(mean, std):
    return TanhMultivariateNormalDiag(loc=mean, scale=std)


def bernoulli_diag(logits):
    return MultivariateBernoulliDiag(logits=logits)


class ImageMultivariateNormalDiag(EqxDistribution):
    args: PyTree[Any]
    kwargs: PyTree[Any]
    dist_type = "image_normal"

    def dist(self):
        return distrax.Independent(distrax.Normal(*self.args, **self.kwargs), 3)

    def log_prob(self, x):
        dist = self.dist()
        return dist.log_prob(x) + 0.5 * jnp.log(2 * jnp.pi) * jnp.prod(jnp.array(dist.event_shape))

    def kl_penalty(self):
        kwargs = dict(
            loc=jnp.zeros_like(self.kwargs["loc"]), scale=jnp.ones_like(self.kwargs["scale"])
        )
        return self.dist().kl_divergence(distrax.Independent(distrax.Normal(**kwargs), 3))

    def mean(self):
        return self.kwargs["loc"]

    def std(self):
        return self.kwargs["scale"]

    def var(self):
        return jnp.square(self.kwargs["scale"])

    @property
    def params(self):
        return {
            "mean": self.kwargs["loc"],
            "std": self.kwargs["scale"],
            "dist_type": self.dist_type,
        }


class MultivariateBetaDiag(EqxDistribution):
    args: PyTree[Any]
    kwargs: PyTree[Any]
    dist_type = "beta"

    def dist(self):
        return distrax.Independent(distrax.Transformed(distrax.Beta(*self.args, **self.kwargs), distrax.ScalarAffine(scale=2, shift=-1)), 1)

    def kl_penalty(self):
        # how to set alpha & beta for penalty?
        kwargs = dict(alpha=jnp.ones(self.kwargs["alpha"]) * 3, beta=jnp.ones_like(self.kwargs["beta"]) * 3)
        return self.dist().kl_divergence(distrax.Independent(distrax.Transformed(distrax.Beta(**kwargs), distrax.ScalarAffine(scale=2, shift=-1)), 1))

    def feature_log_prob(self, x):
        return self.dist()._distribution.log_prob(jax.lax.clamp(min=-1 + eps, x=x, max=1 - eps))

    def sample_and_log_prob(self, seed, sample_shape=()):
        dist = self.dist()
        sample = dist.sample(seed=seed, sample_shape=sample_shape)
        log_prob = dist.log_prob(jax.lax.clamp(min=-1 + eps, x=sample, max=1 - eps))

        return sample, log_prob

    def log_prob(self, x):
        return self.dist().log_prob(jax.lax.clamp(min=-1 + eps, x=x, max=1 - eps))

    def mode(self):
        return self.dist().mode()

    @property
    def params(self):
        # Hack to make beta distribution compatible with mean and std representation
        return {
            "mean": self.kwargs["alpha"],
            "std": self.kwargs["beta"],
            "dist_type": self.dist_type,
        }

    def entropy(self):
        return self.dist().entropy()


def beta_multivariate_normal_diag(mean, std):
    # Hack to make beta distribution compatible with mean and std representation
    return MultivariateBetaDiag(alpha=mean, beta=std)
