import os
os.environ['MUJOCO_GL'] = 'egl'

import numpy as np
from dm_control import suite
from dm_env import specs

from gymnasium import spaces
from gymnasium.envs.registration import EnvSpec


class DMControlSuiteEnv:
    def __init__(self, name, task, render_size=(64, 64), action_repeat=2, image_observations=True, camera_id=None, dataset_name=None):

        self.dataset_name = dataset_name

        self.env = suite.load(domain_name=name, task_name=task)

        self.action_space = self.convert_spec_to_space(self.env.action_spec())

        if image_observations:
            self.observation_space = spaces.Box(0, 255, (3,) + render_size, dtype=np.uint8)
        else:
            self.observation_space = self.convert_ordered_dict_to_space(self.env.observation_spec())

        self.spec = EnvSpec(id=name, entry_point=None, max_episode_steps=int(self.env._step_limit / action_repeat))
        self.render_size = render_size
        self.action_repeat = action_repeat
        self.image_observations = image_observations

        if camera_id in (-1, None):
            self.camera_id = dict(
                quadruped_walk=2, quadruped_run=2, quadruped_escape=2,
                quadruped_fetch=2, locom_rodent_maze_forage=1,
                locom_rodent_two_touch=1,
            ).get(name, 0)

    def normalize_reward(self, reward):
        return reward / 1000 * 100

    def preprocess_observation(self, observation):
        if self.image_observations:
            return observation / 255.0
        else:
            return observation

    def step(self, action):
        reward = 0.0
        for _ in range(self.action_repeat):
            timestep = self.env.step(action)
            reward += timestep.reward or 0.0
            if timestep.last():
                break

        if self.image_observations:
            observation = self.env.physics.render(*self.render_size, camera_id=self.camera_id)
            observation = np.moveaxis(observation, -1, 0)
        else:
            observation = np.concatenate([v.flatten() for v in timestep.observation.values()])

        terminated = timestep.last()
        truncated = False
        info = {}

        return observation, reward, terminated, truncated, info

    def reset(self, *args, **kwargs):
        timestep = self.env.reset()
        if self.image_observations:
            observation = self.env.physics.render(*self.render_size, camera_id=self.camera_id)
            observation = np.moveaxis(observation, -1, 0)
        else:
            observation = np.concatenate([v.flatten() for v in timestep.observation.values()])

        return observation, {}

    def convert_spec_to_space(self, spec):
        # Box
        if type(spec) is specs.Array:
            return spaces.Box(-np.inf, np.inf, shape=spec.shape)
        elif type(spec) is specs.BoundedArray:
            _min = spec.minimum
            _max = spec.maximum

            if np.isscalar(_min) and np.isscalar(_max):
                # same min and max for every element
                return spaces.Box(_min, _max, shape=spec.shape)
            else:
                # different min and max for every element
                return spaces.Box(_min + np.zeros(spec.shape), _max + np.zeros(spec.shape))
        else:
            raise ValueError("Unknown spec!")

    def convert_ordered_dict_to_space(self, odict):
        if len(odict.keys()) == 1:
            # no concatenation
            return self.convert_spec_to_space(list(odict.values())[0])
        else:
            # concatentation
            numdim = sum([np.prod(odict[key].shape) for key in odict])
            return spaces.Box(-np.inf, np.inf, shape=(int(numdim),))

