import gymnasium as gym
import jax
import jax.numpy as jnp
import numpy as np
from gymnasium.spaces import Box


class NEnvs(gym.Wrapper):
    def __init__(self, cfg, n_envs=4):
        from rlab_jax.envs.env import create_single_env_from_cfg

        self.n_envs = n_envs
        self.envs = [create_single_env_from_cfg(cfg) for _ in range(n_envs)]

        if hasattr(cfg, "image_observations"):
            self.image_observations = cfg.image_observations
        else:
            self.image_observations = False

        super().__init__(env=self.envs[0])

    def step(self, action):
        observations, rewards, terminateds, truncateds, infos = [], [], [], [], []

        for i, env in enumerate(self.envs):
            observation, reward, terminated, truncated, info = env.step(action[i])
            observations.append(observation)
            rewards.append(reward)
            terminateds.append(terminated)
            truncateds.append(truncated)
            infos.append(info)

        return (
            np.stack(observations),
            np.stack(rewards).reshape(-1, 1),
            np.stack(terminateds),
            np.stack(truncateds),
            infos,
        )

    def reset(self, seed=None, *args, **kwargs):
        observations, infos = [], []

        if seed is not None:
            seeds = jax.random.split(seed, self.n_envs)

        for i, env in enumerate(self.envs):
            if seed is not None:
                observation, info = env.reset(seed=int(seeds[i, 0]), *args, **kwargs)
            else:
                observation, info = env.reset(*args, **kwargs)
            observations.append(observation)
            infos.append(info)

        return np.stack(observations), infos


class PreprocessObservation(gym.Wrapper):
    def preprocess_observation(self, observation):
        return observation


class TimeLimitFix(gym.Wrapper):
    def __init__(self, env, max_episode_steps=None):
        super().__init__(env)
        if max_episode_steps is None and self.env.spec is not None:
            max_episode_steps = env.spec.max_episode_steps
        if self.env.spec is not None:
            self.env.spec.max_episode_steps = max_episode_steps
        self._max_episode_steps = max_episode_steps
        self._elapsed_steps = None

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        self._elapsed_steps += 1
        if self._elapsed_steps >= self._max_episode_steps:
            truncated = True
        else:
            truncated = False
        return observation, reward, done, truncated, info

    def reset(self, **kwargs):
        self._elapsed_steps = 0
        return self.env.reset(**kwargs)


class GymToGymnasiumEnv(gym.Wrapper):
    def reset(self, *args, **kwargs):
        observation = self.env.reset(*args, **kwargs)
        return observation, {}

    def preprocess_observation(self, observation):
        return observation


class GymToGymnasiumEnvNoResetSeed(gym.Wrapper):
    def reset(self, *args, **kwargs):
        observation = self.env.reset()
        return observation, {}

    def preprocess_observation(self, observation):
        return observation


class RescaleAction(gym.ActionWrapper):
    def __init__(self, env: gym.Env):
        super().__init__(env)

        # save old action space
        self.old_min_action = np.copy(self.action_space.low)
        self.old_max_action = np.copy(self.action_space.high)

        # update action space
        self.action_space.high = np.ones_like(self.action_space.high)
        self.action_space.low = -np.ones_like(self.action_space.low)
        self.action_space.high_repr = "1.0"
        self.action_space.low_repr = "-1.0"

        self.min_action = np.array(self.action_space.low)
        self.max_action = np.array(self.action_space.high)

    def action(self, action):
        action = self.old_min_action + (self.old_max_action - self.old_min_action) * (
            (action - self.min_action) / (self.max_action - self.min_action)
        )
        action = np.clip(a=action, a_min=self.old_min_action, a_max=self.old_max_action)
        return action


class Normalizer:
    def __init__(self, min, max):
        self.min = min
        self.max = max

    def forward(self, x):
        x_norm = (x - self.min) / (self.max - self.min)
        return x_norm * 2 - 1

    def backward(self, x):
        return (x + 1) / 2 * (self.max - self.min) + self.min


class Normalize(gym.Wrapper):
    def __init__(self, env, constraints):
        super().__init__(env)
        self.env = env
        self.constraints = constraints

        self.observation_normalizer = Normalizer(
            np.array(self.constraints["observation"]["min"]),
            np.array(self.constraints["observation"]["max"]),
        )
        self.reward_normalizer = Normalizer(
            self.constraints["reward"]["min"], self.constraints["reward"]["max"]
        )

    def reset(self, *args, **kwargs):
        observation, info = self.env.reset(*args, **kwargs)
        return self.observation_normalizer.forward(observation).astype(np.float32), info

    def step(self, *args, **kwargs):
        next_observation, reward, terminated, truncated, info = self.env.step(*args, **kwargs)
        return (
            self.observation_normalizer.forward(next_observation).astype(np.float32),
            self.reward_normalizer.forward(reward).astype(np.float32),
            terminated,
            truncated,
            info,
        )

    def denormalize_rewards(self, rewards):
        return self.reward_normalizer.backward(rewards)


class ActionRepeat(gym.Wrapper):
    def __init__(self, env, repeat):
        super().__init__(env)

        self.env = env
        self.env.spec.max_episode_steps = int(self.env.spec.max_episode_steps / repeat)
        self.repeat = repeat

    def step(self, action):
        done = False
        total_reward = 0
        current_step = 0
        while current_step < self.repeat and not done:
            observation, reward, terminated, truncated, info = self.env.step(action)
            done = terminated or truncated
            total_reward += reward
            current_step += 1
        return observation, total_reward, terminated, truncated, info


