import random
from abc import abstractmethod

import equinox as eqx
import jax
import jax.numpy as jnp


class Evaluation:
    subclasses = {}
    _requires_multisample_predictions = False
    _requires_multisample_filtered_predictions = False
    _requires_multisample_filtered_reward_predictions = False
    _requires_multisample_filtered_done_predictions = False
    _requires_dream_rollout = False
    _requires_predictions = False

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "_EVALUATION_TYPE"):
            cls.subclasses[cls._EVALUATION_TYPE] = cls

    @classmethod
    def create(cls, evaluation_type, *args, **kwargs):
        if evaluation_type not in cls.subclasses:
            raise ValueError(f"Bad evaluation type {evaluation_type}")

        evaluation = cls.subclasses[evaluation_type](*args, **kwargs)

        return evaluation

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


class ExperimentEvaluation:
    def __init__(self, cfg, env, agent, sequence_model, writer, evaluations_every_n_epochs, *args, **kwargs):
        self.env = env
        self.agent = agent
        self.sequence_model = sequence_model
        self.writer = writer
        self.evaluations_every_n_epochs = evaluations_every_n_epochs

        self.evaluations = [
            Evaluation.create(
                evaluation_type=evaluation_type,
                cfg=cfg,
                writer=writer,
                env=env,
                agent=agent,
                sequence_model=sequence_model,
                *args,
                **kwargs,
            )
            for evaluation_type in cfg.evaluation.types
        ]

    def __call__(
        self,
        batch,
        key: jax.random.PRNGKey,
        step: int,
        epoch: int,
        model_state=None,
        agent_state=None,
        reward_state=None,
        prefix=""
    ):
        for evaluation in self.evaluations:
            if hasattr(evaluation, "_BATCH_ONLY") and evaluation._BATCH_ONLY is True:
                key, subkey = jax.random.split(key)
                evaluation(batch=batch, step=step, key=subkey, model_state=model_state, agent_state=agent_state, prefix=prefix)
