import copy

import equinox as eqx
import jax.random
import matplotlib.pyplot as plt
import omegaconf
import tqdm
from tensorboardX import SummaryWriter

from rlab_jax.evaluation.evaluation import ExperimentEvaluation
from rlab_jax.utils.cfg import save_cfg
from rlab_jax.utils.keygen import keygen
from rlab_jax.utils.logs import add_log, mean_logs
from rlab_jax.utils.memory import ReplayMemory, SequenceReplayMemory, ImageSequenceReplayMemory
from rlab_jax.utils.rollout import BatchRollout, DreamRollout, MemoryRollout
from rlab_jax.rlab_dataclasses import ModulesState

from .experiment import Experiment


class ModelBasedDreamerExperiment(Experiment):
    _EXPERIMENT_TYPE = "ModelBasedDreamer"

    def __init__(
        self,
        agent,
        sequence_model,
        train_env,
        val_env,
        writer,
        replay_memory_device,
        replay_memory_init,
        replay_memory_size,
        window_length,
        epochs,
        train_rollouts_per_epoch,
        env_steps_per_rollout,
        model_train_steps_ratio,
        agent_train_steps_ratio,
        sequence_model_warmup_epochs,
        batch_size,
        val_every_n_epochs,
        checkpoint_every_n_epochs,
        dream_rollout_batch_size,
        dream_rollout_steps,
        log_dir,
        load_dataset_path,
        save_dataset_path,
        evaluation=None,
        checkpoint_path=None,
        observation_normalizer=False,
        random_replay_memory_init=None,
        validation_ratio=None,
        validation_batch_size=None,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.agent = agent
        self.sequence_model = sequence_model
        self.writer = writer
        self.evaluation = evaluation
        self.train_env = train_env
        self.val_env = val_env
        self.replay_memory_device = replay_memory_device
        self.replay_memory_init = replay_memory_init
        self.replay_memory_size = replay_memory_size
        self.epochs = epochs
        self.train_rollouts_per_epoch = train_rollouts_per_epoch
        self.env_steps_per_rollout = env_steps_per_rollout
        self.model_train_steps_ratio = model_train_steps_ratio
        self.agent_train_steps_ratio = agent_train_steps_ratio
        self.sequence_model_warmup_epochs = sequence_model_warmup_epochs
        self.batch_size = batch_size
        self.val_every_n_epochs = val_every_n_epochs
        self.checkpoint_every_n_epochs = checkpoint_every_n_epochs
        self.window_length = window_length
        self.dream_rollout_batch_size = dream_rollout_batch_size
        self.dream_rollout_steps = dream_rollout_steps
        self.log_dir = log_dir
        self.load_dataset_path = load_dataset_path
        self.save_dataset_path = save_dataset_path
        self.checkpoint_path = checkpoint_path
        self.random_replay_memory_init = random_replay_memory_init
        self.validation_ratio = validation_ratio
        self.validation_batch_size = validation_batch_size

        if env_steps_per_rollout is None:
            self.env_steps_per_rollout = train_env.spec.max_episode_steps

        if train_env.image_observations:
            self.train_rollout = MemoryRollout(
                env=train_env,
                memory=ImageSequenceReplayMemory(
                    size=int(self.replay_memory_size / train_env.spec.max_episode_steps),
                    env=self.train_env,
                    window_size=self.window_length,
                ),
                act_fn=agent.act,
            )
        else:
            self.train_rollout = MemoryRollout(
                env=train_env,
                memory=SequenceReplayMemory(
                    size=int(self.replay_memory_size / train_env.spec.max_episode_steps),
                    env=self.train_env,
                    window_size=self.window_length,
                ),
                act_fn=agent.act,
            )

        self.train_rollout.filter_fn = sequence_model.filter

        self.val_rollout = BatchRollout(env=val_env, act_fn=agent.evaluate, sequence_rollout=True)
        self.val_rollout.filter_fn = sequence_model.filter

        self.dream_rollout = DreamRollout(
            sequence_model=sequence_model,
            act_fn=agent.dream_act,
            n_rollouts=self.dream_rollout_batch_size,
            steps_per_rollout=self.dream_rollout_steps,
        )
        self.agent.dream_generate_fn = self.dream_rollout.generate

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, **_ignored):
        from rlab_jax.envs.env import env_from_cfg
        from rlab_jax.modules import Agent, SequenceModel
        from rlab_jax.utils.logs import get_outputs_path

        train_env = env_from_cfg(cfg.env, n_envs=cfg.experiment.train_rollouts_per_epoch)
        val_env = env_from_cfg(cfg.env, n_envs=cfg.experiment.val_rollouts_per_epoch)

        sequence_model = SequenceModel.create(
            cfg=cfg.seqm,
            n_observations=train_env.observation_space.shape if len(train_env.observation_space.shape) > 1 else train_env.observation_space.shape[0],
            n_actions=train_env.action_space.shape[0],
        )

        agent = Agent.create(
            cfg=cfg.agent,
            n_inputs=sequence_model.n_latent,
            n_actions=train_env.action_space.shape[0],
            sequence_model=sequence_model,
        )

        log_dir = get_outputs_path(cfg)
        writer = SummaryWriter(log_dir=log_dir)
        save_cfg(log_dir, cfg)

        evaluation = ExperimentEvaluation(
            cfg=cfg,
            env=val_env,
            agent=agent,
            sequence_model=sequence_model,
            writer=writer,
            evaluations_every_n_epochs=cfg.evaluation.every_n_epochs
        )

        print(
            f"""
        ################################
        Env: {cfg.env.name}
        Experiment: {cfg.experiment.type}
        SequenceModel: {cfg.seqm.type}
        Agent: {cfg.agent.type}
        LogDir: {log_dir}
        ################################
        """
        )

        return cls(
            train_env=train_env,
            val_env=val_env,
            agent=agent,
            sequence_model=sequence_model,
            writer=writer,
            evaluation=evaluation,
            log_dir=log_dir,
            **cfg.experiment,
        )

    @eqx.filter_jit
    def agent_training_epoch(
        self, memory_state, model_state, agent_state, num_updates_per_epoch, key
    ):
        print("agent training compiling")

        agent_params, agent_static = eqx.partition(agent_state, eqx.is_array)

        def update(carry, _):
            agent_params_, key_ = carry

            keys_ = jax.random.split(key_, 5)

            agent_state_ = eqx.combine(agent_params_, agent_static)

            # Sample batch
            batch_ = self.train_rollout.memory.sample(
                memory_state=memory_state, batch_size=self.batch_size, key=keys_[0]
            )

            # Predictions
            predictions_ = self.sequence_model(
                model=model_state.module,
                observation_trajectory=batch_.observation_trajectory,
                action_trajectory=batch_.action_trajectory,
                key=keys_[1],
                scaler=model_state.scaler,
            )

            # Agent training step
            agent_state_, agent_log_ = self.agent.training_step(
                model_state=model_state,
                agent_state=agent_state_,
                batch=batch_,
                predictions=predictions_,
                key=keys_[3],
            )

            agent_params_, _ = eqx.partition(agent_state_, eqx.is_array)

            return (agent_params_, keys_[4]), {**agent_log_}

        init_carry = (agent_params, key)
        (agent_params, key), train_logs = jax.lax.scan(
            f=update, init=init_carry, xs=None, length=num_updates_per_epoch
        )

        agent_state = eqx.combine(agent_params, agent_static)

        train_logs = jax.tree_map(lambda v: v.mean(), train_logs)

        return agent_state, train_logs

    @eqx.filter_jit
    def sequence_model_validation_epoch(
        self, memory_state, model_state, key
    ):
        print("sequence model evaluation compiling")

        keys = jax.random.split(key)

        # Sample batch
        batch = self.train_rollout.memory.sample(
            memory_state=memory_state, batch_size=self.batch_size, key=keys[0], validation=True
        )

        # Model training step
        model_state, predictions, model_log = self.sequence_model.training_step(
            model_state=model_state, batch=batch, key=keys[1]
        )

        return model_log

    @eqx.filter_jit
    def sequence_model_training_epoch(
        self, memory_state, model_state, num_updates_per_epoch, key, agent_state=None
    ):
        print("sequence model training compiling")

        model_params, model_static = eqx.partition(model_state, eqx.is_array)

        def update(carry, _):
            model_params_, key_ = carry

            keys_ = jax.random.split(key_, 5)

            model_state_ = eqx.combine(model_params_, model_static)
            # agent_state_ = eqx.combine(agent_params_, agent_static)

            # Sample batch
            batch_ = self.train_rollout.memory.sample(
                memory_state=memory_state, batch_size=self.batch_size, key=keys_[0]
            )

            # Model training step

            model_state_, predictions_, model_log_ = self.sequence_model.training_step(
                model_state=model_state_, batch=batch_, key=keys_[1]
            )

            model_params_, _ = eqx.partition(model_state_, eqx.is_array)

            return (model_params_, keys_[4]), {**model_log_}

        init_carry = (model_params, key)
        (model_params, key), train_logs = jax.lax.scan(
            f=update, init=init_carry, xs=None, length=num_updates_per_epoch
        )
        model_state = eqx.combine(model_params, model_static)
        # agent_state = eqx.combine(agent_params, agent_static)

        train_logs = jax.tree_map(lambda v: v.mean(), train_logs)

        return model_state, train_logs

    def initial_state(self):
        modules_state = ModulesState(agent_state=self.agent.initial_state(),
                                     model_state=self.sequence_model.initial_state())

        if self.checkpoint_path is not None:
            modules_state = self.load_checkpoint(path=self.checkpoint_path, modules_state=modules_state)

        return modules_state

    def train(self):
        modules_state = self.initial_state()
        memory_state = self.train_rollout.memory.initial_state(key=keygen())

        key, subkey = jax.random.split(keygen())

        if self.random_replay_memory_init:
            @eqx.filter_jit
            def random_act(key, *args, **kwargs):
                key, subkey = jax.random.split(key)
                action, latent_action = self.agent.act(key=key, *args, **kwargs)
                action.sample = jax.random.uniform(key=subkey, shape=action.sample.shape, minval=-1.0, maxval=1.0)
                return action, latent_action

            self.train_rollout.act_fn = random_act

        memory_state = self.train_rollout.generate(
            memory_state=memory_state,
            model_state=modules_state.model_state,
            agent_state=modules_state.agent_state,
            steps_per_rollout=int(self.replay_memory_init / self.train_rollouts_per_epoch),
            key=subkey,
        )

        if self.random_replay_memory_init:
            self.train_rollout.act_fn = self.agent.act

        for epoch in tqdm.tqdm(range(1, self.epochs + 1), desc="epochs"):
            step = (
                self.replay_memory_init
                + epoch * self.train_rollouts_per_epoch * self.env_steps_per_rollout
            )

            # TRAIN ROLLOUT
            key, subkey = jax.random.split(key)
            memory_state = self.train_rollout.generate(
                memory_state=memory_state,
                model_state=modules_state.model_state,
                agent_state=modules_state.agent_state,
                steps_per_rollout=self.env_steps_per_rollout,
                key=subkey,
            )

            # SEQUENCE MODEL & REWARD TRAINING
            num_updates_per_epoch = int(
                self.train_rollouts_per_epoch
                * self.env_steps_per_rollout
                * self.model_train_steps_ratio
            )
            key, subkey = jax.random.split(key)
            modules_state.model_state, train_logs = self.sequence_model_training_epoch(
                memory_state=memory_state,
                model_state=modules_state.model_state,
                agent_state=modules_state.agent_state,
                num_updates_per_epoch=num_updates_per_epoch,
                key=subkey,
            )
            for k, v in train_logs.items():
                self.writer.add_scalar(k, v.item(), step)

            # AGENT TRAINING
            if epoch >= self.sequence_model_warmup_epochs:
                num_updates_per_epoch = int(
                    self.train_rollouts_per_epoch
                    * self.env_steps_per_rollout
                    * self.agent_train_steps_ratio
                )
                key, subkey = jax.random.split(key)
                modules_state.agent_state, train_logs = self.agent_training_epoch(
                    memory_state=memory_state,
                    agent_state=modules_state.agent_state,
                    model_state=modules_state.model_state,
                    num_updates_per_epoch=num_updates_per_epoch,
                    key=subkey,
                )

                for k, v in train_logs.items():
                    self.writer.add_scalar(k, v.item(), step)

            # VALIDATION
            if epoch % self.val_every_n_epochs == 0:

                # VAL ROLLOUT
                key, subkey = jax.random.split(key)
                batch = self.val_rollout.generate(
                    model_state=modules_state.model_state,
                    agent_state=modules_state.agent_state,
                    steps_per_rollout=self.env_steps_per_rollout,
                    key=subkey,
                )

                if self.evaluation is not None:
                    key, subkey = jax.random.split(key)
                    self.evaluation(
                        batch=batch,
                        model_state=modules_state.model_state,
                        agent_state=modules_state.agent_state,
                        key=subkey,
                        step=step,
                        epoch=epoch
                    )

            if epoch % self.checkpoint_every_n_epochs == 0:
                self.save_checkpoint(step=step, modules_state=modules_state)

