import chex
import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf
import optax

from rlab_jax.modules.agents.agent import Agent
from rlab_jax.modules.networks import Mlp, Module
from rlab_jax.rlab_dataclasses import TrainState, Transition, RandomVariable
from rlab_jax.utils.keygen import keygen


@chex.dataclass
class DDPGAgentState:
    policy: TrainState
    critic: TrainState
    target_critic: TrainState
    target_policy: TrainState


class DDPGAgent(Agent):
    _AGENT_TYPE = "DDPGAgent"

    def __init__(
        self,
        policy: Module,
        critic: Module,
        target_policy: Module,
        target_critic: Module,
        discount_factor: float = 0.99,
        policy_optim_type: str = "optax.adam",
        critic_optim_type: str = "optax.adam",
        policy_optim_settings: dict = {"learning_rate": 1e-4},
        critic_optim_settings: dict = {"learning_rate": 3e-4},
        target_policy_lr: float = 1e-3,
        target_critic_lr: float = 5e-3,
        action_l2_weighting: float = 1.0,
        random_action_prob: float = 0.3,
    ):
        super().__init__()

        self.discount_factor = discount_factor

        policy_optim_settings = dict(policy_optim_settings)
        if "lr" in policy_optim_settings.keys():
            policy_optim_settings["learning_rate"] = policy_optim_settings["lr"]
            del policy_optim_settings["lr"]

        critic_optim_settings = dict(critic_optim_settings)
        if "lr" in critic_optim_settings.keys():
            critic_optim_settings["learning_rate"] = critic_optim_settings["lr"]
            del critic_optim_settings["lr"]

        policy_optim = eval(policy_optim_type)(**policy_optim_settings)
        critic_optim = eval(critic_optim_type)(**critic_optim_settings)

        self.target_policy_lr = target_policy_lr
        self.target_critic_lr = target_critic_lr

        self.action_l2_weighting = action_l2_weighting
        self.random_action_prob = random_action_prob

        policy_state = TrainState(
            module=policy,
            optim=policy_optim,
            optim_state=policy_optim.init(eqx.filter(policy, eqx.is_array)),
        )

        critic_state = TrainState(
            module=critic,
            optim=critic_optim,
            optim_state=critic_optim.init(eqx.filter(critic, eqx.is_array)),
        )

        target_critic_state = TrainState(
            module=target_critic, optim=jnp.zeros(1), optim_state=jnp.zeros(1)
        )

        target_policy_state = TrainState(
            module=target_policy, optim=jnp.zeros(1), optim_state=jnp.zeros(1)
        )

        self.agent_state = DDPGAgentState(
            policy=policy_state,
            critic=critic_state,
            target_critic=target_critic_state,
            target_policy=target_policy_state,
        )

    def initial_state(self):
        return self.agent_state

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, n_inputs, n_actions, **_ignored):
        return cls(
            policy=Module.create(
                type=cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_actions,
                key=keygen(),
                activation=jax.nn.tanh,
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
            ),
            target_policy=Module.create(
                type=cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_actions,
                key=keygen(),
                activation=jax.nn.tanh,
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
            ),
            critic=Module.create(
                "Mlp",
                n_inputs=n_inputs + n_actions,
                n_outputs=1,
                key=keygen(),
                activation=None,
                layers=cfg.critic.layers,
                units=cfg.critic.units,
                hidden_activation=cfg.critic.hidden_activation,
            ),
            target_critic=Module.create(
                "Mlp",
                n_inputs=n_inputs + n_actions,
                n_outputs=1,
                key=keygen(),
                activation=None,
                layers=cfg.critic.layers,
                units=cfg.critic.units,
                hidden_activation=cfg.critic.hidden_activation,
            ),
            discount_factor=cfg.reward_discount_factor,
            policy_optim_type=cfg.policy.optim.type,
            critic_optim_type=cfg.critic.optim.type,
            policy_optim_settings=cfg.policy.optim.settings,
            critic_optim_settings=cfg.critic.optim.settings,
            target_policy_lr=cfg.policy.target_lr,
            target_critic_lr=cfg.critic.target_lr,
            action_l2_weighting=cfg.action_l2_weighting,
            random_action_prob=cfg.random_action_prob,
        )

    @staticmethod
    def compute_critic_values(critic: Module, states: jnp.array, actions: jnp.array):
        return critic(jnp.concatenate((states, actions), -1))

    @staticmethod
    def soft_update(network, target, tau):
        model_params = eqx.filter(network, eqx.is_array)
        target_model_params, target_model_static = eqx.partition(target, eqx.is_array)
        new_target_params = optax.incremental_update(model_params, target_model_params, tau)
        return eqx.combine(new_target_params, target_model_static)

    @staticmethod
    @eqx.filter_jit
    def act(
        agent_state: DDPGAgentState, observation: jnp.array, deterministic=False, key=None, **ignored
    ):
        action = agent_state.policy.module(observation)

        # if not deterministic:
        #     action += torch.normal(
        #         mean=torch.zeros_like(action), std=torch.ones_like(action) * self.exploration_noise
        #     )
        #
        #     action = torch.clamp(action, -1, 1)
        #
        #     if torch.rand(1) < self.random_action_prob:
        #         action = action.uniform_(-1, 1)

        return RandomVariable(sample=action, mean=action, std=jnp.zeros_like(action)), None

    @eqx.filter_jit
    def evaluate(
        self, agent_state: DDPGAgentState, observation: jnp.array, deterministic=True, **ignored
    ):
        return self.act(agent_state=agent_state, observation=observation, deterministic=True, **ignored)

    def policy_loss(self, policy: Module, critic: Module, batch: Transition):
        action = policy(batch.observation)
        q_estimates = self.compute_critic_values(critic, batch.observation, action)

        loss = -q_estimates
        loss = loss.mean() + self.action_l2_weighting * jnp.mean(action**2)

        return loss, {"agent/policy/loss": loss}

    def critic_loss(
        self, critic: Module, target_critic: Module, target_policy: Module, batch: Transition
    ):
        target_action = target_policy(batch.next_observation)
        target_q = self.compute_critic_values(target_critic, batch.next_observation, target_action)

        target = batch.reward + self.discount_factor * target_q

        q_estimates = self.compute_critic_values(critic, batch.observation, batch.action)

        loss = (q_estimates - target) ** 2
        loss = loss.mean()

        return loss, {"agent/q/loss": loss}

    def training_step(
        self, agent_state: DDPGAgentState, batch: Transition, key: jax.random.PRNGKey
    ):
        (_, policy_log), policy_grads = eqx.filter_value_and_grad(self.policy_loss, has_aux=True)(
            agent_state.policy.module, critic=agent_state.critic.module, batch=batch
        )
        (_, critic_log), critic_grads = eqx.filter_value_and_grad(self.critic_loss, has_aux=True)(
            agent_state.critic.module,
            target_critic=agent_state.target_critic.module,
            target_policy=agent_state.target_policy.module,
            batch=batch,
        )

        agent_state.policy = agent_state.policy.apply_updates(policy_grads)
        agent_state.critic = agent_state.critic.apply_updates(critic_grads)

        log = {}
        log.update(policy_log)
        log.update(critic_log)

        agent_state.target_critic.module.mlp = self.soft_update(
            agent_state.critic.module.mlp,
            agent_state.target_critic.module.mlp,
            self.target_critic_lr,
        )
        # TODO: Only works for MLP policies
        agent_state.target_policy.module.mlp = self.soft_update(
            agent_state.policy.module.mlp,
            agent_state.target_policy.module.mlp,
            self.target_policy_lr,
        )

        return agent_state, log
