import chex
import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf
import optax
from rlax import lambda_returns, n_step_bootstrapped_returns

from rlab_jax.modules.agents.agent import Agent
from rlab_jax.modules.networks import Mlp, Module
from rlab_jax.rlab_dataclasses import (DreamRolloutTransition, TrainState,
                                       Transition, random_variable)
from rlab_jax.utils.keygen import keygen
from rlab_jax.distributions import multivariate_normal_diag


@chex.dataclass
class MBACAgentState:
    policy: TrainState
    critic: TrainState
    target_critic: TrainState


class MBACAgent(Agent):
    _AGENT_TYPE = "MBACAgent"
    _requires_dream_rollout = True

    def __init__(
        self,
        policy: Module,
        critic: Module,
        target_critic: Module,
        policy_optim_type: str = "optax.adam",
        critic_optim_type: str = "optax.adam",
        policy_optim_settings: dict = {"learning_rate": 1e-4},
        critic_optim_settings: dict = {"learning_rate": 3e-4},
        policy_max_gradient_norm: int = 100,
        critic_max_gradient_norm: int = 100,
        reward_weighting: float = 0.01,
        entropy_scale: float = None, #1e-4,
        return_estimator_cfg=None,
        target_critic_lr: float = 5e-3,
    ):
        super().__init__()

        self.policy_reward_weighting = reward_weighting
        self.critic_reward_weighting = reward_weighting

        policy_optim_settings = dict(policy_optim_settings)
        if "lr" in policy_optim_settings.keys():
            policy_optim_settings["learning_rate"] = policy_optim_settings["lr"]
            del policy_optim_settings["lr"]

        critic_optim_settings = dict(critic_optim_settings)
        if "lr" in critic_optim_settings.keys():
            critic_optim_settings["learning_rate"] = critic_optim_settings["lr"]
            del critic_optim_settings["lr"]

        policy_optim_settings["learning_rate"] = float(policy_optim_settings["learning_rate"])
        critic_optim_settings["learning_rate"] = float(critic_optim_settings["learning_rate"])

        policy_optim = optax.chain(optax.clip_by_global_norm(policy_max_gradient_norm), eval(policy_optim_type)(**policy_optim_settings))
        critic_optim = optax.chain(optax.clip_by_global_norm(critic_max_gradient_norm), eval(critic_optim_type)(**critic_optim_settings))

        self.target_critic_lr = target_critic_lr
        self.entropy_scale = entropy_scale

        self.return_estimator_cfg = return_estimator_cfg

        policy = TrainState(
            module=policy,
            optim=policy_optim,
            optim_state=policy_optim.init(eqx.filter(policy, eqx.is_array)),
        )

        critic = TrainState(
            module=critic,
            optim=critic_optim,
            optim_state=critic_optim.init(eqx.filter(critic, eqx.is_array)),
        )

        target_critic = TrainState(
            module=target_critic, optim=jnp.zeros(1), optim_state=jnp.zeros(1)
        )

        self.agent_state = MBACAgentState(policy=policy, critic=critic, target_critic=target_critic)

    @staticmethod
    def compute_critic_values(critic: Module, states: jnp.array):
        @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
        def evaluate_ensemble(model, x):
            return model(x)

        return evaluate_ensemble(critic, states)

    def initial_state(self):
        return self.agent_state

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, n_inputs, n_actions, **_ignored):
        @eqx.filter_vmap
        def critic_networks(key):
            return Mlp(
                n_inputs=n_inputs,
                n_outputs=1,
                key=key,
                activation=None,
                layers=cfg.critic.layers,
                units=cfg.critic.units,
                hidden_activation=cfg.critic.hidden_activation,
            )

        return cls(
            policy=Module.create(
                cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_actions,
                key=keygen(),
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
                init_std=cfg.policy.init_std,
            ),
            critic=critic_networks(jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))),
            target_critic=critic_networks(jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))),
            policy_optim_type=cfg.policy.optim.type,
            critic_optim_type=cfg.critic.optim.type,
            policy_optim_settings=cfg.policy.optim.settings,
            critic_optim_settings=cfg.critic.optim.settings,
            reward_weighting=cfg.reward_weighting,
            return_estimator_cfg=cfg.return_estimator,
            target_critic_lr=cfg.critic.target_lr,
            entropy_scale=cfg.entropy_scale,
            policy_max_gradient_norm=cfg.policy.optim.max_gradient_norm,
            critic_max_gradient_norm=cfg.critic.optim.max_gradient_norm
        )

    @staticmethod
    def soft_update(network, target, tau):
        model_params = eqx.filter(network, eqx.is_array)
        target_model_params, target_model_static = eqx.partition(target, eqx.is_array)
        new_target_params = optax.incremental_update(model_params, target_model_params, tau)
        return eqx.combine(new_target_params, target_model_static)

    def torch_return(self, rewards, values):
        discount_factor = self.return_estimator_cfg["discount_factor"]
        horizon = self.return_estimator_cfg["horizon"]

        def single_monte_carlo_return(rewards, discount_factor):
            discounts = discount_factor ** jnp.arange(rewards.shape[0])
            discounted_rewards = discounts * rewards
            return discounted_rewards.sum(0)

        returns = []
        for t in range(rewards.shape[0]):
            returns.append(
                single_monte_carlo_return(rewards[t : t + horizon], discount_factor=discount_factor)
            )
            effective_horizon = min(t + horizon, rewards.shape[0]) - t
            returns[t] += (
                discount_factor**effective_horizon
                * values[min(t + horizon - 1, values.shape[0] - 1)]
            )

        return jnp.stack(returns)

    def penalty(self, weighting, dream_rollout, *args, **kwargs):
        return weighting * dream_rollout.action.dist.log_prob(dream_rollout.action.sample)

    def entropy_penalty(self, dream_rollout):
        return self.entropy_scale * dream_rollout.action.dist.entropy()

    def critic_loss(
        self, critic: Module, target_critic: Module, dream_rollout: DreamRolloutTransition
    ):
        if self.critic_reward_weighting != 0:
            penalty = self.penalty(self.critic_reward_weighting, dream_rollout)
        else:
            penalty = jnp.zeros(1)

        def critic_return_fn(r_t, v_t, discount_t):
            if self.return_estimator_cfg["type"] == "n-step":
                return n_step_bootstrapped_returns(
                    r_t=r_t,
                    v_t=v_t,
                    discount_t=discount_t,
                    n=self.return_estimator_cfg["horizon"],
                    stop_target_gradients=True,
                )
            elif self.return_estimator_cfg["type"] == "lambda":
                return lambda_returns(r_t=r_t,
                                      v_t=v_t,
                                      discount_t=discount_t,
                                      lambda_=jnp.array([self.return_estimator_cfg["lambda"]]),
                                      stop_target_gradients=True)
            else:
                raise ValueError

        # t0 - tN-1
        d = (1 - dream_rollout.done.squeeze(-1)) * self.return_estimator_cfg["discount_factor"]
        r = dream_rollout.reward.squeeze(-1) - penalty
        v = jnp.min(self.compute_critic_values(target_critic, dream_rollout.next_state.sample), axis=0).squeeze(-1)
        target_return = jax.vmap(critic_return_fn, in_axes=1, out_axes=1)(r, v, d)

        value_estimate = self.compute_critic_values(critic, dream_rollout.state.sample).squeeze(-1)

        loss = (value_estimate - jax.lax.stop_gradient(target_return)) ** 2
        loss = loss.mean()

        return loss, {"agent/critic/loss": loss, "agent/critic/critic_penalty": penalty.mean(), "agent/critic/value_estimates_mean": value_estimate.mean()}

    def policy_loss(
        self,
        policy: Module,
        critic: Module,
        model_state,
        batch,
        predictions,
        key: jax.random.PRNGKey,
    ):
        dream_rollout, dream_rollout_log = self.dream_generate_fn_wrapped(
            policy=policy,
            model_state=model_state,
            batch=batch,
            predictions=predictions,
            key=key,
        )

        if self.policy_reward_weighting != 0:
            penalty = self.penalty(self.policy_reward_weighting, dream_rollout, policy=policy, model_state=model_state)
        else:
            penalty = jnp.zeros(1)

        def policy_return_fn(r_t, v_t, discount_t):
            if self.return_estimator_cfg["type"] == "n-step":
                return n_step_bootstrapped_returns(
                    r_t=r_t,
                    v_t=v_t,
                    discount_t=discount_t,
                    n=self.return_estimator_cfg["horizon"],
                    stop_target_gradients=False,
                )
            elif self.return_estimator_cfg["type"] == "lambda":
                return lambda_returns(r_t=r_t,
                                      v_t=v_t,
                                      discount_t=discount_t,
                                      lambda_=jnp.array([self.return_estimator_cfg["lambda"]]),
                                      stop_target_gradients=False)
            else:
                raise ValueError

        # t0 - tN-1
        d = (1 - dream_rollout.done.squeeze(-1)) * self.return_estimator_cfg["discount_factor"]
        r = dream_rollout.reward.squeeze(-1) - penalty
        v = jnp.min(self.compute_critic_values(critic, dream_rollout.next_state.sample), axis=0).squeeze(-1)
        est_return = jax.vmap(policy_return_fn, in_axes=1, out_axes=1)(r, v, d)

        if self.entropy_scale is not None:
            ent_loss = self.entropy_penalty(dream_rollout)
        else:
            ent_loss = 0

        loss = -est_return - ent_loss
        loss = loss.mean()

        return loss, (
            {"agent/loss": loss, "agent/loss/policy_penalty": penalty.mean(), "reward/dream": dream_rollout.reward.sum(0).mean(), "agent/loss/entropy": jnp.mean(ent_loss), **dream_rollout_log},
            dream_rollout,
        )

    @staticmethod
    @eqx.filter_jit
    def act(
        latent_state: jnp.array,
        agent_state: MBACAgentState = None,
        policy: Module = None,
        deterministic=False,
        exploration=True,
        key=None,
        **ignored
    ):
        if policy is not None:
            dist = policy(latent_state.sample)
        else:
            dist = agent_state.policy.module(latent_state.sample)

        key, subkey = jax.random.split(key)

        if deterministic:
            sample = dist.mean()
        else:
            sample = dist.sample(key)

        if exploration:
            sample += multivariate_normal_diag(mean=jnp.zeros_like(sample),
                                               std=jnp.ones_like(sample) * 0.3).sample(subkey)

        return random_variable(sample=sample, **dist.params), None

    @eqx.filter_jit
    def evaluate(self, deterministic=True, exploration=False, *args, **kwargs):
        return self.act(deterministic=True, exploration=False, *args, **kwargs)

    @eqx.filter_jit
    def dream_act(self, exploration=False, *args, **kwargs):
        return self.act(exploration=False, *args, **kwargs)

    def training_step(
        self, model_state, agent_state, batch, predictions, key: jax.random.PRNGKey
    ):
        (_, (policy_log, dream_rollout)), policy_grads = eqx.filter_value_and_grad(
            self.policy_loss, has_aux=True
        )(
            agent_state.policy.module,
            critic=agent_state.critic.module,
            model_state=model_state,
            batch=batch,
            predictions=predictions,
            key=key,
        )

        (_, critic_log), critic_grads = eqx.filter_value_and_grad(self.critic_loss, has_aux=True)(
            agent_state.critic.module,
            target_critic=agent_state.target_critic.module,
            dream_rollout=dream_rollout,
        )

        agent_state.policy = agent_state.policy.apply_updates(policy_grads)
        agent_state.critic = agent_state.critic.apply_updates(critic_grads)

        agent_state.target_critic.module = self.soft_update(
            agent_state.critic.module, agent_state.target_critic.module, self.target_critic_lr
        )

        log = {}
        log.update(policy_log)
        log.update(critic_log)

        return agent_state, log

    def dream_generate_fn_wrapped(self, *args, **kwargs):
        return self.dream_generate_fn(*args, **kwargs), {}
