import chex
import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf
import optax

from rlab_jax.modules.agents.agent import Agent
from rlab_jax.modules.networks import Mlp, Module
from rlab_jax.rlab_dataclasses import TrainState, Transition, random_variable
from rlab_jax.utils.keygen import keygen


@chex.dataclass
class SACAgentState:
    policy: TrainState
    critic: TrainState
    target_critic: TrainState
    entropy_coef: TrainState


class SACAgent(Agent):
    _AGENT_TYPE = "SACAgent"

    def __init__(
        self,
        policy: Module,
        critic: Module,
        target_critic: Module,
        discount_factor: float = 0.99,
        entropy_coef: float = 0.01,
        auto_entropy: bool = True,
        policy_optim_type: str = "optax.adam",
        critic_optim_type: str = "optax.adam",
        policy_optim_settings: dict = {"learning_rate": 1e-4},
        critic_optim_settings: dict = {"learning_rate": 3e-4},
        entropy_optim_type: str = "optax.adam",
        entropy_optim_settings: dict = {"learning_rate": 3e-4},
        target_critic_lr: float = 5e-3,
    ):
        super().__init__()

        self.policy = policy
        self.critic = critic
        self.target_critic = target_critic

        self.discount_factor = discount_factor

        self.target_entropy = -float(self.policy.n_outputs)

        class Entropy(eqx.Module):
            value: jax.Array

            def __init__(self, entropy_coef=1.0):
                # TODO: added clip for mopo and mobile
                self.value = jnp.array([jnp.log(entropy_coef)])

            def __call__(self):
                return jnp.clip(jnp.exp(self.value), 0, 1)

            def log(self):
                return jnp.log(jnp.clip(jnp.exp(self.value), 0, 1))

        self.entropy_coef = Entropy(entropy_coef)
        self.auto_entropy = auto_entropy

        policy_optim_settings = dict(policy_optim_settings)
        if "lr" in policy_optim_settings.keys():
            policy_optim_settings["learning_rate"] = policy_optim_settings["lr"]
            del policy_optim_settings["lr"]

        critic_optim_settings = dict(critic_optim_settings)
        if "lr" in critic_optim_settings.keys():
            critic_optim_settings["learning_rate"] = critic_optim_settings["lr"]
            del critic_optim_settings["lr"]

        self.policy_optim = eval(policy_optim_type)(**policy_optim_settings)
        self.critic_optim = eval(critic_optim_type)(**critic_optim_settings)
        self.entropy_coef_optim = eval(entropy_optim_type)(**entropy_optim_settings)

        self.target_critic_lr = target_critic_lr

    def initial_state(self):
        policy = TrainState(
            module=self.policy,
            optim=self.policy_optim,
            optim_state=self.policy_optim.init(eqx.filter(self.policy, eqx.is_array)),
        )

        critic = TrainState(
            module=self.critic,
            optim=self.critic_optim,
            optim_state=self.critic_optim.init(eqx.filter(self.critic, eqx.is_array)),
        )

        target_critic = TrainState(
            module=self.target_critic, optim=jnp.zeros(1), optim_state=jnp.zeros(1)
        )

        entropy_coef = TrainState(
            module=self.entropy_coef,
            optim=self.entropy_coef_optim,
            optim_state=self.entropy_coef_optim.init(eqx.filter(self.entropy_coef, eqx.is_array)),
        )

        return SACAgentState(
            policy=policy, critic=critic, target_critic=target_critic, entropy_coef=entropy_coef
        )

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, n_inputs, n_actions, **_ignored):
        @eqx.filter_vmap
        def q_networks(key):
            return Mlp(
                n_inputs=n_inputs + n_actions,
                n_outputs=1,
                key=key,
                activation=None,
                layers=cfg.critic.layers,
                units=cfg.critic.units,
                hidden_activation=cfg.critic.hidden_activation,
            )

        return cls(
            policy=Module.create(
                type=cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_actions,
                key=keygen(),
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
                init_std=cfg.policy.init_std,
            ),
            critic=q_networks(
                jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))
            ),
            target_critic=q_networks(
                jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))
            ),
            policy_optim_type=cfg.policy.optim.type,
            critic_optim_type=cfg.critic.optim.type,
            policy_optim_settings=cfg.policy.optim.settings,
            critic_optim_settings=cfg.critic.optim.settings,
            discount_factor=cfg.reward_discount_factor,
            entropy_coef=cfg.sac_entropy_coef,
            auto_entropy=cfg.sac_auto_entropy,
            entropy_optim_settings={"learning_rate": cfg.sac_alpha_lr},
            target_critic_lr=cfg.critic.target_lr,
        )

    @staticmethod
    def compute_critic_values(critic: Module, states: jnp.array, actions: jnp.array):
        inputs = jnp.concatenate((states, actions), -1)

        @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
        def evaluate_ensemble(model, x):
            return model(x)

        return evaluate_ensemble(critic, inputs)

    @staticmethod
    def soft_update(network, target, tau):
        model_params = eqx.filter(network, eqx.is_array)
        target_model_params, target_model_static = eqx.partition(target, eqx.is_array)
        new_target_params = optax.incremental_update(model_params, target_model_params, tau)
        return eqx.combine(new_target_params, target_model_static)

    @staticmethod
    @eqx.filter_jit
    def act(agent_state: SACAgentState, observation: jnp.array, deterministic=False, key=None, **ignored):
        dist = agent_state.policy.module(observation)

        if deterministic:
            sample = dist.mean()
        else:
            sample = dist.sample(key)

        return random_variable(sample=sample, **dist.params), None

    @eqx.filter_jit
    def evaluate(self, agent_state: SACAgentState, observation: jnp.array, deterministic=True, **ignored):
        return self.act(agent_state=agent_state, observation=observation, deterministic=True, **ignored)

    def entropy_coef_loss(
        self, entropy_coef: Module, policy: Module, batch: Transition, key: jax.random.PRNGKey
    ):
        action_dist = policy(batch.observation)
        action, log_prob = action_dist.sample_and_log_prob(seed=key)

        loss = -entropy_coef.log() * (log_prob + self.target_entropy)
        loss = loss.mean()

        return loss, {
            "agent/entropy_coef/loss": loss,
            "agent/entropy_coef/diff": (log_prob + self.target_entropy).mean(),
        }

    def policy_loss(
        self,
        policy: Module,
        critic: Module,
        entropy_coef: Module,
        batch: Transition,
        key: jax.random.PRNGKey,
    ):
        action_dist = policy(batch.observation)
        action, log_prob = action_dist.sample_and_log_prob(seed=key)

        q_estimates = jnp.min(self.compute_critic_values(critic, batch.observation, action), axis=0)

        loss = entropy_coef() * log_prob - q_estimates
        loss = jnp.mean(loss)

        return loss, {
            "agent/policy/loss": loss,
            "agent/policy/q_loss": -jnp.mean(q_estimates),
            "agent/entropy_coef/value": entropy_coef(),
        }

    def critic_loss(
        self,
        critic: Module,
        target_critic: Module,
        policy: Module,
        entropy_coef: Module,
        batch: Transition,
        key: jax.random.PRNGKey,
    ):
        next_action_dist = policy(batch.next_observation)
        next_action, next_log_prob = next_action_dist.sample_and_log_prob(seed=key)

        next_q_estimates = jnp.min(
            self.compute_critic_values(target_critic, batch.next_observation, next_action), axis=0
        )

        target_q = next_q_estimates - entropy_coef() * jnp.expand_dims(next_log_prob, -1)

        target = batch.reward + self.discount_factor * target_q

        q_estimates = self.compute_critic_values(critic, batch.observation, batch.action)

        loss = (q_estimates - target) ** 2
        loss = loss.mean()

        return loss, {"agent/q/loss": loss}

    def training_step(self, agent_state: SACAgentState, batch: Transition, key: jax.random.PRNGKey):
        keys = jax.random.split(key, 3)

        (_, policy_log), policy_grads = eqx.filter_value_and_grad(self.policy_loss, has_aux=True)(
            agent_state.policy.module,
            critic=agent_state.critic.module,
            entropy_coef=agent_state.entropy_coef.module,
            batch=batch,
            key=keys[0],
        )
        (_, critic_log), critic_grads = eqx.filter_value_and_grad(self.critic_loss, has_aux=True)(
            agent_state.critic.module,
            target_critic=agent_state.target_critic.module,
            policy=agent_state.policy.module,
            entropy_coef=agent_state.entropy_coef.module,
            batch=batch,
            key=keys[1],
        )

        agent_state.policy = agent_state.policy.apply_updates(policy_grads)
        agent_state.critic = agent_state.critic.apply_updates(critic_grads)

        log = {}
        log.update(policy_log)
        log.update(critic_log)

        if self.auto_entropy:
            (_, entropy_coef_log), entropy_coef_grads = eqx.filter_value_and_grad(
                self.entropy_coef_loss, has_aux=True
            )(agent_state.entropy_coef.module, agent_state.policy.module, batch, keys[2])
            log.update(entropy_coef_log)

            agent_state.entropy_coef = agent_state.entropy_coef.apply_updates(entropy_coef_grads)

        agent_state.target_critic.module = self.soft_update(
            agent_state.critic.module, agent_state.target_critic.module, self.target_critic_lr
        )

        return agent_state, log
