import math

import equinox as eqx
import equinox.nn as nn
import jax
import jax.numpy as jnp

from rlab_jax.distributions import (multivariate_normal_diag,
                                    tanh_multivariate_normal_diag,
                                    ImageMultivariateNormalDiag)
from rlab_jax.modules.networks.module import Module
from rlab_jax.modules.networks.utils import multi_vmap
from rlab_jax.modules.networks.initialization import init_linear_weight, init_linear_bias

from equinox.nn import Conv2d, ConvTranspose2d


class CnnEncoder(Module):
    _MODULE_TYPE = "CnnEncoder"
    _layers: list
    cnn: eqx.Module

    def __init__(
        self,
        n_inputs,
        key: jax.random.PRNGKey,
        activation=None,
        hidden_activation=jax.nn.relu,
        norm=False,
        init_weight_type=None,
        init_bias_type=None,
        cnn_channels=[32, 64, 128, 256],
        cnn_kernels=[4, 4, 4, 4],
        cnn_stride=2,
        **ignored
    ):
        super().__init__()

        object.__setattr__(self, "subclasses", None)

        if isinstance(hidden_activation, str):
            hidden_activation = eval(hidden_activation)

        if isinstance(activation, str):
            activation = eval(activation)

        self._layers = []
        keys = jax.random.split(key, len(cnn_kernels) + 1)
        in_channel = n_inputs[0]
        for i, (kernel, channel) in enumerate(zip(cnn_kernels, cnn_channels)):
            self._layers.append(Conv2d(in_channels=in_channel, out_channels=channel, kernel_size=kernel, stride=cnn_stride, key=keys[i], use_bias=not norm == "LayerNorm")
            )

            if i < len(cnn_kernels) - 1:

                # InstanceNorm
                if norm == "LayerNorm":
                    self._layers.append(nn.Lambda(
                        jax.vmap(jax.vmap(nn.LayerNorm(shape=channel), in_axes=1, out_axes=1), in_axes=1, out_axes=1)))

                self._layers.append(nn.Lambda(hidden_activation))
            in_channel = channel

        if activation is not None:
            self._layers.append(nn.Lambda(activation))

        self.cnn = nn.Sequential(self._layers)

        if init_weight_type is not None:
            self.cnn = init_linear_weight(self.cnn, init_weight_type, key)
        if init_bias_type is not None:
            self.cnn = init_linear_bias(self.cnn, init_bias_type, key)

    def __call__(self, x):
        batch_shape = x.shape[:-3] + (-1,)
        if len(x.shape) == 4:
            x = jax.vmap(self.cnn)(x)
        elif len(x.shape) == 5:
            x = jax.vmap(jax.vmap(self.cnn))(x)
        else:
            raise ValueError
        return x.reshape(batch_shape)


class CnnDecoder(Module):
    _MODULE_TYPE = "CnnDecoder"
    _layers: list
    cnn: eqx.Module
    cnn_input: eqx.Module

    def __init__(
        self,
        n_inputs,
        key: jax.random.PRNGKey,
        activation=jax.nn.sigmoid,
        hidden_activation=jax.nn.relu,
        norm=False,
        init_weight_type=None,
        init_bias_type=None,
        cnn_channels=[128, 64, 32, 3],
        cnn_kernels=[5, 5, 6, 6],
        cnn_stride=2,
        n_cnn_inputs=1024,
        **ignored
    ):
        super().__init__()

        object.__setattr__(self, "subclasses", None)

        if isinstance(hidden_activation, str):
            hidden_activation = eval(hidden_activation)

        if isinstance(activation, str):
            activation = eval(activation)

        key, subkey = jax.random.split(key)
        self.cnn_input = nn.Linear(in_features=n_inputs, out_features=n_cnn_inputs, key=key)

        self._layers = []
        keys = jax.random.split(subkey, len(cnn_kernels) + 1)
        in_channel = n_cnn_inputs
        for i, (kernel, channel) in enumerate(zip(cnn_kernels, cnn_channels)):
            self._layers.append(ConvTranspose2d(in_channels=in_channel, out_channels=channel, kernel_size=kernel, stride=cnn_stride, key=keys[i], use_bias=not norm == "LayerNorm")
            )

            if i < len(cnn_kernels) - 1:

                # InstanceNorm
                if norm == "LayerNorm":
                    self._layers.append(nn.Lambda(
                        jax.vmap(jax.vmap(nn.LayerNorm(shape=channel), in_axes=1, out_axes=1), in_axes=1, out_axes=1)))

                self._layers.append(nn.Lambda(hidden_activation))
            in_channel = channel

        if activation is not None:
            self._layers.append(nn.Lambda(activation))

        self.cnn = nn.Sequential(self._layers)

        if init_weight_type is not None:
            self.cnn = init_linear_weight(self.cnn, init_weight_type, key)
        if init_bias_type is not None:
            self.cnn = init_linear_bias(self.cnn, init_bias_type, key)

    def __call__(self, x):

        y = multi_vmap(fn=self.cnn_input, shape=x.shape, x=x)
        y = jnp.expand_dims(jnp.expand_dims(y, -1), -1)

        if len(x.shape) == 3:
            mean = jax.vmap(jax.vmap(self.cnn))(y)
        else:
            mean = jax.vmap(self.cnn)(y)
        return ImageMultivariateNormalDiag(loc=mean, scale=jnp.ones_like(mean))

