import jax
import equinox as eqx


def init_fn(weight: jax.Array, key: jax.random.PRNGKey, init_type: str) -> jax.Array:
    if init_type == "trunc_normal":
        out_, in_ = weight.shape
        stddev = 1 / (2 * in_ ** 0.5)
        return stddev * jax.random.truncated_normal(key, shape=(out_, in_), lower=-2, upper=2)
    else:
        in_axis = 1
        out_axis = 0

        if init_type == "xavier_normal":
            initializer = jax.nn.initializers.xavier_normal(in_axis=in_axis, out_axis=out_axis)
        elif init_type == "selu_normal":
            initializer = jax._src.nn.initializers.variance_scaling(1.0, "fan_in", "normal", in_axis=in_axis, out_axis=out_axis)
        else:
            raise ValueError

        return initializer(key, weight.shape, jax.numpy.float32)


def zero_init(weight: jax.Array, key: jax.random.PRNGKey) -> jax.Array:
    return jax.numpy.zeros_like(weight)


def init_linear_weight(model, init_type, key, scaling=1.0):
    is_weight = lambda x: isinstance(x, eqx.nn.Linear) or isinstance(x, eqx.nn.Conv) or isinstance(x, eqx.nn.ConvTranspose)
    get_weights = lambda m: [x.weight
                           for x in jax.tree_util.tree_leaves(m, is_leaf=is_weight)
                           if is_weight(x)]
    weights = get_weights(model)
    new_weights = [init_fn(weight, subkey, init_type) * scaling
                 for weight, subkey in zip(weights, jax.random.split(key, len(weights)))]
    new_model = eqx.tree_at(get_weights, model, new_weights)
    return new_model


def init_linear_bias(model, init_type, key):
    if init_type == "zero":
        init_fn = zero_init
    else:
        raise NotImplementedError

    is_weight = lambda x: isinstance(x, eqx.nn.Linear) or isinstance(x, eqx.nn.Conv) or isinstance(x, eqx.nn.ConvTranspose)
    get_weights = lambda m: [x.bias
                           for x in jax.tree_util.tree_leaves(m, is_leaf=is_weight)
                           if is_weight(x) and x.bias is not None]
    weights = get_weights(model)
    new_weights = [init_fn(weight, subkey)
                 for weight, subkey in zip(weights, jax.random.split(key, len(weights)))]
    new_model = eqx.tree_at(get_weights, model, new_weights)
    return new_model
