import math

import equinox as eqx
import equinox.nn as nn
import jax
import jax.numpy as jnp

from rlab_jax.distributions import (multivariate_normal_diag,
                                    tanh_multivariate_normal_diag,
                                    beta_multivariate_normal_diag,
                                    bernoulli_diag)
from rlab_jax.modules.networks.module import Module
from rlab_jax.modules.networks.utils import multi_vmap
from rlab_jax.modules.networks.initialization import init_linear_weight, init_linear_bias


class Mlp(Module):
    _MODULE_TYPE = "Mlp"
    _layers: list
    mlp: eqx.Module

    def __init__(
        self,
        n_inputs,
        n_outputs,
        key: jax.random.PRNGKey,
        activation=None,
        layers=1,
        units=128,
        hidden_activation=jax.nn.relu,
        norm=False,
        init_weight_type=None,
        init_bias_type=None,
        **ignored
    ):
        super().__init__()

        object.__setattr__(self, "subclasses", None)

        if isinstance(hidden_activation, str):
            hidden_activation = eval(hidden_activation)

        if isinstance(activation, str):
            activation = eval(activation)

        # hidden layers
        n_hidden_inputs = n_inputs
        self._layers = []
        keys = jax.random.split(key, layers + 1)
        for i in range(layers):
            self._layers.append(
                nn.Linear(in_features=n_hidden_inputs, out_features=units, key=keys[i], use_bias=not norm == "LayerNorm")
            )

            if norm == "LayerNorm":
                self._layers.append(nn.LayerNorm(shape=units))

            self._layers.append(nn.Lambda(hidden_activation))
            n_hidden_inputs = units

        # output layer
        self._layers.append(nn.Linear(n_hidden_inputs, n_outputs, key=keys[-1]))

        if activation is not None:
            self._layers.append(nn.Lambda(activation))
        if norm == "PenultimateLayerNorm":
            self._layers.append(nn.LayerNorm(shape=n_outputs))
        elif isinstance(norm, str) and "PenultimateSimNorm" in norm:

            def sim_norm(norm):
                import re
                v = int(re.split('(\d+)', norm)[1])

                def sn(x):
                    shape = x.shape
                    z = x.reshape(*shape[:-1], -1, v)
                    z = jax.nn.softmax(z, axis=-1)
                    return z.reshape(shape)

                return sn

            self._layers.append(nn.Lambda(sim_norm(norm)))

        self.mlp = nn.Sequential(self._layers)

        if init_weight_type is not None:
            self.mlp = init_linear_weight(self.mlp, init_weight_type, key)
        if init_bias_type is not None:
            self.mlp = init_linear_bias(self.mlp, init_bias_type, key)

    def call(self, x):
        return self.mlp(x)

    def __call__(self, x):
        return multi_vmap(fn=self.call, shape=x.shape, x=x)


class BernoulliMlp(Mlp):
    _MODULE_TYPE = "BernoulliMlp"
    _dist_type = "bernoulli"

    def call(self, x):
        return bernoulli_diag(self.mlp(x))


class GaussMlp(Mlp):
    _MODULE_TYPE = "GaussMlp"
    _dist_type = "normal"
    n_inputs: int
    n_outputs: int
    mean_layer: eqx.Module
    std_layer: eqx.Module
    min_std: jnp.array = eqx.field(static=True)
    init_std: jnp.array = eqx.field(static=True)

    def __init__(
        self,
        n_inputs,
        n_outputs,
        key,
        units=128,
        layers=1,
        hidden_activation=jax.nn.relu,
        norm=False,
        init_std=None,
        min_std=math.sqrt(1e-5),
        init_weight_type=None,
        init_bias_type=None,
        last_layer_scaling=1.0,
        **ignored
    ):
        keys = jax.random.split(key, 7)
        super().__init__(
            n_inputs=n_inputs,
            n_outputs=units,
            key=keys[0],
            activation=hidden_activation,
            hidden_activation=hidden_activation,
            norm=norm,
            layers=layers - 1,
            units=units,
            init_weight_type=init_weight_type,
            init_bias_type=init_bias_type,
        )

        self.n_inputs = n_inputs
        self.n_outputs = n_outputs

        self.mean_layer = nn.Linear(units, n_outputs, key=keys[1])
        self.std_layer = nn.Linear(units, n_outputs, key=keys[2])

        self.min_std = min_std

        if init_std is None:
            self.init_std = jnp.zeros(1)
        else:
            self.init_std = jnp.log(jnp.exp(jnp.array(init_std)) - 1)

        if init_weight_type is not None:
            self.mean_layer = init_linear_weight(self.mean_layer, init_weight_type, keys[3], scaling=last_layer_scaling)
            self.std_layer = init_linear_weight(self.std_layer, init_weight_type, keys[4], scaling=last_layer_scaling)

        if init_bias_type is not None:
            self.mean_layer = init_linear_bias(self.mean_layer, init_bias_type, keys[5])
            self.std_layer = init_linear_bias(self.std_layer, init_bias_type, keys[6])

    def call(self, x):
        y = self.mlp(x)
        mean = self.mean_layer(y)
        std = jax.nn.softplus(self.std_layer(y) + self.init_std)
        std = jax.lax.clamp(min=self.min_std, x=std, max=jnp.inf)

        return multivariate_normal_diag(mean=mean, std=std)


class Sigmoid2GaussMlp(GaussMlp):
    _MODULE_TYPE = "Sigmoid2GaussMlp"

    def call(self, x):
        y = self.mlp(x)
        mean = self.mean_layer(y)
        std = 2 * jax.nn.sigmoid(self.std_layer(y) / 2) + self.min_std

        return multivariate_normal_diag(mean=mean, std=std)


class PETSGaussMlp(GaussMlp):
    _MODULE_TYPE = "PETSGaussMlp"
    min_logvar: jnp.array
    max_logvar: jnp.array

    def __init__(self, n_outputs, *args, **kwargs):
        super().__init__(n_outputs=n_outputs, *args, **kwargs)
        self.min_logvar = jnp.ones((n_outputs)) * -10
        self.max_logvar = jnp.ones((n_outputs)) * 0.5

    def call(self, x):
        y = self.mlp(x)
        mean = self.mean_layer(y)
        logvar = self.std_layer(y)
        logvar = self.max_logvar - jax.nn.softplus(self.max_logvar - logvar)
        logvar = self.min_logvar + jax.nn.softplus(logvar - self.min_logvar)
        std = jnp.exp(0.5 * logvar)
        return multivariate_normal_diag(mean=mean, std=std)


class TanhGaussMlp(GaussMlp):
    _MODULE_TYPE = "TanhGaussMlp"
    _dist_type = "tanh_normal"
    mean_scale: int
    max_std: int

    def __init__(self, mean_scale=5, max_std=jnp.inf, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mean_scale = mean_scale
        self.max_std = max_std

    def call(self, x):
        y = self.mlp(x)
        mean = jnp.tanh(self.mean_layer(y) / self.mean_scale) * self.mean_scale
        std = jax.nn.softplus(self.std_layer(y) + self.init_std)
        std = jax.lax.clamp(min=self.min_std, x=std, max=self.max_std)

        return tanh_multivariate_normal_diag(mean, std)


class MobileTanhGaussMlp(GaussMlp):
    _MODULE_TYPE = "MobileTanhGaussMlp"
    _dist_type = "tanh_normal"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def call(self, x):
        y = self.mlp(x)
        mean = self.mean_layer(y)
        std = self.std_layer(y)
        std = jnp.exp(jax.lax.clamp(min=-5.0, x=std, max=2.0))

        return tanh_multivariate_normal_diag(mean, std)


class BetaMlp(GaussMlp):
    _MODULE_TYPE = "BetaMlp"
    _dist_type = "beta"

    def __init__(self, mean_scale=5, max_std=jnp.inf, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def call(self, x):
        y = self.mlp(x)
        # 2 as bias for initialization
        alpha = jax.nn.softplus(self.mean_layer(y) + 2) + 1 + 1e-5
        beta = jax.nn.softplus(self.std_layer(y) + 2) + 1 + 1e-5
        # Hack to make beta distribution compatible with mean and std representation
        return beta_multivariate_normal_diag(mean=alpha, std=beta)
