from rlab_jax.modules import Module
from rlab_jax.rlab_dataclasses import RandomVariable
from rlab_jax.modules.networks import BernoulliMlp
import jax.numpy as jnp
import jax
from rlab_jax import keygen


def get_done_fn(env_name):
    if 'hopper' in env_name.lower():
        return hopper_is_done
    elif 'walker' in env_name.lower():
        return walker2d_is_done
    elif 'halfcheetah' in env_name.lower():
        return halfcheetah_is_done
    else:
        raise NotImplementedError


def hopper_is_done(next_observation):
    height = next_observation[..., 0]
    angle = next_observation[..., 1]
    not_done = jnp.isfinite(next_observation).all(axis=-1) \
               * (jnp.abs(next_observation[..., 1:]) < 100).all(axis=-1) \
               * (height > .7) \
               * (jnp.abs(angle) < .2)
    done = ~not_done
    return jnp.expand_dims(done.astype(jnp.float32), -1)


def walker2d_is_done(next_observation):
    height = next_observation[..., 0]
    ang = next_observation[..., 1]
    done = ~((height > 0.8) & (height < 2.0) &
             (ang > -1.0) & (ang < 1.0))
    return jnp.expand_dims(done.astype(jnp.float32), -1)


def halfcheetah_is_done(next_observation):
    return jnp.expand_dims(jnp.zeros_like(next_observation[..., 0]), -1)


class RSSMOracleDone(Module):
    _MODULE_TYPE = "RSSMOracleDone"
    done_fn: callable

    def __init__(self, env_name, *args, **kwargs):
        self.done_fn = get_done_fn(env_name)

    def __call__(
        self,
        model_state,
        action: jnp.array,
        next_state: RandomVariable,
        key: jax.random.PRNGKey
    ):
        next_observation_dist = model_state.module.sequence_model.decoder(next_state.state.sample_deterministic_stochastic)
        next_observation = next_observation_dist.mean()
        next_observation = model_state.scaler.inverse_transform(x=next_observation, scaler=model_state.scaler)

        return self.done_fn(next_observation)

    def loss(self, *args, **kwargs):
        return 0, {}


class RSSMLearnedDone(Module):
    _MODULE_TYPE = "RSSMLearnedDone"
    done_network: Module

    def __init__(self, n_inputs, key, **cfg):

        self.done_network = BernoulliMlp(**cfg, n_inputs=n_inputs,  n_outputs=1, key=key)

    def __call__(
        self,
        model_state,
        action: jnp.array,
        next_state: RandomVariable,
        key: jax.random.PRNGKey
    ):
        if hasattr(next_state, "state"):
            next_state = next_state.state

        inputs = jnp.concatenate((next_state.sample_deterministic_stochastic, action), -1)

        return self.done_network(inputs).mode()

    def loss(
        self,
        next_state: RandomVariable,
        action: jnp.array,
        target: jnp.array,
    ):
        if hasattr(next_state, "state"):
            next_state = next_state.state

        inputs = jnp.concatenate((next_state.sample_deterministic_stochastic, action), -1)
        dist = self.done_network(inputs)

        nll = -dist.log_prob(target).mean()

        return nll, {"discount_function/loss": nll}
