import jax
import jax.numpy as jnp

from rlab_jax.modules.networks import GaussMlp
from rlab_jax.rlab_dataclasses import RandomVariable
from rlab_jax.utils.keygen import keygen

from rlab_jax.modules import Module


class StochasticLearnedReward(Module):
    _MODULE_TYPE = "StochasticLearnedReward"
    reward_network: Module

    def __init__(self, n_inputs, key, **cfg):

        self.reward_network = GaussMlp(**cfg, n_inputs=n_inputs, n_outputs=1, key=key)

    def __call__(
        self,
        action: jnp.array,
        next_state: RandomVariable,
        key: jax.random.PRNGKey
    ):
        inputs = jnp.concatenate((next_state.sample, action), -1)

        return self.reward_network(inputs).mean()

    def loss(
        self,
        next_state: RandomVariable,
        action: jnp.array,
        target: jnp.array,
    ):
        inputs = jnp.concatenate((next_state.sample, action), -1)
        reward_dist = self.reward_network(inputs)

        nll = -reward_dist.log_prob(target).mean()

        return nll, {"reward_function/loss": nll}


class RSSMStochasticLearnedReward(StochasticLearnedReward):
    _MODULE_TYPE = "RSSMStochasticLearnedReward"

    def __call__(
        self,
        action: jnp.array,
        next_state: RandomVariable,
        key: jax.random.PRNGKey
    ):
        if hasattr(next_state, "state"):
            next_state = next_state.state

        inputs = jnp.concatenate((next_state.sample_deterministic_stochastic, action), -1)

        return self.reward_network(inputs).mean()

    def loss(
        self,
        next_state: RandomVariable,
        action: jnp.array,
        target: jnp.array,
    ):
        if hasattr(next_state, "state"):
            next_state = next_state.state

        inputs = jnp.concatenate((next_state.sample_deterministic_stochastic, action), -1)
        reward_dist = self.reward_network(inputs)

        nll = -reward_dist.log_prob(target).mean()

        return nll, {"reward_function/loss": nll}
