import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf
import optax

from rlab_jax.modules.networks import GaussMlp, Mlp, Module, multi_vmap
from rlab_jax.modules.sequence_models import SequenceModel
from rlab_jax.rlab_dataclasses import (
    ScalerAnnealingTrainState, DeterministicStochasticRandomVariable,
    DeterministicStochasticRandomVariableList, Prediction, random_variable,
    random_variable_list, Transition)
from rlab_jax.utils.keygen import keygen

class RSSMModel(eqx.Module):
    encoder: Module
    decoder: Module
    transition: Module
    stochastic_state_inference: Module
    reward: Module
    done: Module


class RSSMTransition(Module):
    _MODULE_TYPE = "RSSMTransition"
    rnn: eqx.Module
    rnn_input: eqx.Module
    deterministic_to_stochastic: eqx.Module

    def __init__(self, rnn: eqx.nn.GRUCell, rnn_input: Mlp, deterministic_to_stochastic: GaussMlp):
        self.rnn = rnn
        self.rnn_input = rnn_input
        self.deterministic_to_stochastic = deterministic_to_stochastic

    @classmethod
    def from_cfg(
        cls,
        cfg: omegaconf.DictConfig,
        n_stochastic_states: int,
        n_deterministic_states: int,
        n_actions: int,
        **kwargs,
    ):

        rnn = eqx.nn.GRUCell(input_size=cfg.n_rnn_inputs, hidden_size=n_deterministic_states, key=keygen())

        rnn_input = Module.create(
            **cfg.rnn_input_model,
            n_inputs=n_stochastic_states + n_actions,
            n_outputs=cfg.n_rnn_inputs,
            key=keygen(),
        )

        deterministic_to_stochastic = Module.create(
            **cfg.deterministic_to_stochastic,
            n_inputs=n_deterministic_states,
            n_outputs=n_stochastic_states,
            key=keygen(),
        )

        return cls(
            rnn=rnn, rnn_input=rnn_input, deterministic_to_stochastic=deterministic_to_stochastic
        )

    def __call__(
        self, deterministic_state: jnp.array, stochastic_state: jnp.array, action: jnp.array, key: jax.random.PRNGKey
    ):
        rnn_input = self.rnn_input(jnp.concatenate((stochastic_state, action), -1))

        # deterministic transition
        next_deterministic_state = multi_vmap(fn=self.rnn, shape=rnn_input.shape, input=rnn_input, hidden=deterministic_state)

        # stochastic state
        next_stochastic_dist = self.deterministic_to_stochastic(next_deterministic_state)

        return next_deterministic_state, next_stochastic_dist


class RSSMTransitionEnsemble(Module):
    _MODULE_TYPE = "RSSMTransitionEnsemble"
    rnn: eqx.Module
    rnn_input: eqx.Module
    deterministic_to_stochastic: eqx.Module
    n_networks: int

    def __init__(self, rnn: eqx.nn.GRUCell, rnn_input: Mlp, deterministic_to_stochastic: GaussMlp, n_networks: int):
        self.rnn = rnn
        self.rnn_input = rnn_input
        self.deterministic_to_stochastic = deterministic_to_stochastic
        self.n_networks = n_networks

    @classmethod
    def from_cfg(
        cls,
        cfg: omegaconf.DictConfig,
        n_stochastic_states: int,
        n_deterministic_states: int,
        n_actions: int,
        **kwargs,
    ):

        rnn = eqx.nn.GRUCell(input_size=cfg.n_rnn_inputs, hidden_size=n_deterministic_states, key=keygen())

        rnn_input = Module.create(
            **cfg.rnn_input_model,
            n_inputs=n_stochastic_states + n_actions,
            n_outputs=cfg.n_rnn_inputs,
            key=keygen(),
        )

        @eqx.filter_vmap
        def ensemble(key):
            return Module.create(
            **cfg.deterministic_to_stochastic,
            n_inputs=n_deterministic_states,
            n_outputs=n_stochastic_states,
            key=key,
        )

        deterministic_to_stochastic = ensemble(jax.random.split(keygen(), cfg.n_networks))

        return cls(
            rnn=rnn, rnn_input=rnn_input, deterministic_to_stochastic=deterministic_to_stochastic, n_networks=cfg.n_networks
        )

    def __call__(
        self, deterministic_state: jnp.array, stochastic_state: jnp.array, action: jnp.array, key: jax.random.PRNGKey
    ):
        rnn_input = self.rnn_input(jnp.concatenate((stochastic_state, action), -1))

        # deterministic transition
        next_deterministic_state = multi_vmap(fn=self.rnn, shape=rnn_input.shape, input=rnn_input, hidden=deterministic_state)

        # select network
        k = jax.random.randint(key, minval=0, maxval=self.n_networks, shape=(1,))

        def fn(tree):
            if eqx.is_array(tree):
                return tree[k].squeeze(0)
            else:
                return tree

        deterministic_to_stochastic = jax.tree_util.tree_map(fn, self.deterministic_to_stochastic)

        # stochastic state
        next_stochastic_dist = deterministic_to_stochastic(next_deterministic_state)

        return next_deterministic_state, next_stochastic_dist


class RSSM(SequenceModel):
    _SEQUENCE_MODEL_TYPE = "RSSM"

    def __init__(
        self,
        n_actions: int,
        n_observations: int,
        n_deterministic_states: int,
        n_stochastic_states: int,
        encoder: Module,
        decoder: Module,
        transition: Module,
        stochastic_state_inference: Module,
        reward: Module,
        done: Module,
        optim_type: str = "optax.adam",
        optim_settings: dict = {"lr": 3e-4},
        beta: float = 1,
        temperature: float = 1,
        max_gradient_norm: float = 100,
        free_nats: float = 1.0,
        kl_balance: float = 0.8,
        n_initial_filtered_steps: int = 1,
    ):
        """Instantiate a RecurrentStateSpaceModel."""
        super().__init__()

        self.n_actions = n_actions
        self.n_observations = n_observations
        self.n_deterministic_states = n_deterministic_states
        self.n_stochastic_states = n_stochastic_states
        self.n_latent = n_stochastic_states + n_deterministic_states

        self.beta = beta
        self.temperature = jnp.array(temperature)
        self.max_gradient_norm = max_gradient_norm
        self.free_nats = free_nats
        self.kl_balance = kl_balance

        self.n_initial_filtered_steps = n_initial_filtered_steps

        optim_settings = dict(optim_settings)
        if "lr" in optim_settings.keys():
            optim_settings["learning_rate"] = optim_settings["lr"]
            del optim_settings["lr"]

        if optim_type in ["sam_adam", "sam_sgd"]:
            optim_type = "optax.adam"
        optim = optax.chain(
            optax.clip_by_global_norm(self.max_gradient_norm), eval(optim_type)(**optim_settings)
        )

        model_state = RSSMModel(
            encoder=encoder,
            decoder=decoder,
            transition=transition,
            stochastic_state_inference=stochastic_state_inference,
            reward=reward,
            done=done
        )

        self.train_state = ScalerAnnealingTrainState(
            module=model_state,
            optim=optim,
            optim_state=optim.init(eqx.filter(model_state, eqx.is_array)),
            annealing=jnp.array(1 / temperature),
            scaler=None
        )

    def initial_state(self):
        return self.train_state

    @classmethod
    def from_cfg(cls, cfg, n_observations, n_actions, env_name=None, n_transition_actions=None, **_ignored):
        encoder = Module.create(
            **cfg.encoder,
            n_inputs=n_observations,
            n_outputs=cfg.n_observation_embedding,
            key=keygen(),
        )

        stochastic_state_inference = Module.create(
            **cfg.stochastic_state_inference,
            n_inputs=cfg.n_deterministic_states + cfg.n_observation_embedding,
            n_outputs=cfg.n_stochastic_states,
            key=keygen(),
        )

        decoder = Module.create(
            **cfg.decoder,
            n_inputs=cfg.n_deterministic_states + cfg.n_stochastic_states,
            n_outputs=n_observations,
            key=keygen(),
        )
        if n_transition_actions is None:
            n_transition_actions = n_actions

        transition = Module.create(
            type=cfg.transition.type,
            cfg=cfg.transition,
            n_actions=n_transition_actions,
            n_deterministic_states=cfg.n_deterministic_states,
            n_stochastic_states=cfg.n_stochastic_states,
            from_cfg=True,
        )

        n_inputs = cfg.n_stochastic_states + n_actions
        if cfg.reward.type == "RSSMStochasticLearnedReward":
            n_inputs += cfg.n_deterministic_states

        reward = Module.create(**cfg.reward, n_inputs=n_inputs, key=keygen())

        done = Module.create(**cfg.done, n_inputs=cfg.n_deterministic_states + cfg.n_stochastic_states + n_actions, key=keygen(), env_name=env_name)

        return cls(
            n_actions=n_actions,
            n_observations=n_observations,
            n_deterministic_states=cfg.n_deterministic_states,
            n_stochastic_states=cfg.n_stochastic_states,
            stochastic_state_inference=stochastic_state_inference,
            encoder=encoder,
            decoder=decoder,
            transition=transition,
            reward=reward,
            done=done,
            beta=cfg.beta,
            temperature=cfg.temperature,
            optim_type=cfg.optim.type,
            optim_settings=cfg.optim.settings,
            max_gradient_norm=cfg.max_gradient_norm,
            free_nats=cfg.free_nats,
            kl_balance=cfg.kl_balance,
            n_initial_filtered_steps=cfg.n_initial_filtered_steps,
        )

    def __call__(
        self, model: RSSMModel, observation_trajectory: jnp.array, action_trajectory: jnp.array, key: jax.random.PRNGKey, scaler
    ) -> Prediction:

        keys = jax.random.split(key, 2)

        prior, posterior = self._encode(
            model=model, observation_trajectory=observation_trajectory, action_trajectory=action_trajectory, key=keys[0]
        )

        reconstruction = self._decode(model=model, latent_states=posterior, key=keys[1])

        return Prediction(posterior=posterior, prior=prior, reconstruction=reconstruction)

    def _encode(
        self, model: RSSMModel, observation_trajectory: jnp.array, action_trajectory: jnp.array, key: jax.random.PRNGKey
    ):
        keys = jax.random.split(key, 2)

        observation_embedding = model.encoder(observation_trajectory)

        prior_states = DeterministicStochasticRandomVariableList(
            self.initial_latent_state(model=model, batch_size=observation_trajectory.shape[1])
        )
        posterior_states = DeterministicStochasticRandomVariableList(
            self.initial_latent_state(model=model, batch_size=observation_trajectory.shape[1])
        )

        for i in range(action_trajectory.shape[0] - 1):
            keys = jax.random.split(keys[0], 3)
            prior_state = self.one_step(
                model=model, latent_state=posterior_states[i], action=action_trajectory[i + 1], key=keys[1]
            )

            posterior_state = self.filter_after_one_step(
                model=model,
                latent_state=prior_state,
                observation_embedding=observation_embedding[i + 1],
                key=keys[2],
            )

            prior_states.append(prior_state)
            posterior_states.append(posterior_state)

        return prior_states.stack(), posterior_states.stack()

    def _decode(
        self,
        model: RSSMModel,
        latent_states: DeterministicStochasticRandomVariable,
        key: jax.random.PRNGKey,
    ):
        reconstructions_dist = model.decoder(latent_states.sample_deterministic_stochastic)
        reconstructions = reconstructions_dist.sample(key)

        return random_variable(sample=reconstructions, **reconstructions_dist.params)

    def one_step(
        self,
        model: RSSMModel,
        latent_state: DeterministicStochasticRandomVariable,
        action: jnp.array,
        key: jax.random.PRNGKey,
        **ignored,
    ):
        if isinstance(action, jax.Array):
            action = action
        else:
            action = action.sample

        key, subkey = jax.random.split(key)

        next_deterministic_state, next_stochastic_dist = model.transition(
            deterministic_state=latent_state.deterministic,
            stochastic_state=latent_state.stochastic.sample,
            action=action,
            key=key,
        )

        return DeterministicStochasticRandomVariable(
            deterministic=next_deterministic_state,
            stochastic=random_variable(
                sample=next_stochastic_dist.sample(subkey), **next_stochastic_dist.params
            ),
        )

    def filter_after_one_step(
        self,
        model: RSSMModel,
        latent_state: DeterministicStochasticRandomVariable,
        observation_embedding: jnp.array,
        key: jax.random.PRNGKey,
    ):
        inputs = jnp.concatenate([latent_state.deterministic, observation_embedding], -1)
        stochastic_state_dist = model.stochastic_state_inference(inputs)

        return DeterministicStochasticRandomVariable(
            deterministic=latent_state.deterministic,
            stochastic=random_variable(
                sample=stochastic_state_dist.sample(key), **stochastic_state_dist.params
            ),
        )

    def initial_latent_state(self, model: RSSMModel, batch_size: int):
        return DeterministicStochasticRandomVariable(
            deterministic=jnp.zeros((batch_size, self.n_deterministic_states)),
            stochastic=random_variable(
                sample=jnp.zeros((batch_size, self.n_stochastic_states)),
                mean=jnp.zeros((batch_size, self.n_stochastic_states)),
                std=jnp.ones((batch_size, self.n_stochastic_states)),
                dist_type=model.stochastic_state_inference._dist_type,
            ),
        )

    def loss(
        self,
        model: RSSMModel,
        batch: Transition,
        key: jax.random.PRNGKey,
        annealing: jnp.array,
        scaler
    ):
        predictions = self.__call__(model, action_trajectory=batch.action_trajectory, observation_trajectory=batch.observation_trajectory, key=key, scaler=scaler)

        # ignore s_{0} and z_{0}
        nll_sample_wise_state = -predictions.reconstruction.dist.log_prob(batch.observation_trajectory)
        nll_sample_wise_state = nll_sample_wise_state.at[0].set(0)
        nll = jnp.mean(nll_sample_wise_state)

        if self.kl_balance is not None:
            kl_prior = jax.lax.stop_gradient(predictions.posterior.stochastic.dist).kl_divergence(predictions.prior.stochastic.dist)
            kl_posterior = predictions.posterior.stochastic.dist.kl_divergence(jax.lax.stop_gradient(predictions.prior.stochastic.dist))
            kl_prior = kl_prior.at[0].set(0)
            kl_posterior = kl_posterior.at[0].set(0)

            if self.free_nats is not None:
                kl_prior = jnp.maximum(kl_prior.mean(), self.free_nats)
                kl_posterior = jnp.maximum(kl_posterior.mean(), self.free_nats)
            else:
                kl_prior = kl_prior.mean()
                kl_posterior = kl_posterior.mean()

            kl = self.kl_balance * kl_prior + (1 - self.kl_balance) * kl_posterior

        else:
            kl_sample_wise_state = predictions.posterior.stochastic.dist.kl_divergence(predictions.prior.stochastic.dist)
            kl_sample_wise_state = kl_sample_wise_state.at[0].set(0)

            if self.free_nats is not None:
                kl_sample_wise_state = jnp.maximum(kl_sample_wise_state, self.free_nats)

            kl = jnp.mean(kl_sample_wise_state)

        kl = kl * self.beta * annealing

        loss = nll + kl

        reward_loss, reward_log = model.reward.loss(next_state=predictions.posterior.item(jnp.array(range(1, predictions.posterior.sample.shape[0]))),
                                                    action=batch.action,
                                                    target=batch.reward)

        done_loss, done_log = model.done.loss(next_state=predictions.posterior.item(jnp.array(range(1, predictions.posterior.sample.shape[0]))),
                                              action=batch.action,
                                              target=batch.done)

        loss += reward_loss
        loss += done_loss

        annealing = jnp.minimum(annealing + (1.0 / self.temperature), 1.0)

        log = {
            "sequence_model/loss": loss,
            "sequence_model/nll": nll,
            "sequence_model/kl/z": kl,
            "sequence_model/warmup_kl/z": annealing,
            **reward_log,
            **done_log
        }

        return loss, (log, predictions, annealing)

    def training_step(
        self, model_state: ScalerAnnealingTrainState, batch: Transition, key: jax.random.PRNGKey
    ):
        (_, (log, predictions, annealing)), grads = eqx.filter_value_and_grad(
            self.loss, has_aux=True
        )(
            model_state.module,
            batch=batch,
            key=key,
            annealing=model_state.annealing,
            scaler=model_state.scaler,
        )

        model_state = model_state.apply_updates(grads)
        model_state.annealing = annealing

        return model_state, predictions, log

    @eqx.filter_jit
    def filter(
        self,
        model_state: ScalerAnnealingTrainState,
        next_observation: jnp.array,
        latent_state: DeterministicStochasticRandomVariable,
        action: jnp.array,
        key: jax.random.PRNGKey,
        reset=False,
        **ignored,
    ):
        if reset:
            return self.initial_latent_state(
                model=model_state.module, batch_size=next_observation.shape[0]
            )

        keys = jax.random.split(key, 3)

        observation_embedding = model_state.module.encoder(next_observation)

        prior_state = self.one_step(
            model=model_state.module, latent_state=latent_state, action=action.sample, key=keys[1]
        )

        posterior_state = self.filter_after_one_step(
            model=model_state.module,
            latent_state=prior_state,
            observation_embedding=observation_embedding,
            key=keys[2],
        )

        return posterior_state

    def generate(
        self,
        model_state: ScalerAnnealingTrainState,
        next_observations: jnp.array,
        actions: jnp.array,
        key: jax.random.PRNGKey,
    ) -> Prediction:
        latent_state = self.initial_latent_state(model=model_state.module, batch_size=next_observations.shape[1])

        observation_embedding = model_state.module.encoder(next_observations)

        for i in range(actions.shape[0]):
            key, subkey = jax.random.split(key)
            latent_state = self.one_step(
                model=model_state.module,
                latent_state=latent_state,
                action=actions[i],
                key=subkey,
            )

            if i < self.n_initial_filtered_steps:
                key, subkey = jax.random.split(key)
                latent_state = self.filter_after_one_step(
                    model=model_state.module,
                    latent_state=latent_state,
                    observation_embedding=observation_embedding[i],
                    key=subkey,
                )

            if i == 0:
                latent_states = random_variable_list(latent_state)
            else:
                latent_states.append(latent_state)

        latent_states = latent_states.stack()

        key, subkey = jax.random.split(key, 2)
        reconstruction = self._decode(model=model_state.module, latent_states=latent_states, key=subkey)

        return Prediction(reconstruction=reconstruction, posterior=None, prior=latent_states)
