from typing import Union

import chex
import jax.numpy as jnp

from rlab_jax.distributions import (multivariate_normal_diag,
                                    tanh_multivariate_normal_diag,
                                    beta_multivariate_normal_diag,
                                    ImageMultivariateNormalDiag)


def random_variable(sample, mean, std, dist_type="normal"):
    if dist_type == "normal":
        return NormalRandomVariable(sample=sample, mean=mean, std=std)
    elif dist_type == "image_normal":
        return ImageNormalRandomVariable(sample=sample, mean=mean, std=std)
    elif dist_type == "tanh_normal":
        return TanhNormalRandomVariable(sample=sample, mean=mean, std=std)
    elif dist_type == "beta":
        return BetaRandomVariable(sample=sample, mean=mean, std=std)
    else:
        raise ValueError


@chex.dataclass
class Variable:
    sample: jnp.array

    def item(self, item):
        return Variable(sample=self.sample[item])

@chex.dataclass
class RandomVariable:
    sample: jnp.array
    mean: jnp.array
    std: jnp.array

    def __init__(self, sample: jnp.array, mean: jnp.array, std: jnp.array):
        self.sample = sample
        self.mean = mean
        self.std = std

    @property
    def dist(self):
        return None

    @property
    def list_cls(self):
        return RandomVariableList



@chex.dataclass
class NormalRandomVariable(RandomVariable):
    @property
    def dist(self):
        return multivariate_normal_diag(mean=self.mean, std=self.std)

    def item(self, item):
        return NormalRandomVariable(
            sample=self.sample[item],
            mean=self.mean[item],
            std=self.std[item]
        )


@chex.dataclass
class ImageNormalRandomVariable(RandomVariable):
    @property
    def dist(self):
        return ImageMultivariateNormalDiag(loc=self.mean, scale=self.std)

    def item(self, item):
        return ImageMultivariateNormalDiag(
            sample=self.sample[item],
            mean=self.mean[item],
            std=self.std[item]
        )


@chex.dataclass
class TanhNormalRandomVariable(RandomVariable):
    @property
    def dist(self):
        return tanh_multivariate_normal_diag(mean=self.mean, std=self.std)

    def item(self, item):
        return TanhNormalRandomVariable(
            sample=self.sample[item],
            mean=self.mean[item],
            std=self.std[item]
        )

@chex.dataclass
class BetaRandomVariable(RandomVariable):
    @property
    def dist(self):
        return beta_multivariate_normal_diag(mean=self.mean, std=self.std)

    def item(self, item):
        return BetaRandomVariable(
            sample=self.sample[item],
            mean=self.mean[item],
            std=self.std[item]
        )

@chex.dataclass
class DeterministicStochasticRandomVariable:
    deterministic: jnp.array
    stochastic: RandomVariable

    def __init__(self, deterministic: jnp.array, stochastic: RandomVariable):
        self.deterministic = deterministic
        self.stochastic = stochastic

    @property
    def sample(self):
        return jnp.concatenate((self.deterministic, self.stochastic.sample), axis=-1)

    @property
    def sample_deterministic_stochastic(self):
        return jnp.concatenate((self.deterministic, self.stochastic.sample), axis=-1)

    @property
    def mean(self):
        return jnp.concatenate((self.deterministic, self.stochastic.mean), axis=-1)

    @property
    def list_cls(self):
        return DeterministicStochasticRandomVariableList

    def item(self, item):
        return DeterministicStochasticRandomVariable(
            deterministic=self.deterministic[item],
            stochastic=self.stochastic.item(item)
        )


@chex.dataclass
class StateActionRandomVariable:
    action: RandomVariable
    state: Union[RandomVariable, DeterministicStochasticRandomVariable]

    def __init__(
        self,
        action: RandomVariable,
        state: Union[RandomVariable, DeterministicStochasticRandomVariable],
    ):
        self.action = action
        self.state = state

    @property
    def sample(self):
        return self.state.sample

    def item(self, item):
        return StateActionRandomVariable(
            action=self.action.item(item),
            state=self.state.item(item)
        )

    @property
    def mean(self):
        return self.state.mean

    @property
    def std(self):
        return self.state.std

    @property
    def list_cls(self):
        return StateActionRandomVariableList


def random_variable_list(rv, initial_input=True):
    if initial_input:
        return rv.list_cls(rv)
    else:
        return rv.list_cls()


class VariableList:
    def __init__(self, v: Variable = None):
        self.sample = []

        if v is not None:
            self.sample.append(v.sample)

    def append(self, v: Variable):
        self.sample.append(v.sample)

    def __getitem__(self, item):
        return Variable(sample=self.sample[item])

    def stack(self):
        return Variable(sample=jnp.stack(self.sample))

    def stack_with_ind(self, start=None, end=None):
        return Variable(sample=jnp.stack(self.sample)[start:end])


class RandomVariableList:
    def __init__(self, rv: RandomVariable = None):
        self.sample = []
        self.mean = []
        self.std = []
        self.dist_type = None

        if rv is not None:
            self.sample.append(rv.sample)
            self.mean.append(rv.mean)
            self.std.append(rv.std)
            self.dist_type = rv.dist.params["dist_type"]

    def append(self, rv: RandomVariable):
        self.sample.append(rv.sample)
        self.mean.append(rv.mean)
        self.std.append(rv.std)

        if self.dist_type is None:
            self.dist_type = rv.dist.params["dist_type"]

    def __getitem__(self, item):
        return random_variable(
            sample=self.sample[item],
            mean=self.mean[item],
            std=self.std[item],
            dist_type=self.dist_type,
        )

    def stack(self):
        return random_variable(
            sample=jnp.stack(self.sample),
            mean=jnp.stack(self.mean),
            std=jnp.stack(self.std),
            dist_type=self.dist_type,
        )

    def stack_with_ind(self, start=None, end=None):
        return random_variable(
            sample=jnp.stack(self.sample)[start:end],
            mean=jnp.stack(self.mean)[start:end],
            std=jnp.stack(self.std)[start:end],
            dist_type=self.dist_type,
        )


class StateActionRandomVariableList:
    def __init__(self, rv: StateActionRandomVariable = None):
        self.state = None
        self.action = None

        if rv is not None:
            self.state = random_variable_list(rv.state)
            self.action = random_variable_list(rv.action)

    def append(self, rv: StateActionRandomVariable):
        if self.state is None and self.action is None:
            self.state = random_variable_list(rv.state)
            self.action = random_variable_list(rv.action)
        else:
            self.state.append(rv.state)
            self.action.append(rv.action)

    def __getitem__(self, item):
        return StateActionRandomVariable(state=self.state[item], action=self.action[item])

    def stack(self):
        return StateActionRandomVariable(state=self.state.stack(), action=self.action.stack())

    def stack_with_ind(self, start=None, end=None):
        return StateActionRandomVariable(
            state=self.state.stack_with_ind(start=start, end=end),
            action=self.action.stack_with_ind(start=start, end=end),
        )


class DeterministicStochasticRandomVariableList:
    def __init__(self, rv: DeterministicStochasticRandomVariable = None):
        self.deterministic = list()
        self.stochastic = RandomVariableList()

        if rv is not None:
            self.append(rv)

    def append(self, rv: DeterministicStochasticRandomVariable):
        self.deterministic.append(rv.deterministic)
        self.stochastic.append(rv.stochastic)

    def __getitem__(self, item):
        return DeterministicStochasticRandomVariable(
            deterministic=self.deterministic[item], stochastic=self.stochastic[item]
        )

    def stack(self):
        return DeterministicStochasticRandomVariable(
            deterministic=jnp.stack(self.deterministic), stochastic=self.stochastic.stack()
        )

    def stack_with_ind(self, start=None, end=None):
        return DeterministicStochasticRandomVariable(
            deterministic=jnp.stack(self.deterministic)[start:end],
            stochastic=self.stochastic.stack_with_ind(start=start, end=end),
        )
