import chex
import equinox as eqx
import jax
import jax.numpy as jnp

from rlab_jax.rlab_dataclasses import Transition


@chex.dataclass
class ReplayMemoryState:
    data: jnp.array
    size: jnp.array
    current_size: jnp.array
    index: jnp.array


class SequenceReplayMemory:
    def __init__(self, size, env, window_size):
        self.size = size
        self.n_observations = env.observation_space.shape[0]
        self.n_actions = env.action_space.shape[0]
        self.episode_length = env.spec.max_episode_steps
        self.window_size = window_size

    def initial_state(self, key, size=None, validation_ratio=None):
        if size is None:
            size = self.size
        self.size = size

        if validation_ratio is not None:
            validation_size = int(validation_ratio * self.size)
        else:
            validation_size = 0

        key, subkey = jax.random.split(key)
        self.validation_indices = jax.random.choice(key=key, a=jnp.arange(self.size), shape=(validation_size,), replace=False)
        self.train_indices = jnp.arange(self.size)
        self.train_indices = self.train_indices[jnp.logical_not(jnp.isin(self.train_indices, self.validation_indices))]

        return ReplayMemoryState(
            data=Transition(
                action=jnp.zeros((size, self.episode_length, self.n_actions)),
                observation=jnp.zeros((size, self.episode_length, self.n_observations)),
                next_observation=jnp.zeros((size, self.episode_length, self.n_observations)),
                reward=jnp.zeros((size, self.episode_length, 1)),
                done=jnp.zeros((size, self.episode_length, 1)),
                mask=jnp.zeros((size, self.episode_length, 1)),
            ),
            size=jnp.array(size),
            current_size=jnp.array(0),
            index=jnp.array(0),
        )

    def sample(self, memory_state: ReplayMemoryState, batch_size: int, key: jax.random.PRNGKey, validation=False):
        _, batch_key, window_key = jax.random.split(key, 3)

        trajectory_len = memory_state.data.mask.sum(1)
        batch_indices_weight = trajectory_len.squeeze(-1) / trajectory_len.sum() * jnp.array(trajectory_len >= self.window_size, dtype=jnp.float32).squeeze(-1)

        if validation is False:
            batch_indices = jax.random.choice(key=batch_key, a=self.train_indices, p=batch_indices_weight[self.train_indices], shape=(batch_size,))
        else:
            batch_indices = jax.random.choice(key=batch_key, a=self.validation_indices, p=batch_indices_weight[self.validation_indices], shape=(batch_size,))

        # prioritize start and end
        def sample_window_ind(i, w_key):
            return jnp.clip(jax.random.randint(w_key, shape=(1,), minval=-self.window_size, maxval=trajectory_len[i]), a_min=0, a_max=trajectory_len[i] - self.window_size).astype(jnp.int32)[0]

        window_indices = jax.vmap(sample_window_ind)(batch_indices, jax.random.split(window_key, batch_size))

        batch = jax.tree_map(lambda data: data[batch_indices], memory_state.data)

        add_window = jax.vmap(
            lambda data, i: data[i, window_indices[i] + jnp.arange(self.window_size)],
            in_axes=[None, 0],
            out_axes=1,
        )

        return jax.tree_map(lambda data: add_window(data, jnp.arange(batch_size)), batch)

    @eqx.filter_jit
    def push(
        self,
        memory_state: ReplayMemoryState,
        observation: jnp.array,
        next_observation: jnp.array,
        action: jnp.array,
        reward: jnp.array,
        done: jnp.array,
        mask: jnp.array
    ):
        transition = Transition(observation=observation, next_observation=next_observation, action=action, reward=reward, done=done, mask=mask)

        def add(data, transition):
            # TODO: dynamic_update_slice_in_dim will adjust index if batch_size is to large to fit into the buffer
            return jax.lax.dynamic_update_slice_in_dim(data, transition, memory_state.index, axis=0)

        memory_state.data = jax.tree_map(add, memory_state.data, transition)
        memory_state.index = (
            jnp.minimum(memory_state.index + transition.observation.shape[0], memory_state.size)
            % memory_state.size
        )
        memory_state.current_size = jnp.minimum(
            memory_state.current_size + transition.observation.shape[0], memory_state.size
        )

        return memory_state


class ReplayMemory(SequenceReplayMemory):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, window_size=1)

    def sample(self, memory_state: ReplayMemoryState, batch_size: int, key: jax.random.PRNGKey, validation=False):
        _, batch_key, window_key = jax.random.split(key, 3)

        trajectory_len = memory_state.data.mask.sum(1)
        batch_indices_weight = trajectory_len.squeeze(-1) / trajectory_len.sum() * jnp.array(trajectory_len >= self.window_size, dtype=jnp.float32).squeeze(-1)

        if validation is False:
            batch_indices = jax.random.choice(key=batch_key, a=self.train_indices, p=batch_indices_weight[self.train_indices], shape=(batch_size,))
        else:
            batch_indices = jax.random.choice(key=batch_key, a=self.validation_indices, p=batch_indices_weight[self.validation_indices], shape=(batch_size,))

        def sample_ind(i, w_key):
            return jax.random.randint(w_key, shape=(1,), minval=0, maxval=trajectory_len[i])[0]

        sample_indices = jax.vmap(sample_ind)(batch_indices, jax.random.split(window_key, batch_size))

        return jax.tree_map(lambda data: data[batch_indices, sample_indices], memory_state.data)


class ImageSequenceReplayMemory:
    def __init__(self, size, env, window_size):
        self.size = size
        self.observation_size = list(env.observation_space.shape)
        self.n_observations = env.observation_space.shape[0] * env.observation_space.shape[1] * env.observation_space.shape[2]
        self.n_actions = env.action_space.shape[0]
        self.episode_length = env.spec.max_episode_steps
        self.window_size = window_size
        self.preprocess_observation = env.preprocess_observation

    def initial_state(self, key, size=None, validation_ratio=None):
        if size is None:
            size = self.size
        else:
            self.size = size

        if validation_ratio is not None:
            validation_size = int(validation_ratio * self.size)
        else:
            validation_size = 0

        key, subkey = jax.random.split(key)
        self.validation_indices = jax.random.choice(key=key, a=jnp.arange(self.size), shape=(validation_size,), replace=False)
        self.train_indices = jnp.arange(self.size)
        self.train_indices = self.train_indices[jnp.logical_not(jnp.isin(self.train_indices, self.validation_indices))]

        return ReplayMemoryState(
            data=Transition(
                action=jnp.zeros((size, self.episode_length, self.n_actions)),
                observation=jnp.zeros((size, self.episode_length + 1, self.n_observations), dtype=jnp.uint8),
                reward=jnp.zeros((size, self.episode_length, 1)),
                done=jnp.zeros((size, self.episode_length, 1)),
                mask=jnp.zeros((size, self.episode_length, 1)),
            ),
            size=jnp.array(size),
            current_size=jnp.array(0),
            index=jnp.array(0),
        )

    def sample(self, memory_state: ReplayMemoryState, batch_size: int, key: jax.random.PRNGKey, validation=False):
        _, batch_key, window_key = jax.random.split(key, 3)

        trajectory_len = memory_state.data.mask.sum(1)
        batch_indices_weight = trajectory_len.squeeze(-1) / trajectory_len.sum() * jnp.array(trajectory_len >= self.window_size, dtype=jnp.float32).squeeze(-1)

        if validation is False:
            batch_indices = jax.random.choice(key=batch_key, a=self.train_indices, p=batch_indices_weight[self.train_indices], shape=(batch_size,))
        else:
            batch_indices = jax.random.choice(key=batch_key, a=self.validation_indices, p=batch_indices_weight[self.validation_indices], shape=(batch_size,))

        def sample_window_ind(i, w_key):
            return jnp.clip(jax.random.randint(w_key, shape=(1,), minval=-self.window_size, maxval=trajectory_len[i]), a_min=0, a_max=trajectory_len[i] - self.window_size).astype(jnp.int32)[0]

        window_indices = jax.vmap(sample_window_ind)(batch_indices, jax.random.split(window_key, batch_size))

        b = jax.tree_map(lambda data: data[batch_indices], memory_state.data)

        add_window = jax.vmap(
            lambda data, i: data[i, window_indices[i] + jnp.arange(self.window_size)],
            in_axes=[None, 0],
            out_axes=1,
        )

        add_window_p1 = jax.vmap(
            lambda data, i: data[i, window_indices[i] + jnp.arange(self.window_size) + 1],
            in_axes=[None, 0],
            out_axes=1,
        )

        batch = jax.tree_map(lambda data: add_window(data, jnp.arange(batch_size)), b)
        batch_p1 = jax.tree_map(lambda data: add_window_p1(data, jnp.arange(batch_size)), b)
        batch.next_observation = batch_p1.observation

        shape = [batch.observation.shape[0], batch.observation.shape[1]]
        shape.extend(self.observation_size)

        batch.observation = batch.observation.reshape(shape)
        batch.next_observation = batch.next_observation.reshape(shape)

        batch.observation = self.preprocess_observation(batch.observation)
        batch.next_observation = self.preprocess_observation(batch.next_observation)

        return batch

    @eqx.filter_jit
    def push(
        self,
        memory_state: ReplayMemoryState,
        observation: jnp.array,
        next_observation: jnp.array,
        action: jnp.array,
        reward: jnp.array,
        done: jnp.array,
        mask: jnp.array,
        *args,
        **kwargs
    ):
        obs = jnp.concatenate((observation, jnp.expand_dims(next_observation[:, -1], 1)), axis=1)
        transition = Transition(
            observation=obs.reshape(obs.shape[0], obs.shape[1], -1), action=action, reward=reward, done=done, mask=mask
        )

        def add(data, transition):
            # TODO: dynamic_update_slice_in_dim will adjust index if batch_size is to large to fit into the buffer
            return jax.lax.dynamic_update_slice_in_dim(data, transition, memory_state.index, axis=0)

        memory_state.data = jax.tree_map(add, memory_state.data, transition)
        memory_state.index = (
            jnp.minimum(memory_state.index + transition.observation.shape[0], memory_state.size)
            % memory_state.size
        )
        memory_state.current_size = jnp.minimum(
            memory_state.current_size + transition.observation.shape[0], memory_state.size
        )

        return memory_state