import chex
import equinox as eqx
import jax.numpy as jnp
import jax.random
import numpy as np

from rlab_jax.rlab_dataclasses import (DreamRolloutTransition,
                                       RandomVariableList,
                                       Transition, random_variable,
                                       random_variable_list)


class MemoryRollout:
    def __init__(self, env, memory, act_fn):
        self.env = env
        self.memory = memory
        self.act_fn = act_fn

    def filter_fn(
        self, model_state, next_observation, latent_state, action, latent_action, key, reset=False
    ):
        return next_observation

    def generate(self, memory_state, agent_state, steps_per_rollout, key, model_state=None):
        observations, next_observations, actions, rewards, dones, masks = [], [], [], [], [], []

        keys = jax.random.split(key, 3)
        observation, info = self.env.reset(seed=keys[2])
        latent_state = self.filter_fn(
            model_state=model_state,
            next_observation=self.env.preprocess_observation(observation),
            latent_state=None,
            action=None,
            latent_action=None,
            key=keys[1],
            reset=True,
        )

        # TODO: FIX ROLLOUT FOR ENVS WHICH CAN TERMINATE EARLY
        was_done = np.zeros((observation.shape[0]))
        for step in range(steps_per_rollout):
            keys = jax.random.split(keys[0], 3)

            action, latent_action = self.act_fn(
                agent_state=agent_state, model_state=model_state, latent_state=latent_state, key=keys[1], observation=observation,
            )

            next_observation, reward, terminated, truncated, info = self.env.step(
                np.asarray(action.sample)
            )
            mask = np.logical_not(was_done)

            observations.append(observation)
            actions.append(action.sample)
            rewards.append(reward)
            dones.append(terminated)
            masks.append(mask)
            next_observations.append(next_observation)

            latent_state = self.filter_fn(
                model_state=model_state,
                next_observation=self.env.preprocess_observation(next_observation),
                latent_state=latent_state,
                action=action,
                latent_action=latent_action,
                key=keys[2],
            )
            observation = next_observation
            was_done = np.logical_or(np.logical_or(terminated, truncated), was_done)

            if was_done.all():
                was_done = np.zeros((observation.shape[0]))
                keys = jax.random.split(keys[0], 2)
                observation, info = self.env.reset()
                latent_state = self.filter_fn(
                    model_state=model_state,
                    next_observation=self.env.preprocess_observation(observation),
                    latent_state=None,
                    action=None,
                    latent_action=None,
                    key=keys[1],
                    reset=True,
                )

        observations = np.stack(observations, axis=1)
        next_observations = np.stack(next_observations, axis=1)
        actions = np.stack(actions, axis=1)
        rewards = np.stack(rewards, axis=1)
        dones = np.stack(dones, axis=1)
        masks = np.stack(masks, axis=1)

        if observations.shape[1] > self.env.spec.max_episode_steps:

            def split_trajectories(x):
                shape = list(x.shape)
                shape[0] = -1
                shape[1] = self.env.spec.max_episode_steps
                return x.reshape(shape)

            observations = split_trajectories(observations)
            next_observations = split_trajectories(next_observations)
            actions = split_trajectories(actions)
            rewards = split_trajectories(rewards)
            dones = split_trajectories(dones)
            masks = split_trajectories(masks)

        dones = np.expand_dims(dones, -1).astype(float)
        masks = np.expand_dims(masks, -1).astype(float)

        memory_state = self.memory.push(
            memory_state=memory_state,
            observation=observations,
            next_observation=next_observations,
            action=actions,
            reward=rewards,
            done=dones,
            mask=masks
        )

        return memory_state


class BatchRollout:
    def __init__(self, env, act_fn, sequence_rollout=False):
        self.env = env
        self.act_fn = act_fn
        self.sequence_rollout = sequence_rollout

    def filter_fn(
        self, model_state, next_observation, latent_state, action, latent_action, key, reset=False
    ):
        return next_observation

    def generate(self, agent_state, steps_per_rollout, key, model_state=None):
        observations, next_observations, actions, latent_actions, rewards, masks, dones = [], [], [], [], [], [], []

        keys = jax.random.split(key, 3)
        observation, info = self.env.reset(seed=keys[2])
        observation = self.env.preprocess_observation(observation)
        latent_state = self.filter_fn(
            model_state=model_state,
            next_observation=observation,
            latent_state=None,
            action=None,
            latent_action=None,
            key=keys[1],
            reset=True,
        )

        if hasattr(latent_state, "list_cls"):
            random_variable_state = True
            latent_states = random_variable_list(latent_state, initial_input=False)
        else:
            random_variable_state = False
            latent_states = []

        was_done = np.zeros((observation.shape[0]))
        for step in range(steps_per_rollout):
            keys = jax.random.split(keys[0], 3)

            action, latent_action = self.act_fn(
                agent_state=agent_state, model_state=model_state, latent_state=latent_state, key=keys[1], observation=observation
            )

            next_observation, reward, terminated, truncated, info = self.env.step(np.asarray(action.sample))
            next_observation = self.env.preprocess_observation(next_observation)
            mask = np.logical_not(was_done)

            observations.append(observation)
            actions.append(action.sample)
            rewards.append(reward)
            masks.append(mask)
            dones.append(terminated)
            next_observations.append(next_observation)
            latent_states.append(latent_state)
            if latent_action is not None:
                latent_actions.append(latent_action.sample)
            latent_state = self.filter_fn(
                model_state=model_state,
                next_observation=next_observation,
                latent_state=latent_state,
                action=action,
                latent_action=latent_action,
                key=keys[2],
            )
            observation = next_observation
            was_done = np.logical_or(terminated, was_done)

            if was_done.all() or truncated.all():
                latent_states.append(latent_state)
                break

        observations = jnp.stack(observations)
        next_observations = jnp.stack(next_observations)
        actions = jnp.stack(actions)
        dones = jnp.stack(dones)
        rewards = np.stack(rewards)
        masks = np.expand_dims(np.stack(masks), -1).astype(float)

        # set rewards to zero for terminated environments
        rewards[jnp.logical_not(masks.astype(jnp.bool_))] = 0.0

        if latent_action is not None:
            latent_actions = jnp.stack(latent_actions)
        else:
            latent_actions = None

        if random_variable_state is True:
            latent_state = latent_states.stack_with_ind(end=-1)
            next_latent_state = latent_states.stack_with_ind(start=1)
        else:
            latent_state = jnp.stack(latent_states[:-1])
            next_latent_state = jnp.stack(latent_states[1:])

        return Transition(
            observation=observations,
            next_observation=next_observations,
            action=actions,
            reward=jnp.array(rewards),
            latent_action=latent_actions,
            latent_state=latent_state,
            next_latent_state=next_latent_state,
            mask=masks,
            done=dones,
        )


class DreamRollout:
    def __init__(self, sequence_model, act_fn, steps_per_rollout, n_rollouts):
        self.sequence_model = sequence_model
        self.act_fn = act_fn
        self.steps_per_rollout = steps_per_rollout
        self.n_rollouts = n_rollouts

    def generate(self, policy, model_state, batch, predictions, key: jax.random.PRNGKey):

        def sample_init_state_fn(k):
            ks = jax.random.split(k, 2)

            i = jax.random.randint(
                ks[0], minval=self.sequence_model.n_initial_filtered_steps, maxval=predictions.posterior.sample.shape[0], shape=(1,)
            )
            j = jax.random.randint(
                ks[1], minval=0, maxval=predictions.posterior.sample.shape[1], shape=(1,)
            )
            initial_state = jax.tree_map(lambda data: data[i, j].squeeze(0), predictions.posterior)
            initial_done = jax.tree_map(lambda data: data[i, j].squeeze(0), batch.done)

            return initial_state, initial_done

        latent_state, initial_done = jax.lax.stop_gradient(
            jax.vmap(sample_init_state_fn)(jax.random.split(key, self.n_rollouts))
        )

        key, subkey = jax.random.split(key, 2)

        latent_states = random_variable_list(latent_state)
        actions = RandomVariableList()
        latent_actions = RandomVariableList()

        keys = jax.random.split(key, 2)

        for step in range(self.steps_per_rollout):
            keys = jax.random.split(keys[0], 6)
            action, latent_action = self.act_fn(
                policy=policy, latent_state=latent_state, model_state=model_state, key=keys[1])

            latent_state = self.sequence_model.one_step(
                model=model_state.module,
                latent_state=latent_state,
                action=action,
                latent_action=latent_action,
                key=keys[2],
            )

            latent_states.append(latent_state)
            actions.append(action)
            if latent_action is not None:
                latent_actions.append(latent_action)

        actions = actions.stack()

        if latent_action is not None:
            latent_actions = latent_actions.stack()
        else:
            latent_actions = None

        states = latent_states.stack_with_ind(end=-1)
        next_states = latent_states.stack_with_ind(start=1)

        rewards = model_state.module.reward(
            action=actions.sample,
            next_state=next_states,
            key=keys[4],
        )

        dones = model_state.module.done(
            action=actions.sample,
            next_state=next_states,
            key=keys[5],
            model_state=model_state
        )

        # overwrite first time step with value from batch
        dones = jnp.concatenate((jnp.expand_dims(initial_done, 0), dones[1:]), axis=0)

        # time step corrected!
        was_done = jnp.concatenate((jnp.zeros_like(jnp.expand_dims(initial_done, 0)), dones[:-1]), axis=0)

        return DreamRolloutTransition(
            state=states, next_state=next_states, action=actions, reward=rewards, done=dones, latent_action=latent_actions, was_done=was_done
        )
