import os

import h5py
import jax.numpy as jnp
import tqdm
import matplotlib.pyplot as plt
import jax


def get_keys(h5file):
    keys = []

    def visitor(name, item):
        if isinstance(item, h5py.Dataset):
            keys.append(name)

    h5file.visititems(visitor)
    return keys


def load(h5path, env):
    data_dict = {}
    with h5py.File(h5path, "r") as dataset_file:
        for k in tqdm.tqdm(get_keys(dataset_file), desc="load datafile"):
            try:  # first try loading as an array
                data_dict[k] = dataset_file[k][:]
            except ValueError as e:  # try loading as a scalar
                data_dict[k] = dataset_file[k][()]

    # Run a few quick sanity checks
    for key in ["observations", "actions", "rewards", "terminals"]:
        assert key in data_dict, "Dataset is missing key %s" % key
    N_samples = data_dict["observations"].shape[0]
    if env.observation_space.shape is not None:
        assert (
            data_dict["observations"].shape[1:] == env.observation_space.shape
        ), "Observation shape does not match env: %s vs %s" % (
            str(data_dict["observations"].shape[1:]),
            str(env.observation_space.shape),
        )
    assert (
        data_dict["actions"].shape[1:] == env.action_space.shape
    ), "Action shape does not match env: %s vs %s" % (
        str(data_dict["actions"].shape[1:]),
        str(env.action_space.shape),
    )
    if data_dict["rewards"].shape == (N_samples, 1):
        data_dict["rewards"] = data_dict["rewards"][:, 0]
    assert data_dict["rewards"].shape == (N_samples,), "Reward has wrong shape: %s" % (
        str(data_dict["rewards"].shape)
    )
    if data_dict["terminals"].shape == (N_samples, 1):
        data_dict["terminals"] = data_dict["terminals"][:, 0]
    assert data_dict["terminals"].shape == (N_samples,), "Terminals has wrong shape: %s" % (
        str(data_dict["rewards"].shape)
    )
    return data_dict


def get_offline_dataset(env, memory, prng_key, writer=None, validation_ratio=None):
    dataset = env.get_dataset()

    # metadata might cause an error when using d4rl.sequence_dataset
    keys = list(dataset.keys())
    for key in keys:
        if key not in ["observations", "actions", "next_observations", "rewards", "terminals", "timeouts"]:
            del dataset[key]

    from d4rl import sequence_dataset
    sequence_dataset = sequence_dataset(env=env, dataset=dataset)

    trajectories = []
    max_len = 0
    min_len = env.spec.max_episode_steps
    for trajectory in sequence_dataset:
        trajectories.append(trajectory)
        max_len = max(max_len, trajectory["actions"].shape[0])
        min_len = min(min_len, trajectory["actions"].shape[0])

    memory.episode_length = max_len
    memory_state = memory.initial_state(size=len(trajectories), validation_ratio=validation_ratio, key=prng_key)

    for trajectory in trajectories:
        if "next_observations" in trajectory.keys():
            memory_state = memory.push(
                memory_state=memory_state,
                observation=jnp.expand_dims(trajectory["observations"], 0),
                next_observation=jnp.expand_dims(trajectory["next_observations"], 0),
                action=jnp.expand_dims(trajectory["actions"], 0),
                reward=jnp.expand_dims(jnp.expand_dims(trajectory["rewards"], 0), -1),
                done=jnp.expand_dims(jnp.expand_dims(trajectory["terminals"].astype(jnp.float32), 0), -1),
                mask=jnp.ones((1, trajectory["actions"].shape[0], 1))
            )
        else:
            memory_state = memory.push(
                memory_state=memory_state,
                observation=jnp.expand_dims(trajectory["observations"][:-1], 0),
                next_observation=jnp.expand_dims(trajectory["observations"][1:], 0),
                action=jnp.expand_dims(trajectory["actions"][:-1], 0),
                reward=jnp.expand_dims(jnp.expand_dims(trajectory["rewards"][:-1], 0), -1),
                done=jnp.expand_dims(jnp.expand_dims(trajectory["terminals"].astype(jnp.float32), 0), -1),
                mask=jnp.ones((1, trajectory["actions"][:-1].shape[0], 1))
            )


    if writer is not None:
        dataset_reward_histogram(memory_state.data.reward, writer)

    return memory, memory_state


def dataset_reward_histogram(reward, writer):
    fig, axs = plt.subplots(figsize=(10, 10))
    axs.set_xlabel("step")
    axs.set_ylabel("reward")
    y_ticks = jnp.array([5, 10, 15, 20, 25, 30, 35, 40, 45])
    axs.set_yticks(y_ticks)
    fig.tight_layout()

    bins = jnp.linspace(start=-5, stop=20, num=51)
    hist = jax.vmap(
        lambda l: jnp.histogram(reward[:, l], bins=bins, range=None, weights=None, density=True))(
        jnp.arange(reward.shape[1]))

    axs.imshow(hist[0].T, aspect="auto", origin='lower')
    axs.yaxis.set_ticks(jnp.arange(5, 50, 5))

    new_y_ticks = hist[1][0][y_ticks].round(1)
    axs.set_yticklabels(new_y_ticks, minor=False)

    writer.add_figure("DatasetRewardHistFigure", fig, 0)