import jax.random
import omegaconf
from tensorboardX import SummaryWriter

from rlab_jax.evaluation.evaluation import ExperimentEvaluation
from rlab_jax.experiments.mb_dreamer import ModelBasedDreamerExperiment
from rlab_jax.utils.cfg import save_cfg
from rlab_jax.utils.rollout import BatchRollout

class LatentActionModelBasedDreamerExperiment(ModelBasedDreamerExperiment):
    _EXPERIMENT_TYPE = "LatentActionModelBasedDreamer"

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        def prior_act(model_state, latent_state, key, **ignored):
            key, subkey = jax.random.split(key)

            latent_action = self.sequence_model.latent_action_model._prior(model=model_state.module.latent_action_model, state=latent_state.sample, key=key)

            # reconstruct action
            action_reconstructed = self.agent.decode_fn(
                model=model_state.module.latent_action_model,
                latent_action=latent_action.sample,
                state=latent_state.sample,
                deterministic=True,
                key=subkey,
            )

            return action_reconstructed, latent_action

        self.prior_val_rollout = BatchRollout(env=self.val_env, act_fn=prior_act, sequence_rollout=True)
        self.prior_val_rollout.filter_fn = self.sequence_model.filter

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, **_ignored):
        from rlab_jax.envs.env import env_from_cfg
        from rlab_jax.modules import Agent, SequenceModel
        from rlab_jax.utils.logs import get_outputs_path

        train_env = env_from_cfg(cfg.env, n_envs=cfg.experiment.train_rollouts_per_epoch)
        val_env = env_from_cfg(cfg.env, n_envs=cfg.experiment.val_rollouts_per_epoch)

        # num latent actions = 2 * num actions
        if cfg.seqm.latent_action_model.n_latent_actions is None:
            cfg.seqm.latent_action_model.n_latent_actions = train_env.action_space.shape[0] * 2

        sequence_model = SequenceModel.create(
            cfg=cfg.seqm,
            n_observations=train_env.observation_space.shape if len(train_env.observation_space.shape) > 1 else train_env.observation_space.shape[0],
            n_actions=train_env.action_space.shape[0],
            env_name=cfg.env.name,
        )

        agent = Agent.create(
            cfg=cfg.agent,
            n_inputs=sequence_model.n_latent,
            n_actions=cfg.seqm.latent_action_model.n_latent_actions,
            decode_fn=sequence_model.latent_action_model._decode,
        )

        log_dir = get_outputs_path(cfg)
        writer = SummaryWriter(log_dir=log_dir)
        save_cfg(log_dir, cfg)

        evaluation = ExperimentEvaluation(
            cfg=cfg,
            env=val_env,
            agent=agent,
            sequence_model=sequence_model,
            writer=writer,
            evaluations_every_n_epochs=cfg.evaluation.every_n_epochs
        )

        print(
            f"""
        ################################
        Env: {cfg.env.name}
        Experiment: {cfg.experiment.type}
        SequenceModel: {cfg.seqm.type}
        Agent: {cfg.agent.type}
        LogDir: {log_dir}
        ################################
        """
        )

        return cls(
            train_env=train_env,
            val_env=val_env,
            agent=agent,
            sequence_model=sequence_model,
            writer=writer,
            evaluation=evaluation,
            log_dir=log_dir,
            **cfg.experiment,
        )
