import equinox as eqx
import jax.random
import omegaconf
import tqdm
from tensorboardX import SummaryWriter

from rlab_jax.evaluation.evaluation import ExperimentEvaluation
from rlab_jax.experiments import ModelFreeOffPolicyExperiment
from rlab_jax.rlab_dataclasses import ModulesState
from rlab_jax.utils.cfg import save_cfg
from rlab_jax.utils.keygen import keygen


class LatentActionModelFreeOffPolicyExperiment(ModelFreeOffPolicyExperiment):
    _EXPERIMENT_TYPE = "LatentActionModelFreeOffPolicyExperiment"

    def __init__(
        self,
        latent_action_model,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.latent_action_model = latent_action_model

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, **_ignored):
        from rlab_jax.envs.env import env_from_cfg
        from rlab_jax.modules.agents import Agent
        from rlab_jax.utils.logs import get_outputs_path
        from rlab_jax_projects.latent_actions import LatentActionVAE

        train_env = env_from_cfg(cfg.env, n_envs=cfg.experiment.train_rollouts_per_epoch)
        val_env = env_from_cfg(cfg.env, n_envs=cfg.experiment.val_rollouts_per_epoch)

        if cfg.seqm.latent_action_model.type == "LatentActionVAE":
            latent_action_model = LatentActionVAE.from_cfg(
                cfg=cfg.seqm.latent_action_model,
                n_states=train_env.observation_space.shape[0],
                n_actions=train_env.action_space.shape[0],
            )
        else:
            raise ValueError

        agent = Agent.create(
            cfg=cfg.agent,
            n_inputs=train_env.observation_space.shape[0],
            n_actions=train_env.action_space.shape[0],
            n_latent_actions=cfg.seqm.latent_action_model.n_latent_actions,
            decode_fn=latent_action_model._decode,
        )

        log_dir = get_outputs_path(cfg)
        writer = SummaryWriter(log_dir=log_dir)
        save_cfg(log_dir, cfg)

        evaluation = ExperimentEvaluation(
            cfg=cfg, env=val_env, agent=agent, sequence_model=None, reward=None, writer=writer, evaluations_every_n_epochs=cfg.evaluation.every_n_epochs
        )

        print(
            f"""
        ################################
        Env: {cfg.env.name}
        Experiment: {cfg.experiment.type}
        Agent: {cfg.agent.type}
        LogDir: {log_dir}
        ################################
        """
        )

        return cls(
            train_env=train_env,
            val_env=val_env,
            agent=agent,
            latent_action_model=latent_action_model,
            writer=writer,
            evaluation=evaluation,
            log_dir=log_dir,
            **cfg.experiment,
        )

    def initial_state(self):
        modules_state = ModulesState(agent_state=self.agent.initial_state(),
                                     model_state=self.latent_action_model.initial_state())

        if self.checkpoint_path is not None:
            modules_state = self.load_checkpoint(path=self.checkpoint_path, modules_state=modules_state)

        return modules_state

    @eqx.filter_jit
    def agent_training_epoch(self, memory_state, modules_state, num_updates_per_epoch, key):
        print("compiling agent training")

        modules_params, modules_static = eqx.partition(modules_state, eqx.is_array)

        def update(carry, _):
            modules_params_, key_ = carry

            keys_ = jax.random.split(key_, 3)

            batch_ = self.train_rollout.memory.sample(
                memory_state=memory_state, batch_size=self.batch_size, key=keys_[0]
            )

            modules_state_ = eqx.combine(modules_params_, modules_static)
            modules_state_.agent_state, agent_log_ = self.agent.training_step(
                agent_state=modules_state_.agent_state, model_state=modules_state_.model_state, batch=batch_, key=keys_[1]
            )
            modules_params_, _ = eqx.partition(modules_state_, eqx.is_array)

            return (modules_params_, keys_[2]), agent_log_

        init_carry = (modules_params, key)
        (modules_params, key), train_logs = jax.lax.scan(
            f=update, init=init_carry, xs=None, length=num_updates_per_epoch
        )
        modules_state = eqx.combine(modules_params, modules_static)

        train_logs = jax.tree_map(lambda v: v.mean(), train_logs)

        return modules_state, train_logs

    @eqx.filter_jit
    def model_training_epoch(self, memory_state, modules_state, num_updates_per_epoch, key):
        print("compiling model training")

        modules_params, modules_static = eqx.partition(modules_state, eqx.is_array)

        def update(carry, _):
            modules_params_, key_ = carry

            keys_ = jax.random.split(key_, 3)

            batch_ = self.train_rollout.memory.sample(
                memory_state=memory_state, batch_size=self.batch_size, key=keys_[0]
            )

            modules_state_ = eqx.combine(modules_params_, modules_static)
            modules_state_.model_state, model_log_ = self.latent_action_model.training_step(
                model_state=modules_state_.model_state, batch=batch_, key=keys_[1]
            )
            modules_params_, _ = eqx.partition(modules_state_, eqx.is_array)

            return (modules_params_, keys_[2]), model_log_

        init_carry = (modules_params, key)
        (modules_params, key), train_logs = jax.lax.scan(
            f=update, init=init_carry, xs=None, length=num_updates_per_epoch
        )
        modules_state = eqx.combine(modules_params, modules_static)

        train_logs = jax.tree_map(lambda v: v.mean(), train_logs)

        return modules_state, train_logs

    def train(self):
        memory_state = self.train_rollout.memory.initial_state()
        modules_state = self.initial_state()

        key, subkey = jax.random.split(keygen())
        memory_state = self.train_rollout.generate(
            memory_state=memory_state,
            model_state=modules_state.model_state,
            agent_state=modules_state.agent_state,
            steps_per_rollout=int(self.replay_memory_init / self.train_rollouts_per_epoch),
            key=subkey,
        )

        for epoch in tqdm.tqdm(range(1, self.epochs + 1), desc="epochs"):
            step = (
                self.replay_memory_init
                + epoch * self.train_rollouts_per_epoch * self.env_steps_per_rollout
            )

            # TRAIN ROLLOUT
            key, subkey = jax.random.split(key)
            memory_state = self.train_rollout.generate(
                memory_state=memory_state,
                agent_state=modules_state.agent_state,
                model_state=modules_state.model_state,
                steps_per_rollout=self.env_steps_per_rollout,
                key=subkey,
            )

            # AGENT TRAINING
            num_updates_per_epoch = int(
                self.train_rollouts_per_epoch * self.env_steps_per_rollout * self.train_steps_ratio
            )
            key, subkey = jax.random.split(key)
            modules_state, train_logs = self.agent_training_epoch(
                memory_state=memory_state,
                modules_state=modules_state,
                num_updates_per_epoch=num_updates_per_epoch,
                key=subkey,
            )

            for k, v in train_logs.items():
                self.writer.add_scalar(k, v.item(), step)

            # MODEL TRAINING
            num_updates_per_epoch = int(
                self.train_rollouts_per_epoch * self.env_steps_per_rollout * self.train_steps_ratio
            )
            key, subkey = jax.random.split(key)
            modules_state, train_logs = self.model_training_epoch(
                memory_state=memory_state,
                modules_state=modules_state,
                num_updates_per_epoch=num_updates_per_epoch,
                key=subkey,
            )

            for k, v in train_logs.items():
                self.writer.add_scalar(k, v.item(), step)

            # VAL
            if epoch % self.val_every_n_epochs == 0:

                # VAL ROLLOUT
                key, subkey = jax.random.split(key)
                batch = self.val_rollout.generate(
                    agent_state=modules_state.agent_state,
                    model_state=modules_state.model_state,
                    steps_per_rollout=self.env_steps_per_rollout,
                    key=subkey,
                )

                self.best_checkpoint(batch=batch,
                                     epoch=epoch,
                                     modules_state=modules_state)

                if self.evaluation is not None:
                    key, subkey = jax.random.split(key)
                    self.evaluation(
                        agent_state=modules_state.agent_state,
                        model_state=modules_state.model_state,
                        batch=batch,
                        step=step,
                        epoch=epoch,
                        key=subkey,
                    )

            if epoch % self.checkpoint_every_n_epochs == 0:
                self.save_checkpoint(step=step, modules_state=modules_state)

    def eval(self, key=keygen()):
        modules_state = self.initial_state()

        batch = self.val_rollout.generate(
            agent_state=modules_state.agent_state,
            model_state=modules_state.model_state,
            steps_per_rollout=self.env_steps_per_rollout,
            key=key,
        )
        return batch.reward.sum(0)
