from rlab_jax.experiments import ModelFreeOffPolicyExperiment
from rlab_jax_projects.latent_actions.d4rl_datasets import get_offline_dataset
from rlab_jax.rlab_dataclasses import ModulesState
from rlab_jax.utils.cfg import save_cfg
from rlab_jax.utils.keygen import keygen
import jax
import tqdm
from tensorboardX import SummaryWriter
import omegaconf
from rlab_jax.evaluation.evaluation import ExperimentEvaluation
import equinox as eqx
from rlab_jax.utils.memory import ReplayMemory
from rlab_jax_projects.latent_actions.modules.mbpo_ensemble import MBPODreamRollout
from copy import deepcopy


class MBPOExperiment(ModelFreeOffPolicyExperiment):
    _EXPERIMENT_TYPE = "MBPOExperiment"

    def __init__(self, model, model_training_epochs, dream_replay_memory_size, dream_rollout_steps, dream_rollout_batch_size, dataset_ratio, validation_ratio, max_model_training_epochs_since_improvement, validation_batch_size, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        self.model_training_epochs = model_training_epochs

        self.dream_rollout_steps = dream_rollout_steps
        self.dream_rollout_batch_size = dream_rollout_batch_size
        self.dataset_ratio = dataset_ratio
        self.validation_ratio = validation_ratio
        self.validation_batch_size = validation_batch_size
        self.max_model_training_epochs_since_improvement = max_model_training_epochs_since_improvement

        self.dream_replay_memory = ReplayMemory(
                size=dream_replay_memory_size,
                env=self.train_env,
            )
        self.dream_replay_memory.episode_length = self.dream_rollout_steps

        self.dream_rollout = MBPODreamRollout(
            model=self.model,
            act_fn=self.agent.act,
            steps_per_rollout=self.dream_rollout_steps,
            memory=self.dream_replay_memory,
        )

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, **_ignored):
        from rlab_jax.envs.env import env_from_cfg
        from rlab_jax.modules import Agent, SequenceModel
        from rlab_jax.utils.logs import get_outputs_path

        train_env = env_from_cfg(cfg.env, n_envs=cfg.experiment.train_rollouts_per_epoch)
        val_env = env_from_cfg(cfg.env, n_envs=cfg.experiment.val_rollouts_per_epoch)

        model = SequenceModel.create(type=cfg.model.type,
                              cfg=cfg.model,
                              n_observations=train_env.observation_space.shape[0],
                              n_actions=train_env.action_space.shape[0],
                              env_name=cfg.env.name,
                              from_cfg=True)

        agent = Agent.create(
            cfg=cfg.agent,
            n_inputs=train_env.observation_space.shape[0],
            n_actions=train_env.action_space.shape[0],
            observation_predict_fn=model.__call__,
            decay_steps=train_env.spec.max_episode_steps * (cfg.experiment.epochs - cfg.experiment.model_training_epochs),
        )



        log_dir = get_outputs_path(cfg)
        writer = SummaryWriter(log_dir=log_dir)
        save_cfg(log_dir, cfg)

        evaluation = ExperimentEvaluation(
            cfg=cfg, env=val_env, agent=agent, sequence_model=None, reward=None, writer=writer, evaluations_every_n_epochs=cfg.evaluation.every_n_epochs
        )

        print(
            f"""
        ################################
        Env: {cfg.env.name}
        Experiment: {cfg.experiment.type}
        Agent: {cfg.agent.type}
        LogDir: {log_dir}
        ################################
        """
        )

        return cls(
            train_env=train_env,
            val_env=val_env,
            agent=agent,
            model=model,
            writer=writer,
            evaluation=evaluation,
            log_dir=log_dir,
            **cfg.experiment,
        )

    def initial_state(self, dont_load_checkpoint=False):
        modules_state = ModulesState(agent_state=self.agent.initial_state(), model_state=self.model.initial_state())

        if not dont_load_checkpoint and self.checkpoint_path is not None:
            # TODO: remove .model state
            modules_state.model_state = self.load_checkpoint(path=self.checkpoint_path, modules_state=modules_state).model_state

        return modules_state

    @eqx.filter_jit
    def model_training_epoch(
        self, memory_state, model_state, num_updates_per_epoch, key
    ):
        print("model training compiling")

        model_params, model_static = eqx.partition(model_state, eqx.is_array)

        def update(carry, _):
            model_params_, key_ = carry

            keys_ = jax.random.split(key_, 5)

            model_state_ = eqx.combine(model_params_, model_static)

            # Sample batch
            batch_ = self.train_rollout.memory.sample(
                memory_state=memory_state, batch_size=self.batch_size, key=keys_[0]
            )

            # Model training step
            model_state_, model_log_ = self.model.training_step(
                model_state=model_state_, batch=batch_, key=keys_[1]
            )

            model_params_, _ = eqx.partition(model_state_, eqx.is_array)

            return (model_params_, keys_[4]), {**model_log_}

        init_carry = (model_params, key)
        (model_params, key), train_logs = jax.lax.scan(
            f=update, init=init_carry, xs=None, length=num_updates_per_epoch
        )
        model_state = eqx.combine(model_params, model_static)

        train_logs = jax.tree_map(lambda v: v.mean(), train_logs)

        return model_state, train_logs

    @eqx.filter_jit
    def agent_training_epoch(self, memory_state, dream_memory_state, modules_state, num_updates_per_epoch, key):
        print("compiling agent training")

        modules_params, modules_static = eqx.partition(modules_state, eqx.is_array)

        def update(carry, _):
            modules_params_, key_ = carry

            keys_ = jax.random.split(key_, 4)

            batch_size = int(self.dataset_ratio * self.batch_size)
            dream_batch_size = self.batch_size - batch_size

            data_batch_ = self.train_rollout.memory.sample(
                memory_state=memory_state, batch_size=batch_size, key=keys_[0]
            )

            dream_batch_ = self.dream_rollout.memory.sample(
                memory_state=dream_memory_state, batch_size=dream_batch_size, key=keys_[1]
            )

            modules_state_ = eqx.combine(modules_params_, modules_static)
            modules_state_.agent_state, agent_log_ = self.agent.training_step(
                agent_state=modules_state_.agent_state, model_state=modules_state_.model_state, data_batch=data_batch_, dream_batch=dream_batch_, key=keys_[2]
            )
            modules_params_, _ = eqx.partition(modules_state_, eqx.is_array)

            return (modules_params_, keys_[3]), agent_log_

        init_carry = (modules_params, key)
        (modules_params, key), train_logs = jax.lax.scan(
            f=update, init=init_carry, xs=None, length=num_updates_per_epoch
        )
        modules_state = eqx.combine(modules_params, modules_static)

        train_logs = jax.tree_map(lambda v: v.mean(), train_logs)

        return modules_state, train_logs

    @eqx.filter_jit
    def model_validation_epoch(self, memory_state, model_state, key):
        print("compiling model validation")
        keys = jax.random.split(key, 2)
        batch = self.train_rollout.memory.sample(memory_state=memory_state, batch_size=self.validation_batch_size, key=keys[0], validation=True)

        model_log = self.model.validation_step(
            model_state=model_state, batch=batch, key=keys[1]
        )

        return model_log

    @staticmethod
    def save_best_model(modules_state, best_modules_state, save_ensemble_indexes):

        for i in save_ensemble_indexes:
            def fn(tree, rest):
                if eqx.is_array(tree):
                    if tree.shape[0] == 7:
                        return tree.at[i].set(jax.numpy.copy(rest[i]))
                    return rest
                else:
                    return tree

            best_modules_state.model_state.module = jax.tree_util.tree_map(fn, best_modules_state.model_state.module, modules_state.model_state.module)

        return best_modules_state

    def train(self):

        dream_memory_state = self.dream_replay_memory.initial_state(key=keygen())

        self.train_rollout.memory, memory_state = get_offline_dataset(env=self.train_env, memory=self.train_rollout.memory, writer=self.writer, prng_key=keygen(), validation_ratio=self.validation_ratio)

        modules_state = self.initial_state(dont_load_checkpoint=True)
        modules_state.model_state = self.model.scaler.fit(data=memory_state.data, model_state=modules_state.model_state)
        if self.checkpoint_path is not None:
            modules_state = self.initial_state()

        key, subkey = jax.random.split(keygen())

        val_loss = jax.numpy.ones((self.model.n_networks)) * 10**5
        epochs_since_last_improvement = 0

        best_modules_state = deepcopy(modules_state)
        stop_model_training = False

        modules_state.model_state.elite_indexes = jax.numpy.arange(self.model.n_elites)

        for epoch in tqdm.tqdm(range(1, self.epochs + 1), desc="epochs"):
            step = epoch * self.train_rollouts_per_epoch * self.env_steps_per_rollout
            num_updates_per_epoch = int(
                self.train_rollouts_per_epoch * self.env_steps_per_rollout * self.train_steps_ratio
            )

            if epoch < self.model_training_epochs + 1:
                if not stop_model_training:
                    # MODEL TRAINING
                    key, subkey = jax.random.split(key)
                    modules_state.model_state, train_logs = self.model_training_epoch(
                        memory_state=memory_state,
                        model_state=modules_state.model_state,
                        num_updates_per_epoch=int(jax.numpy.sum(memory_state.data.mask.sum(1)[self.train_rollout.memory.train_indices]) / self.batch_size),
                        key=subkey,
                    )

                    del train_logs["ensemble/indv_loss"]
                    for k, v in train_logs.items():
                        self.writer.add_scalar(k, v.item(), step)

                    key, subkey = jax.random.split(key)
                    model_val_log = self.model_validation_epoch(memory_state=memory_state, model_state=modules_state.model_state, key=subkey)

                    save_ensemble_indexes = []
                    for i, (old_loss, new_loss) in enumerate(zip(val_loss, model_val_log["ensemble/ind_validation_loss"])):
                        if new_loss < old_loss:
                            if new_loss < 0 and old_loss < 0:
                                improvement = -(old_loss - new_loss) / old_loss
                            else:
                                improvement = (old_loss - new_loss) / old_loss
                        else:
                            improvement = -1

                        if improvement > 0.01:
                            val_loss = val_loss.at[i].set(new_loss)
                            save_ensemble_indexes.append(i)

                    if len(save_ensemble_indexes) > 0:
                        best_modules_state = self.save_best_model(modules_state=modules_state, best_modules_state=best_modules_state, save_ensemble_indexes=save_ensemble_indexes)
                        epochs_since_last_improvement = 0
                    else:
                        epochs_since_last_improvement += 1

                    if epochs_since_last_improvement >= self.max_model_training_epochs_since_improvement or epoch == self.model_training_epochs:
                        stop_model_training = True
                        modules_state = best_modules_state
                        modules_state.model_state.elite_indexes = jax.lax.top_k(-val_loss, k=self.model.n_elites)[1]
                        model_val_log["ensemble/ind_validation_loss"] = jax.numpy.mean(val_loss)

                    model_val_log["ensemble/ind_validation_loss"] = jax.numpy.mean(model_val_log["ensemble/ind_validation_loss"])
                    for k, v in model_val_log.items():
                        self.writer.add_scalar("validation/" + k, v.item(), step)

                else:
                    print("model training stopped")
                    for k, v in model_val_log.items():
                        self.writer.add_scalar("validation/" + k, v.item(), step)

            else:
                # DREAM ROLLOUT
                key, subkey = jax.random.split(key)
                batch = self.train_rollout.memory.sample(memory_state=memory_state,
                                                         batch_size=self.dream_rollout_batch_size,
                                                         key=subkey)

                key, subkey = jax.random.split(key)
                dream_memory_state = self.dream_rollout.generate(agent_state=modules_state.agent_state,
                                                                 model_state=modules_state.model_state,
                                                                 initial_observation=batch.observation,
                                                                 key=subkey,
                                                                 memory_state=dream_memory_state)


                # AGENT TRAINING
                key, subkey = jax.random.split(key)
                modules_state, train_logs = self.agent_training_epoch(
                    memory_state=memory_state,
                    dream_memory_state=dream_memory_state,
                    modules_state=modules_state,
                    num_updates_per_epoch=num_updates_per_epoch,
                    key=subkey,
                )

                train_logs["agent/policy_lr"] = self.agent.lr_schedule(modules_state.agent_state.policy.optim_state[1].count)

                for k, v in train_logs.items():
                    self.writer.add_scalar(k, v.item(), step)

            # VAL
            if epoch % self.val_every_n_epochs == 0:

                # VAL ROLLOUT
                key, subkey = jax.random.split(key)
                batch = self.val_rollout.generate(
                    agent_state=modules_state.agent_state,
                    steps_per_rollout=self.env_steps_per_rollout,
                    key=subkey,
                )

                self.best_checkpoint(batch=batch,
                                     epoch=epoch,
                                     modules_state=modules_state)

                if self.evaluation is not None:
                    key, subkey = jax.random.split(key)
                    self.evaluation(batch=batch, step=step, epoch=epoch, key=subkey)

            if epoch % self.checkpoint_every_n_epochs == 0:
                self.save_checkpoint(step=step, modules_state=modules_state)

        self.save_checkpoint(modules_state=modules_state)
