import jax
import tqdm
import equinox as eqx

from rlab_jax_projects.latent_actions.experiments import LatentActionModelBasedDreamerExperiment
from rlab_jax.utils.keygen import keygen
from rlab_jax_projects.latent_actions.offline_datasets import get_offline_dataset
from rlab_jax.rlab_dataclasses import random_variable
from rlab_jax.rlab_dataclasses import ModulesState


class OfflineLatentActionModelBasedDreamerExperiment(LatentActionModelBasedDreamerExperiment):
    _EXPERIMENT_TYPE = "OfflineLatentActionModelBasedDreamerExperiment"

    def __init__(self, model_training_epochs, seed, dataset_path=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_training_epochs = model_training_epochs
        self.seed = seed
        self.dataset_path = dataset_path

    def initial_state(self):
        modules_state = ModulesState(agent_state=self.agent.initial_state(),
                                     model_state=self.sequence_model.initial_state())

        if self.checkpoint_path is not None:
            modules_state = self.load_checkpoint(path=self.checkpoint_path, modules_state=modules_state)

        return modules_state

    def eval(self, key=keygen()):
        modules_state = self.initial_state()

        # VAL ROLLOUT
        batch = self.val_rollout.generate(
            model_state=modules_state.model_state,
            agent_state=modules_state.agent_state,
            steps_per_rollout=self.env_steps_per_rollout,
            key=key,
        )
        return batch.reward.sum(0)

    @eqx.filter_jit
    def policy_training_epoch(
        self, memory_state, model_state, agent_state, num_updates_per_epoch, key
    ):
        print("policy training compiling")

        agent_params, agent_static = eqx.partition(agent_state, eqx.is_array)

        def update(carry, _):
            agent_params_, key_ = carry

            keys_ = jax.random.split(key_, 5)

            agent_state_ = eqx.combine(agent_params_, agent_static)

            # Sample batch
            batch_ = self.train_rollout.memory.sample(
                memory_state=memory_state, batch_size=self.batch_size, key=keys_[0]
            )

            # Predictions
            predictions_ = self.sequence_model(
                model=model_state.module,
                observation_trajectory=batch_.observation_trajectory,
                action_trajectory=batch_.action_trajectory,
                key=keys_[1],
            )

            # Agent training step
            agent_state_, agent_log_ = self.agent.policy_training_step(
                model_state=model_state,
                agent_state=agent_state_,
                batch=batch_,
                predictions=predictions_,
                key=keys_[3],
            )

            agent_params_, _ = eqx.partition(agent_state_, eqx.is_array)

            return (agent_params_, keys_[4]), {**agent_log_}

        init_carry = (agent_params, key)
        (agent_params, key), train_logs = jax.lax.scan(
            f=update, init=init_carry, xs=None, length=num_updates_per_epoch
        )

        agent_state = eqx.combine(agent_params, agent_static)

        train_logs = jax.tree_map(lambda v: v.mean(), train_logs)

        return agent_state, train_logs

    @eqx.filter_jit
    def critic_training_epoch(
        self, memory_state, model_state, agent_state, num_updates_per_epoch, key
    ):
        print("critic training compiling")

        agent_params, agent_static = eqx.partition(agent_state, eqx.is_array)

        def update(carry, _):
            agent_params_, key_ = carry

            keys_ = jax.random.split(key_, 5)

            agent_state_ = eqx.combine(agent_params_, agent_static)

            # Sample batch
            batch_ = self.train_rollout.memory.sample(
                memory_state=memory_state, batch_size=self.batch_size, key=keys_[0]
            )

            # Predictions
            predictions_ = self.sequence_model(
                model=model_state.module,
                observation_trajectory=batch_.observation_trajectory,
                action_trajectory=batch_.action_trajectory,
                key=keys_[1],
            )

            # Agent training step
            agent_state_, agent_log_ = self.agent.critic_on_predictions_training_step(
                model_state=model_state,
                agent_state=agent_state_,
                batch=batch_,
                predictions=predictions_,
                key=keys_[3],
            )

            agent_params_, _ = eqx.partition(agent_state_, eqx.is_array)

            return (agent_params_, keys_[4]), {**agent_log_}

        init_carry = (agent_params, key)
        (agent_params, key), train_logs = jax.lax.scan(
            f=update, init=init_carry, xs=None, length=num_updates_per_epoch
        )

        agent_state = eqx.combine(agent_params, agent_static)

        train_logs = jax.tree_map(lambda v: v.mean(), train_logs)

        return agent_state, train_logs

    @eqx.filter_jit
    def sequence_model_validation_epoch(
            self, memory_state, model_state, key
    ):
        print("sequence model evaluation compiling")

        keys = jax.random.split(key, 3)

        # Sample batch
        batch = self.train_rollout.memory.sample(
            memory_state=memory_state, batch_size=self.validation_batch_size, key=keys[0], validation=True
        )

        # Model training step
        model_state, predictions, model_log = self.sequence_model.training_step(
            model_state=model_state, batch=batch, key=keys[1]
        )

        done_pred = model_state.module.done(
            next_state=predictions.posterior.item(jax.numpy.array(range(1, predictions.posterior.sample.shape[0]))),
            action=batch.action,
            key=keys[2],
            model_state=model_state)

        return model_log, batch, done_pred

    def train(self):
        modules_state = self.initial_state()

        self.train_rollout.memory, memory_state = get_offline_dataset(env=self.train_env, memory=self.train_rollout.memory, writer=self.writer, validation_ratio=self.validation_ratio, prng_key=keygen(), dataset_path=self.dataset_path)
        modules_state.model_state = self.sequence_model.scaler.fit(data=memory_state.data, model_state=modules_state.model_state)

        key, subkey = jax.random.split(keygen())

        for epoch in tqdm.tqdm(range(1, self.epochs + 1), desc="epochs"):
            step = (
                self.replay_memory_init
                + epoch * self.train_rollouts_per_epoch * self.env_steps_per_rollout
            )

            if epoch < self.model_training_epochs + 1:

                # SEQUENCE MODEL & REWARD TRAINING
                num_updates_per_epoch = int(
                    self.train_rollouts_per_epoch
                    * self.env_steps_per_rollout
                    * self.model_train_steps_ratio
                )
                key, subkey = jax.random.split(key)
                modules_state.model_state, train_logs = self.sequence_model_training_epoch(
                    memory_state=memory_state,
                    model_state=modules_state.model_state,
                    num_updates_per_epoch=num_updates_per_epoch,
                    key=subkey,
                )
                for k, v in train_logs.items():
                    self.writer.add_scalar(k, v.item(), step)

                # SEQUENCE MODEL VALIDATION
                val_logs, batch, done_pred = self.sequence_model_validation_epoch(
                    memory_state=memory_state,
                    model_state=modules_state.model_state,
                    key=subkey,
                )

                for k, v in val_logs.items():
                    self.writer.add_scalar("validation/" + k, v.item(), step)

            else:

                # AGENT TRAINING
                num_updates_per_epoch = int(
                    self.train_rollouts_per_epoch
                    * self.env_steps_per_rollout
                    * self.agent_train_steps_ratio
                )
                key, subkey = jax.random.split(key)

                modules_state.agent_state, train_logs = self.agent_training_epoch(
                    memory_state=memory_state,
                    agent_state=modules_state.agent_state,
                    model_state=modules_state.model_state,
                    num_updates_per_epoch=num_updates_per_epoch,
                    key=subkey,
                )

                for k, v in train_logs.items():
                    self.writer.add_scalar(k, v.item(), step)

            # VALIDATION
            if epoch % self.val_every_n_epochs == 0:

                # VAL ROLLOUT
                key, subkey = jax.random.split(key)
                batch = self.val_rollout.generate(
                    model_state=modules_state.model_state,
                    agent_state=modules_state.agent_state,
                    steps_per_rollout=self.env_steps_per_rollout,
                    key=subkey,
                )

                if self.evaluation is not None:
                    key, subkey = jax.random.split(key)
                    self.evaluation(
                        batch=batch,
                        model_state=modules_state.model_state,
                        agent_state=modules_state.agent_state,
                        key=subkey,
                        step=step,
                        epoch=epoch
                    )

                # PRIOR VAL ROLLOUT
                batch = self.prior_val_rollout.generate(
                    model_state=modules_state.model_state,
                    agent_state=modules_state.agent_state,
                    steps_per_rollout=self.env_steps_per_rollout,
                    key=keygen(),
                )

                reward = batch.reward.sum(0).mean()
                self.writer.add_scalar("reward/prior_real_world", reward.item(), step)

            if epoch % self.checkpoint_every_n_epochs == 0:
                self.save_checkpoint(step=step, modules_state=modules_state)

        self.save_checkpoint(modules_state=modules_state.model_state)

    def prior_act_fn_sample(self, agent_state, model_state, observation, latent_state, key):
        latent_action_dist = model_state.module.latent_action_model.prior(latent_state.sample)
        latent_action = latent_action_dist.mean()

        # reconstruct action
        action_reconstructed = self.agent.decode_fn(
            model=model_state.module.latent_action_model,
            latent_action=latent_action,
            state=latent_state.mean,
            deterministic=True,
            key=key,
        )

        return action_reconstructed, random_variable(sample=latent_action, **latent_action_dist.params)
