import jax
import tqdm

from rlab_jax.experiments.mb_dreamer import ModelBasedDreamerExperiment
from rlab_jax.utils.keygen import keygen
from rlab_jax_projects.latent_actions.offline_datasets import get_offline_dataset


class OfflineModelBasedDreamerExperiment(ModelBasedDreamerExperiment):
    _EXPERIMENT_TYPE = "OfflineModelBasedDreamerExperiment"

    def __init__(self, model_training_epochs, dataset_path=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_training_epochs = model_training_epochs
        self.dataset_path = dataset_path

    def train(self):
        modules_state = self.initial_state()

        self.train_rollout.memory, memory_state = get_offline_dataset(env=self.train_env, memory=self.train_rollout.memory, writer=self.writer, validation_ratio=self.validation_ratio, prng_key=keygen(), dataset_path=self.dataset_path)

        key, subkey = jax.random.split(keygen())

        for epoch in tqdm.tqdm(range(1, self.epochs + 1), desc="epochs"):
            step = (self.replay_memory_init + epoch * self.train_rollouts_per_epoch * self.env_steps_per_rollout)

            if epoch < self.model_training_epochs + 1:

                # SEQUENCE MODEL & REWARD TRAINING
                num_updates_per_epoch = int(
                    self.train_rollouts_per_epoch
                    * self.env_steps_per_rollout
                    * self.model_train_steps_ratio
                )
                key, subkey = jax.random.split(key)
                modules_state.model_state, train_logs = self.sequence_model_training_epoch(
                    memory_state=memory_state,
                    model_state=modules_state.model_state,
                    num_updates_per_epoch=num_updates_per_epoch,
                    key=subkey,
                )
                for k, v in train_logs.items():
                    self.writer.add_scalar(k, v.item(), step)

            # AGENT TRAINING
            else:
                num_updates_per_epoch = int(
                    self.train_rollouts_per_epoch
                    * self.env_steps_per_rollout
                    * self.agent_train_steps_ratio
                )
                key, subkey = jax.random.split(key)
                modules_state.agent_state, train_logs = self.agent_training_epoch(
                    memory_state=memory_state,
                    agent_state=modules_state.agent_state,
                    model_state=modules_state.model_state,
                    num_updates_per_epoch=num_updates_per_epoch,
                    key=subkey,
                )

                for k, v in train_logs.items():
                    self.writer.add_scalar(k, v.item(), step)

            # VALIDATION
            if epoch % self.val_every_n_epochs == 0:

                # VAL ROLLOUT
                key, subkey = jax.random.split(key)
                batch = self.val_rollout.generate(
                    model_state=modules_state.model_state,
                    agent_state=modules_state.agent_state,
                    steps_per_rollout=self.env_steps_per_rollout,
                    key=subkey,
                )

                if self.evaluation is not None:
                    key, subkey = jax.random.split(key)
                    self.evaluation(
                        batch=batch,
                        model_state=modules_state.model_state,
                        agent_state=modules_state.agent_state,
                        key=subkey,
                        step=step,
                        epoch=epoch
                    )

            if epoch % self.checkpoint_every_n_epochs == 0:
                self.save_checkpoint(step=step, modules_state=modules_state)

        self.save_checkpoint(modules_state=modules_state)