from rlab_jax.modules.agents import MBACAgent
import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf
from rlab_jax.modules.networks import Mlp, Module
from rlab_jax.utils.keygen import keygen


class DisagreementMBACAgent(MBACAgent):
    _AGENT_TYPE = "DisagreementMBACAgent"

    def __init__(self, disagreement_type, disagreement_scaling, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.disagreement_type = disagreement_type
        self.disagreement_scaling = disagreement_scaling

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, n_inputs, n_actions, **_ignored):
        @eqx.filter_vmap
        def critic_networks(key):
            return Mlp(
                n_inputs=n_inputs,
                n_outputs=1,
                key=key,
                activation=None,
                layers=cfg.critic.layers,
                units=cfg.critic.units,
                hidden_activation=cfg.critic.hidden_activation,
            )

        return cls(
            policy=Module.create(
                cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_actions,
                key=keygen(),
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
                init_std=cfg.policy.init_std,
            ),
            critic=critic_networks(jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))),
            target_critic=critic_networks(jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))),
            policy_optim_type=cfg.policy.optim.type,
            critic_optim_type=cfg.critic.optim.type,
            policy_optim_settings=cfg.policy.optim.settings,
            critic_optim_settings=cfg.critic.optim.settings,
            reward_weighting=cfg.reward_weighting,
            return_estimator_cfg=cfg.return_estimator,
            target_critic_lr=cfg.critic.target_lr,
            entropy_scale=cfg.entropy_scale,
            policy_max_gradient_norm=cfg.policy.optim.max_gradient_norm,
            critic_max_gradient_norm=cfg.critic.optim.max_gradient_norm,
            disagreement_type=cfg.disagreement_type,
            disagreement_scaling=cfg.disagreement_scaling
        )

    def dream_generate_fn_wrapped(self, model_state, *args, **kwargs):
        dream_rollout = self.dream_generate_fn(model_state=model_state, *args, **kwargs)

        @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
        def vmap_ensemble(e, x):
            return e(x)

        stochastic_dist = vmap_ensemble(model_state.module.transition.deterministic_to_stochastic, dream_rollout.state.deterministic)

        if self.disagreement_type == "mean_disagreement":
            ensemble_mean = jnp.expand_dims(stochastic_dist.mean().mean(0), 0)
            disagreement = jnp.mean(jnp.linalg.norm((stochastic_dist.mean() - ensemble_mean), axis=-1), 0)
            disagreement = jnp.expand_dims(disagreement, -1)
        elif self.disagreement_type == "std_logprob":
            logp = stochastic_dist.dist().log_prob(dream_rollout.state.stochastic.sample)
            disagreement = jnp.std(logp, axis=0)
            disagreement = jnp.expand_dims(disagreement, -1)
        else:
            raise ValueError

        dream_rollout.reward -= disagreement * self.disagreement_scaling

        return dream_rollout, {"reward/disagreement_dream": disagreement.sum(0).mean()}