import chex
import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf

from rlab_jax.modules.agents.ddpg import (DDPGAgent, DDPGAgentState, Mlp,
                                          Module, Transition, keygen)


class LatentActionDDPGAgent(DDPGAgent):
    _AGENT_TYPE = "LatentActionDDPGAgent"

    def __init__(self, decode_fn, max_latent_action, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.decode_fn = decode_fn
        self.max_latent_action = max_latent_action

    @classmethod
    def from_cfg(
        cls, cfg: omegaconf.DictConfig, n_inputs, n_actions, n_latent_actions, decode_fn, **_ignored
    ):
        @eqx.filter_vmap
        def q_networks(key):
            return Mlp(
                n_inputs=n_inputs + n_actions,
                n_outputs=1,
                key=key,
                activation=None,
                layers=cfg.critic.layers,
                units=cfg.critic.units,
                hidden_activation=cfg.critic.hidden_activation,
            )

        return cls(
            policy=Module.create(
                type=cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_latent_actions,
                key=keygen(),
                activation=jax.nn.tanh,
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
            ),
            target_policy=Module.create(
                type=cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_latent_actions,
                key=keygen(),
                activation=jax.nn.tanh,
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
            ),
            critic=q_networks(
                jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))
            ),
            target_critic=q_networks(
                jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))
            ),
            discount_factor=cfg.reward_discount_factor,
            policy_optim_type=cfg.policy.optim.type,
            critic_optim_type=cfg.critic.optim.type,
            policy_optim_settings=cfg.policy.optim.settings,
            critic_optim_settings=cfg.critic.optim.settings,
            target_policy_lr=cfg.policy.target_lr,
            target_critic_lr=cfg.critic.target_lr,
            max_latent_action=cfg.max_latent_action,
            decode_fn=decode_fn,
        )

    @staticmethod
    def compute_critic_values(critic: Module, states: jnp.array, actions: jnp.array):
        inputs = jnp.concatenate((states, actions), -1)

        @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
        def evaluate_ensemble(model, x):
            return model(x)

        return evaluate_ensemble(critic, inputs)

    @eqx.filter_jit
    def act(
        self,
        agent_state: chex.dataclass,
        model_state: chex.dataclass,
        observation,
        key: jax.random.PRNGKey,
        deterministic=False,
        **ignored
    ):
        key, subkey = jax.random.split(key)
        latent_action, _ = super().act(agent_state=agent_state, observation=observation, key=key, deterministic=deterministic)

        latent_action.sample *= self.max_latent_action

        # reconstruct action
        action_reconstructed = self.decode_fn(
            model=model_state.module,
            latent_action=latent_action.sample,
            state=observation,
            deterministic=deterministic,
            key=subkey,
        )

        return action_reconstructed, latent_action

    def policy_loss(
        self,
        policy: Module,
        critic: Module,
        decoder: Module,
        batch: Transition,
        key: jax.random.PRNGKey,
    ):
        latent_action = policy(batch.observation)
        latent_action *= self.max_latent_action

        action = self.decode_fn(
            model=None,
            decoder=decoder,
            latent_action=latent_action,
            state=batch.observation,
            deterministic=True,
            key=key,
        )

        q_estimates = jnp.min(self.compute_critic_values(critic, batch.observation, action.sample), axis=0)

        loss = -q_estimates
        loss = loss.mean()

        return loss, {"agent/policy/loss": loss}

    def critic_loss(
        self,
        critic: Module,
        target_critic: Module,
        target_policy: Module,
        decoder: Module,
        batch: Transition,
        key: jax.random.PRNGKey,
    ):
        target_latent_action = target_policy(batch.next_observation)
        target_latent_action *= self.max_latent_action

        target_action = self.decode_fn(
            model=None,
            decoder=decoder,
            latent_action=target_latent_action,
            state=batch.next_observation,
            deterministic=True,
            key=key,
        )
        target_q = jnp.min(self.compute_critic_values(target_critic, batch.next_observation, target_action.sample), axis=0)

        target = batch.reward + jnp.logical_not(batch.done) * self.discount_factor * target_q

        q_estimates = self.compute_critic_values(critic, batch.observation, batch.action)

        loss = (q_estimates - target) ** 2
        loss = loss.mean()

        return loss, {"agent/q/loss": loss}

    def training_step(
        self, agent_state: DDPGAgentState, model_state, batch: Transition, key: jax.random.PRNGKey
    ):
        keys = jax.random.split(key, 2)

        (_, policy_log), policy_grads = eqx.filter_value_and_grad(self.policy_loss, has_aux=True)(
            agent_state.policy.module,
            critic=agent_state.critic.module,
            decoder=model_state.module.decoder,
            batch=batch,
            key=keys[0],
        )
        (_, critic_log), critic_grads = eqx.filter_value_and_grad(self.critic_loss, has_aux=True)(
            agent_state.critic.module,
            target_critic=agent_state.target_critic.module,
            target_policy=agent_state.target_policy.module,
            decoder=model_state.module.decoder,
            batch=batch,
            key=keys[1],
        )

        agent_state.policy = agent_state.policy.apply_updates(policy_grads)
        agent_state.critic = agent_state.critic.apply_updates(critic_grads)

        log = {}
        log.update(policy_log)
        log.update(critic_log)

        agent_state.target_critic.module = self.soft_update(
            agent_state.critic.module, agent_state.target_critic.module, self.target_critic_lr
        )
        # TODO: Only works for MLP policies
        agent_state.target_policy.module.mlp = self.soft_update(
            agent_state.policy.module.mlp,
            agent_state.target_policy.module.mlp,
            self.target_policy_lr,
        )

        return agent_state, log
