import equinox as eqx
import jax.numpy as jnp
import jax.random
import omegaconf

from rlab_jax.modules.agents.mbac import (MBACAgent, MBACAgentState, Mlp,
                                          Module, keygen)
from rlab_jax.modules.networks import TanhGaussMlp
from rlab_jax.distributions import multivariate_normal_diag
from rlab_jax.rlab_dataclasses.transitions import DreamRolloutTransition


class TanhGaussMlpLatentActionStdScaling(TanhGaussMlp):
    _MODULE_TYPE = "TanhGaussMlpLatentActionStdScaling"
    _dist_type = "tanh_normal"
    mean_scale: int
    max_std: int
    latent_action_std_scaling: jnp.array

    def __init__(self, mean_scale=5, max_std=jnp.inf, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mean_scale = mean_scale
        self.max_std = max_std
        self.latent_action_std_scaling = jnp.ones((1,))


class LatentActionMBACAgent(MBACAgent):
    _AGENT_TYPE = "LatentActionMBACAgent"

    def __init__(
        self, decode_fn, policy_reward_weighting, critic_reward_weighting, latent_action_std_scaling, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.decode_fn = decode_fn
        self.policy_reward_weighting = policy_reward_weighting
        self.critic_reward_weighting = critic_reward_weighting
        self.latent_action_std_scaling = latent_action_std_scaling

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, n_inputs, n_actions, decode_fn, **_ignored):
        @eqx.filter_vmap
        def critic_networks(key):
            return Mlp(
                **cfg.critic,
                activation=None,
                n_inputs=n_inputs,
                n_outputs=1,
                key=key,
            )

        return cls(
            policy=Module.create(
                **cfg.policy,
                n_inputs=n_inputs,
                n_outputs=n_actions,
                key=keygen(),
            ),
            critic=critic_networks(jax.random.split(keygen(), cfg.critic.get("n_networks", 2))),
            target_critic=critic_networks(jax.random.split(keygen(), cfg.critic.get("n_networks", 2))),
            policy_optim_type=cfg.policy.optim.type,
            critic_optim_type=cfg.critic.optim.type,
            policy_optim_settings=cfg.policy.optim.settings,
            critic_optim_settings=cfg.critic.optim.settings,
            policy_reward_weighting=cfg.policy_reward_weighting,
            critic_reward_weighting=cfg.critic_reward_weighting,
            return_estimator_cfg=cfg.return_estimator,
            target_critic_lr=cfg.critic.target_lr,
            latent_action_std_scaling=cfg.latent_action_std_scaling,
            decode_fn=decode_fn,
            policy_max_gradient_norm=cfg.policy.optim.max_gradient_norm,
            critic_max_gradient_norm=cfg.critic.optim.max_gradient_norm,
            entropy_scale=cfg.entropy_scale,
        )

    def scale_latent_action(self, latent_action, latent_action_prior_dist, latent_action_std_scaling):

        if latent_action_std_scaling is None:
            return latent_action

        else:
            latent_action_min = latent_action_prior_dist.mean() - latent_action_std_scaling * latent_action_prior_dist.std()
            latent_action_max = latent_action_prior_dist.mean() + latent_action_std_scaling * latent_action_prior_dist.std()

            return (latent_action + 1) / 2 * (latent_action_max - latent_action_min) + latent_action_min

    @eqx.filter_jit
    def evaluate(
        self,
        latent_state: jnp.array,
        model_state,
        key: jax.random.PRNGKey,
        agent_state: MBACAgentState = None,
        policy: Module = None,
        deterministic=False,
        **ignored
    ):
        key, subkey = jax.random.split(key, 2)
        latent_action, _ = super().act(latent_state=latent_state, key=key, agent_state=agent_state,
                                       policy=policy, deterministic=True, exploration=False)

        latent_action_prior_dist = model_state.module.latent_action_model.prior(latent_state.sample)

        if hasattr(policy, "latent_action_std_scaling"):
            latent_action_std_scaling = policy.latent_action_std_scaling
        else:
            latent_action_std_scaling = self.latent_action_std_scaling

        latent_action_sample = self.scale_latent_action(latent_action=latent_action.sample,
                                                        latent_action_prior_dist=latent_action_prior_dist,
                                                        latent_action_std_scaling=latent_action_std_scaling)

        action = self.decode_fn(
            model=model_state.module.latent_action_model,
            latent_action=latent_action_sample,
            state=latent_state.sample,
            deterministic=True,
            key=subkey,
        )

        return action, latent_action

    @eqx.filter_jit
    def act(
        self,
        latent_state: jnp.array,
        model_state,
        key: jax.random.PRNGKey,
        agent_state: MBACAgentState = None,
        policy: Module = None,
        deterministic=False,
        **ignored
    ):
        key, subkey = jax.random.split(key, 2)
        latent_action, _ = super().act(latent_state=latent_state, key=key, agent_state=agent_state,
                                       policy=policy, deterministic=deterministic, exploration=False)

        latent_action_prior_dist = model_state.module.latent_action_model.prior(latent_state.sample)

        if hasattr(policy, "latent_action_std_scaling"):
            latent_action_std_scaling = policy.latent_action_std_scaling
        else:
            latent_action_std_scaling = self.latent_action_std_scaling

        latent_action_sample = self.scale_latent_action(latent_action=latent_action.sample,
                                                        latent_action_prior_dist=latent_action_prior_dist,
                                                        latent_action_std_scaling=latent_action_std_scaling)

        action = self.decode_fn(
            model=model_state.module.latent_action_model,
            latent_action=latent_action_sample,
            state=latent_state.sample,
            deterministic=deterministic,
            key=subkey,
        )

        return action, latent_action

    def dream_act(
        self,
        latent_state: jnp.array,
        model_state,
        key: jax.random.PRNGKey,
        agent_state: MBACAgentState = None,
        policy: Module = None,
        deterministic=False,
        **ignored
    ):
        key, subkey = jax.random.split(key, 2)
        latent_action, _ = super().act(latent_state=latent_state, key=key, agent_state=agent_state,
                                       policy=policy, deterministic=deterministic, exploration=False)

        latent_action_prior_dist = model_state.module.latent_action_model.prior(latent_state.sample)

        if hasattr(policy, "latent_action_std_scaling"):
            latent_action_std_scaling = policy.latent_action_std_scaling
        else:
            latent_action_std_scaling = self.latent_action_std_scaling

        latent_action_sample = self.scale_latent_action(latent_action=latent_action.sample,
                                                        latent_action_prior_dist=latent_action_prior_dist,
                                                        latent_action_std_scaling=latent_action_std_scaling)

        action = self.decode_fn(
            model=model_state.module.latent_action_model,
            latent_action=latent_action_sample,
            state=latent_state.sample,
            deterministic=True,
            key=subkey,
        )

        return action, latent_action

    def penalty(self, weighting, dream_rollout, model_state=None, policy=None):
        return weighting * dream_rollout.latent_action.dist.log_prob(dream_rollout.latent_action.sample)

    def entropy_penalty(self, dream_rollout):
        return self.entropy_scale * dream_rollout.latent_action.dist.entropy()

    def training_step(
        self, model_state, agent_state, batch, predictions, key: jax.random.PRNGKey
    ):
        (_, (policy_log, dream_rollout)), policy_grads = eqx.filter_value_and_grad(
            self.policy_loss, has_aux=True
        )(
            agent_state.policy.module,
            critic=agent_state.critic.module,
            model_state=model_state,
            batch=batch,
            predictions=predictions,
            key=key,
        )

        (_, critic_log), critic_grads = eqx.filter_value_and_grad(self.critic_loss, has_aux=True)(
            agent_state.critic.module,
            target_critic=agent_state.target_critic.module,
            dream_rollout=dream_rollout,
        )

        agent_state.policy = agent_state.policy.apply_updates(policy_grads)
        agent_state.critic = agent_state.critic.apply_updates(critic_grads)

        agent_state.target_critic.module = self.soft_update(
            agent_state.critic.module, agent_state.target_critic.module, self.target_critic_lr
        )

        log = {}
        log.update(policy_log)
        log.update(critic_log)
        if hasattr(agent_state.policy.module, "latent_action_std_scaling"):
            log.update({"agent/latent_action_std_scaling": agent_state.policy.module.latent_action_std_scaling})

        return agent_state, log

    def critic_on_predictions_training_step(
        self, model_state, agent_state, batch, predictions, key: jax.random.PRNGKey
    ):
        dream_rollout = DreamRolloutTransition(
            state=predictions.posterior.item(jnp.arange(1, predictions.posterior.state.sample.shape[0] - 1)),
            next_state=predictions.posterior.item(jnp.arange(2, predictions.posterior.state.sample.shape[0])),
            action=batch.action[1:],
            reward=batch.reward[1:],
            done=batch.done[1:],
            was_done=batch.done[:-1],
        )

        self.critic_reward_weighting = 0

        (_, critic_log), critic_grads = eqx.filter_value_and_grad(self.critic_loss, has_aux=True)(
            agent_state.critic.module,
            target_critic=agent_state.target_critic.module,
            dream_rollout=dream_rollout,
        )

        agent_state.critic = agent_state.critic.apply_updates(critic_grads)

        agent_state.target_critic.module = self.soft_update(
            agent_state.critic.module, agent_state.target_critic.module, self.target_critic_lr
        )

        log = {}
        log.update(critic_log)
        if hasattr(agent_state.policy.module, "latent_action_std_scaling"):
            log.update({"agent/latent_action_std_scaling": agent_state.policy.module.latent_action_std_scaling})

        return agent_state, log

    def policy_training_step(
        self, model_state, agent_state, batch, predictions, key: jax.random.PRNGKey
    ):
        (_, (policy_log, dream_rollout)), policy_grads = eqx.filter_value_and_grad(
            self.policy_loss, has_aux=True
        )(
            agent_state.policy.module,
            critic=agent_state.critic.module,
            model_state=model_state,
            batch=batch,
            predictions=predictions,
            key=key,
        )

        agent_state.policy = agent_state.policy.apply_updates(policy_grads)

        log = {}
        log.update(policy_log)
        if hasattr(agent_state.policy.module, "latent_action_std_scaling"):
            log.update({"agent/latent_action_std_scaling": agent_state.policy.module.latent_action_std_scaling})

        return agent_state, log