from rlab_jax.modules.agents import MBACAgent
import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf
from rlab_jax.modules.networks import Mlp, Module
from rlab_jax.utils.keygen import keygen
from rlab_jax_projects.latent_actions.modules.agents import DisagreementMBACAgent
from rlab_jax.rlab_dataclasses.transitions import DreamRolloutTransition


class LOMPOAgent(DisagreementMBACAgent):
    _AGENT_TYPE = "LOMPOAgent"

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, n_inputs, n_actions, **_ignored):
        @eqx.filter_vmap
        def critic_networks(key):
            return Mlp(
                n_inputs=n_inputs + n_actions,
                n_outputs=1,
                key=key,
                activation=None,
                layers=cfg.critic.layers,
                units=cfg.critic.units,
                hidden_activation=cfg.critic.hidden_activation,
            )

        return cls(
            policy=Module.create(
                cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_actions,
                key=keygen(),
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
                init_std=cfg.policy.init_std,
            ),
            critic=critic_networks(jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))),
            target_critic=critic_networks(jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))),
            policy_optim_type=cfg.policy.optim.type,
            critic_optim_type=cfg.critic.optim.type,
            policy_optim_settings=cfg.policy.optim.settings,
            critic_optim_settings=cfg.critic.optim.settings,
            reward_weighting=cfg.reward_weighting,
            return_estimator_cfg=cfg.return_estimator,
            target_critic_lr=cfg.critic.target_lr,
            entropy_scale=cfg.entropy_scale,
            policy_max_gradient_norm=cfg.policy.optim.max_gradient_norm,
            critic_max_gradient_norm=cfg.critic.optim.max_gradient_norm,
            disagreement_type=cfg.disagreement_type,
            disagreement_scaling=cfg.disagreement_scaling
        )

    @staticmethod
    def compute_critic_values(critic: Module, states: jnp.array, actions: jnp.array):
        inputs = jnp.concatenate((states, actions), -1)

        @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
        def evaluate_ensemble(model, x):
            return model(x)

        return evaluate_ensemble(critic, inputs)

    def mixed_batch(self, batch, predictions, model_state, agent_state, key):
        batch_size = batch.observation.shape[1]

        dream_rollout, dream_rollout_log = self.dream_generate_fn_wrapped(
            policy=agent_state.policy.module,
            model_state=model_state,
            batch=batch,
            predictions=predictions,
            key=key,
        )

        def sample_dream_transitions(key):
            key, subkey = jax.random.split(key)
            i = jax.random.randint(key, shape=(1,), minval=0, maxval=dream_rollout.reward.shape[0])
            j = jax.random.randint(key, shape=(1,), minval=0, maxval=dream_rollout.reward.shape[1])

            return DreamRolloutTransition(state=dream_rollout.state.sample[i, j].squeeze(0),
                                          next_state=dream_rollout.next_state.sample[i, j].squeeze(0),
                                          action=dream_rollout.action.sample[i, j].squeeze(0),
                                          reward=dream_rollout.reward[i, j].squeeze(0),
                                          done=dream_rollout.done[i, j].squeeze(0))

        dream_transition = jax.vmap(sample_dream_transitions)(jax.random.split(key, batch_size))

        def sample_pred_transitions(key):
            key, subkey = jax.random.split(key)
            i = jax.random.randint(key, shape=(1,), minval=0, maxval=batch.action.shape[0])
            j = jax.random.randint(key, shape=(1,), minval=0, maxval=batch.action.shape[1])
            return DreamRolloutTransition(state=predictions.posterior.sample[i, j].squeeze(0),
                    next_state=predictions.posterior.sample[i + 1, j].squeeze(0),
                    action=batch.action[i, j].squeeze(0),
                    reward=batch.reward[i, j].squeeze(0),
                    done=batch.done[i, j].squeeze(0))

        pred_transition = jax.vmap(sample_pred_transitions)(jax.random.split(key, batch_size))

        return jax.tree_map(lambda x,y: jnp.concatenate((x,y)), pred_transition, dream_transition)

    def critic_loss(
        self, critic: Module, target_critic: Module, policy: Module, batch: DreamRolloutTransition, key: jax.random.PRNGKey
    ):
        next_action_dist = policy(batch.next_state)
        next_action, next_log_prob = next_action_dist.sample_and_log_prob(seed=key)

        next_q_estimates = jnp.min(
            self.compute_critic_values(target_critic, batch.next_state, next_action), axis=0
        )

        target = batch.reward + self.return_estimator_cfg["discount_factor"] * next_q_estimates

        q_estimates = self.compute_critic_values(critic, batch.state, batch.action)

        loss = (q_estimates - target) ** 2
        loss = loss.mean()

        return loss, {"agent/q/loss": loss}

    def policy_loss(
        self,
        policy: Module,
        critic: Module,
        model_state,
        batch,
        key: jax.random.PRNGKey,
    ):
        action_dist = policy(batch.state)
        action, log_prob = action_dist.sample_and_log_prob(seed=key)

        loss = -jnp.min(self.compute_critic_values(critic, batch.next_state, action), axis=0)
        loss = loss.mean()

        return loss, {"agent/loss": loss}

    def training_step(
        self, model_state, agent_state, batch, predictions, key: jax.random.PRNGKey
    ):
        key, subkey = jax.random.split(key)
        mixed_batch = jax.lax.stop_gradient(self.mixed_batch(batch, predictions, model_state, agent_state, subkey))

        (_, (policy_log)), policy_grads = eqx.filter_value_and_grad(
            self.policy_loss, has_aux=True
        )(
            agent_state.policy.module,
            critic=agent_state.critic.module,
            model_state=model_state,
            batch=mixed_batch,
            key=key,
        )

        (_, critic_log), critic_grads = eqx.filter_value_and_grad(self.critic_loss, has_aux=True)(
            agent_state.critic.module,
            target_critic=agent_state.target_critic.module,
            policy=agent_state.policy.module,
            batch=mixed_batch,
            key=subkey,
        )

        agent_state.policy = agent_state.policy.apply_updates(policy_grads)
        agent_state.critic = agent_state.critic.apply_updates(critic_grads)

        agent_state.target_critic.module = self.soft_update(
            agent_state.critic.module, agent_state.target_critic.module, self.target_critic_lr
        )

        log = {}
        log.update(policy_log)
        log.update(critic_log)

        return agent_state, log

    def dream_generate_fn_wrapped(self, model_state, *args, **kwargs):
        dream_rollout = self.dream_generate_fn(model_state=model_state, *args, **kwargs)

        @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
        def vmap_ensemble(e, x):
            return e(x)

        stochastic_dist = vmap_ensemble(model_state.module.transition.deterministic_to_stochastic, dream_rollout.state.deterministic)

        if self.disagreement_type == "mean_disagreement":
            ensemble_mean = jnp.expand_dims(stochastic_dist.mean().mean(0), 0)
            disagreement = jnp.mean(jnp.linalg.norm((stochastic_dist.mean() - ensemble_mean), axis=-1), 0)
            disagreement = jnp.expand_dims(disagreement, -1)
        elif self.disagreement_type == "std_logprob":
            logp = stochastic_dist.dist().log_prob(dream_rollout.state.stochastic.sample)
            disagreement = jnp.std(logp, axis=0)
            disagreement = jnp.expand_dims(disagreement, -1)
        else:
            raise ValueError

        dream_rollout.reward -= disagreement * self.disagreement_scaling

        return dream_rollout, {"reward/disagreement_dream": disagreement.sum(0).mean()}