from rlab_jax.modules.agents.sac import SACAgent, SACAgentState
import chex
import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf
import optax

from rlab_jax.modules.agents.agent import Agent
from rlab_jax.modules.networks import Mlp, Module
from rlab_jax.rlab_dataclasses import TrainState, Transition, random_variable
from rlab_jax.utils.keygen import keygen
from rlab_jax_projects.latent_actions.modules.mbpo_ensemble import ScalerTrainState


class MobileAgent(SACAgent):
    _AGENT_TYPE = "MobileAgent"

    def __init__(self, observation_predict_fn,
                 penalty_samples,
                 penalty_coef,
                 decay_steps: int,
                 policy_optim_type: str = "optax.adam",
                 policy_optim_settings: dict = {"learning_rate": 1e-4},
                 *args,
                 **kwargs):
        super().__init__(policy_optim_type=policy_optim_type, policy_optim_settings=policy_optim_settings, *args, **kwargs)
        self.observation_predict_fn = observation_predict_fn
        self.penalty_samples = penalty_samples
        self.penalty_coef = penalty_coef

        policy_optim_settings = dict(policy_optim_settings)
        if "lr" in policy_optim_settings.keys():
            policy_optim_settings["learning_rate"] = policy_optim_settings["lr"]
            del policy_optim_settings["lr"]

        self.lr_schedule = optax.cosine_decay_schedule(policy_optim_settings["learning_rate"] , decay_steps=decay_steps)
        policy_optim_settings["learning_rate"] = self.lr_schedule

        self.policy_optim = eval(policy_optim_type)(**policy_optim_settings)


    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, n_inputs, n_actions, observation_predict_fn, decay_steps, **_ignored):
        @eqx.filter_vmap
        def q_networks(key):
            return Mlp(
                n_inputs=n_inputs + n_actions,
                n_outputs=1,
                key=key,
                activation=None,
                layers=cfg.critic.layers,
                units=cfg.critic.units,
                hidden_activation=cfg.critic.hidden_activation,
            )

        return cls(
            policy=Module.create(
                type=cfg.policy.type,
                n_inputs=n_inputs,
                n_outputs=n_actions,
                key=keygen(),
                layers=cfg.policy.layers,
                units=cfg.policy.units,
                hidden_activation=cfg.policy.hidden_activation,
                init_std=cfg.policy.init_std,
            ),
            critic=q_networks(
                jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))
            ),
            target_critic=q_networks(
                jax.random.split(keygen(), cfg.critic.get("number_of_q_networks", 2))
            ),
            policy_optim_type=cfg.policy.optim.type,
            critic_optim_type=cfg.critic.optim.type,
            policy_optim_settings=cfg.policy.optim.settings,
            critic_optim_settings=cfg.critic.optim.settings,
            discount_factor=cfg.reward_discount_factor,
            entropy_coef=cfg.sac_entropy_coef,
            auto_entropy=cfg.sac_auto_entropy,
            entropy_optim_settings={"learning_rate": cfg.sac_alpha_lr},
            target_critic_lr=cfg.critic.target_lr,
            observation_predict_fn=observation_predict_fn,
            penalty_samples=cfg.penalty_samples,
            penalty_coef=cfg.penalty_coef,
            decay_steps=decay_steps
        )

    def penalty(self, policy: Module, target_critic: Module, model_state: ScalerTrainState, dream_batch: Transition, key: jax.random.PRNGKey):
        keys = jax.random.split(key, 2)
        next_observation, _, _ = self.observation_predict_fn(model_state=model_state, observation=dream_batch.observation, action=dream_batch.action, key=keys[0], sample_shape=(self.penalty_samples))

        next_action_dist = policy(next_observation)
        next_action = next_action_dist.sample(keys[0])

        next_q_estimates = jnp.min(self.compute_critic_values(target_critic, next_observation, next_action), axis=0)

        return next_q_estimates.mean(0).std(0)

    def critic_loss(
        self,
        critic: Module,
        target_critic: Module,
        policy: Module,
        entropy_coef: Module,
        model_state: ScalerTrainState,
        batch: Transition,
        dream_batch: Transition,
        data_batch: Transition,
        key: jax.random.PRNGKey,
    ):
        dream_penalty = self.penalty(policy=policy, target_critic=target_critic, model_state=model_state, dream_batch=dream_batch, key=key)
        penalty = jnp.zeros_like(batch.reward)
        penalty = penalty.at[:dream_penalty.shape[0]].set(dream_penalty)

        next_action_dist = policy(batch.next_observation)
        next_action, next_log_prob = next_action_dist.sample_and_log_prob(seed=key)

        next_q_estimates = jnp.min(
            self.compute_critic_values(target_critic, batch.next_observation, next_action), axis=0
        )

        target_q = next_q_estimates - entropy_coef() * jnp.expand_dims(next_log_prob, -1)

        target = batch.reward - penalty * self.penalty_coef + jnp.logical_not(batch.done) * self.discount_factor * target_q
        target = jnp.clip(target, a_min=0)

        q_estimates = self.compute_critic_values(critic, batch.observation, batch.action)

        loss = (q_estimates - target) ** 2
        loss = loss.mean()

        return loss, {"agent/q/loss": loss, "agent/q/penalty": penalty.mean()}

    def training_step(self, agent_state: SACAgentState, model_state: ScalerTrainState, data_batch: Transition, dream_batch: Transition, key: jax.random.PRNGKey):
        keys = jax.random.split(key, 3)

        batch = jax.tree_map(lambda x, y: jax.numpy.concatenate((x, y), axis=0), dream_batch, data_batch)

        (_, policy_log), policy_grads = eqx.filter_value_and_grad(self.policy_loss, has_aux=True)(
            agent_state.policy.module,
            critic=agent_state.critic.module,
            entropy_coef=agent_state.entropy_coef.module,
            batch=batch,
            key=keys[0],
        )
        (_, critic_log), critic_grads = eqx.filter_value_and_grad(self.critic_loss, has_aux=True)(
            agent_state.critic.module,
            target_critic=agent_state.target_critic.module,
            policy=agent_state.policy.module,
            entropy_coef=agent_state.entropy_coef.module,
            model_state=model_state,
            batch=batch,
            dream_batch=dream_batch,
            data_batch=data_batch,
            key=keys[1],
        )

        agent_state.policy = agent_state.policy.apply_updates(policy_grads)
        agent_state.critic = agent_state.critic.apply_updates(critic_grads)

        log = {}
        log.update(policy_log)
        log.update(critic_log)

        if self.auto_entropy:
            (_, entropy_coef_log), entropy_coef_grads = eqx.filter_value_and_grad(
                self.entropy_coef_loss, has_aux=True
            )(agent_state.entropy_coef.module, agent_state.policy.module, batch, keys[2])
            log.update(entropy_coef_log)

            agent_state.entropy_coef = agent_state.entropy_coef.apply_updates(entropy_coef_grads)

        agent_state.target_critic.module = self.soft_update(
            agent_state.critic.module, agent_state.target_critic.module, self.target_critic_lr
        )

        return agent_state, log