import chex
import equinox as eqx
import jax
import jax.numpy as jnp
import optax

from rlab_jax.modules import Module
from rlab_jax.rlab_dataclasses import (Prediction, TrainState, Transition,
                                       random_variable)
from rlab_jax.utils.keygen import keygen


@chex.dataclass
class LatentActionVAEModel:
    decoder: Module
    encoder: Module
    prior: Module
    dynamics: Module


class LatentActionVAE:
    _LATENT_ACTION_MODEL_TYPE = "LatentActionVAE"

    def __init__(
        self,
        encoder: Module,
        decoder: Module,
        prior: Module,
        n_actions,
        n_latent_actions,
        beta,
        temperature,
        annealing_method,
        decoder_state_dropout=None,
        optim_type: str = "optax.adam",
        optim_settings: dict = {"lr": 3e-4},
        latent_action_dynamics: Module = None,
        latent_action_dynamics_loss_scale=0,
        mode_if_deterministic=False,
        free_nats=None,
        kl_balance=None
    ):
        """Latent action VAE

        Parameters
        ----------
        encoder: Encoder
        decoder: Decoder
        n_actions: int
            Number of real world actions
        n_latent_actions: int
            Number of latent actions
        beta: float
            Scaling factor for lam KL. Defaults to 1, the normal ELBO.
        temperature: float
            Temperature for lam KL warm-up scheme. If set to greater than 1,
            the KL is initially scaled down by ``1 / temperature``.
            The factor is then increased linearly after every iteration
            by `1 / temperature`. Therefore, it takes `temperature`
            optimization steps until the warm-up is completed.
        """

        self.n_actions = n_actions
        self.n_latent_actions = n_latent_actions

        self.annealing_method = annealing_method

        self.beta = beta
        self.temperature = temperature
        self.latent_action_dynamics_loss_scale = latent_action_dynamics_loss_scale
        self.latent_action_dynamics = latent_action_dynamics
        self.mode_if_deterministic = mode_if_deterministic
        self.free_nats = free_nats
        self.kl_balance = kl_balance

        self.decoder_state_dropout = decoder_state_dropout

        optim_settings = dict(optim_settings)
        if "lr" in optim_settings.keys():
            optim_settings["learning_rate"] = optim_settings["lr"]
            del optim_settings["lr"]

        optim = eval(optim_type)(**optim_settings)

        model_state = LatentActionVAEModel(
            decoder=decoder, encoder=encoder, prior=prior, dynamics=latent_action_dynamics
        )

        self.train_state = TrainState(
            module=model_state,
            optim=optim,
            optim_state=optim.init(eqx.filter(model_state, eqx.is_array)),
        )

    def initial_state(self):
        return self.train_state

    @classmethod
    def from_cfg(cls, cfg, n_states, n_actions, **_ignored):
        encoder = Module.create(
            **cfg.encoder,
            n_inputs=n_states + n_actions,
            n_outputs=cfg.n_latent_actions,
            key=keygen(),
        )

        decoder = Module.create(
            **cfg.decoder,
            n_inputs=n_states + cfg.n_latent_actions,
            n_outputs=n_actions,
            key=keygen(),
        )

        prior = Module.create(
            **cfg.prior,
            n_inputs=n_states,
            n_outputs=cfg.n_latent_actions,
            key=keygen()
        )

        if (
            hasattr(cfg, "latent_action_dynamics_loss_scale")
            and cfg.latent_action_dynamics_loss_scale > 0
        ):
            latent_action_dynamics = Module.create(
                "GaussMlp",
                n_inputs=n_states + cfg.n_latent_actions,
                n_outputs=n_states,
                layers=3,
                units=512,
                hidden_activation=jax.nn.elu,
                init_std=1.0,
                key=keygen(),
            )
            latent_action_dynamics_loss_scale = cfg.latent_action_dynamics_loss_scale
        else:
            latent_action_dynamics = None
            latent_action_dynamics_loss_scale = 0

        return cls(
            encoder=encoder,
            decoder=decoder,
            prior=prior,
            latent_action_dynamics=latent_action_dynamics,
            latent_action_dynamics_loss_scale=latent_action_dynamics_loss_scale,
            n_actions=n_actions,
            n_latent_actions=cfg.n_latent_actions,
            beta=cfg.beta,
            temperature=cfg.temperature,
            annealing_method=cfg.annealing_method,
            optim_type=cfg.optim.type,
            optim_settings=cfg.optim.settings,
            mode_if_deterministic=cfg.mode_if_deterministic,
            free_nats=cfg.free_nats,
            kl_balance=cfg.kl_balance,
        )

    def __call__(
        self,
        model: LatentActionVAEModel,
        observation: jnp.array,
        action: jnp.array,
        key: jax.random.PRNGKey,
    ):
        keys = jax.random.split(key, 3)

        posterior = self._encode(model=model, action=action, state=observation, key=keys[0])
        prior = self._prior(model=model, state=observation, key=keys[1])
        reconstruction = self._decode(
            model=model, latent_action=posterior.sample, state=observation, key=keys[2]
        )

        return Prediction(posterior=posterior, prior=prior, reconstruction=reconstruction)

    def _prior(self, model: LatentActionVAEModel, state: jnp.array, key: jax.random.PRNGKey):
        dist = model.prior(state)
        sample = dist.sample(key)

        return random_variable(sample=sample, **dist.params)

    def _encode(
        self,
        model: LatentActionVAEModel,
        action: jnp.array,
        state: jnp.array,
        key: jax.random.PRNGKey,
    ):
        dist = model.encoder(jnp.concatenate((state, action), axis=-1))
        sample = dist.sample(key)

        return random_variable(sample=sample, **dist.params)

    def _decode(
        self,
        model: LatentActionVAEModel,
        latent_action: jnp.array,
        state: jnp.array,
        key: jax.random.PRNGKey,
        deterministic=False,
        decoder=None,
        dropout=False
    ):
        if dropout and self.decoder_state_dropout is not None:
            q = 1 - self.decoder_state_dropout
            mask = jax.random.bernoulli(key, q, state.shape)
            state = jnp.where(mask, state / q, 0)

        if decoder is not None:
            dist = decoder(jnp.concatenate((state, latent_action), axis=-1))
        else:
            dist = model.decoder(jnp.concatenate((state, latent_action), axis=-1))

        if deterministic:
            if self.mode_if_deterministic:
                sample = dist.mode()
            else:
                sample = dist.mean()
        else:
            sample = dist.sample(key)

        return random_variable(sample=sample, **dist.params)

    def loss(
        self,
        model: LatentActionVAEModel,
        action: jnp.array,
        observation: jnp.array,
        next_observation: jnp.array,
        key: jax.random.PRNGKey,
    ):
        predictions = self.__call__(model=model, action=action, observation=observation, key=key)

        # Action NLL & KL
        nll_sample_wise_action = -predictions.reconstruction.dist.log_prob(action)
        kl_sample_wise_action = predictions.posterior.dist.kl_divergence(predictions.prior.dist)

        nll = jnp.mean(nll_sample_wise_action)
        kl = jnp.mean(kl_sample_wise_action) * self.beta
        loss = nll + kl

        log = {
            "sequence_model/nll/v": nll,
            "sequence_model/kl/v": kl,
            "latent_action_model/nll": nll,
            "latent_action_model/kl": kl,
            # "latent_action_model/warmup_kl": self.annealing,
            "latent_action_model/loss": loss,
        }

        if self.latent_action_dynamics is not None:
            next_observation_prediction = model.dynamics(
                jnp.concatenate((observation, predictions.posterior.sample), axis=-1)
            )
            latent_action_dynamics_loss = (
                -next_observation_prediction.dist().log_prob(next_observation).mean()
                * self.latent_action_dynamics_loss_scale
            )
            loss += latent_action_dynamics_loss
            log["latent_action_model/loss"] = loss
            log["latent_action_model/dynamics_loss"] = latent_action_dynamics_loss

        return loss, log

    def training_step(self, model_state: TrainState, batch: Transition, key: jax.random.PRNGKey):
        (_, log), grads = eqx.filter_value_and_grad(self.loss, has_aux=True)(
            model_state.module,
            action=batch.action,
            observation=batch.observation,
            next_observation=batch.next_observation,
            key=key,
        )

        model_state = model_state.apply_updates(grads)

        return model_state, log
