import equinox as eqx
import jax
import jax.numpy as jnp
import optax
import chex

from rlab_jax.modules.networks import PETSGaussMlp, Module, GaussMlp
from rlab_jax.rlab_dataclasses import Prediction, TrainState, Transition, DeterministicStochasticRandomVariable, RandomVariable
from rlab_jax.utils.keygen import keygen
from rlab_jax.rlab_dataclasses import Transition, ScalerTrainState
from rlab_jax_projects.latent_actions.modules.mean_std_scaler import MeanStdScaler
from rlab_jax.distributions import multivariate_normal_diag
from rlab_jax.modules.sequence_models import SequenceModel


def get_done_fn(env_name):
    if 'hopper' in env_name.lower():
        return hopper_is_done
    elif 'walker' in env_name.lower():
        return walker2d_is_done
    elif 'halfcheetah' in env_name.lower():
        return halfcheetah_is_done
    else:
        raise NotImplementedError


def hopper_is_done(next_observation):
    height = next_observation[..., 0]
    angle = next_observation[..., 1]
    not_done = jnp.isfinite(next_observation).all(axis=-1) \
               * (jnp.abs(next_observation[..., 1:]) < 100).all(axis=-1) \
               * (height > .7) \
               * (jnp.abs(angle) < .2)
    done = ~not_done
    return jnp.expand_dims(done.astype(jnp.float32), -1)


def walker2d_is_done(next_observation):
    height = next_observation[..., 0]
    ang = next_observation[..., 1]
    done = ~((height > 0.8) & (height < 2.0) &
             (ang > -1.0) & (ang < 1.0))
    return jnp.expand_dims(done.astype(jnp.float32), -1)


def halfcheetah_is_done(next_observation):
    return jnp.expand_dims(jnp.zeros_like(next_observation[..., 0]), -1)


class ObservationActionMeanStdScaler(MeanStdScaler):
    def fit(self, data: Transition, model_state, ):
        inputs = jnp.concatenate((data.observation, data.action), axis=-1)

        mean = jnp.mean(inputs, axis=(0, 1))
        std = jnp.std(inputs, axis=(0, 1))

        model_state.scaler = ObservationActionMeanStdScaler(mean=mean, std=std)

        return model_state


class PETSGaussMlpEnsemble(Module):
    _MODULE_TYPE = "PETSGaussMlpEnsemble"
    min_logvar: jnp.array
    max_logvar: jnp.array
    ensemble: Module
    n_networks: int

    def __init__(self, cfg, n_observations, n_actions, *args, **kwargs):

        self.min_logvar = jnp.ones((n_observations + 1)) * -10
        self.max_logvar = jnp.ones((n_observations + 1)) * 0.5

        class GaussMlpHelper(GaussMlp):
            def call(self, x):
                y = self.mlp(x)
                mean = self.mean_layer(y)
                logvar = self.std_layer(y)
                return mean, logvar

        @eqx.filter_vmap
        def ensemble(key):
            return GaussMlpHelper(
                n_inputs=n_observations + n_actions,
                n_outputs=n_observations + 1,
                key=key,
                layers=cfg.layers,
                units=cfg.units,
                hidden_activation=cfg.hidden_activation,
                init_bias_type="zero",
                init_weight_type="trunc_normal",
            )

        self.ensemble = ensemble(jax.random.split(keygen(), cfg.get("n_networks", 2)))
        self.n_networks = cfg.get("n_networks", 2)

    def __call__(self, x):

        if x.shape[0] == self.n_networks:
            @eqx.filter_vmap(in_axes=(eqx.if_array(0), 0))
            def vmap_ensemble(e, x):
                return e(x)
        else:
            @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
            def vmap_ensemble(e, x):
                return e(x)

        mean, logvar = vmap_ensemble(self.ensemble, x)

        logvar = self.max_logvar - jax.nn.softplus(self.max_logvar - logvar)
        logvar = self.min_logvar + jax.nn.softplus(logvar - self.min_logvar)
        std = jnp.exp(0.5 * logvar)
        return multivariate_normal_diag(mean=mean, std=std)



class MBPOModel(SequenceModel):
    _SEQUENCE_MODEL_TYPE = "MBPOModel"

    def __init__(
        self,
        ensemble: Module,
        done_fn: callable,
        n_networks: int,
        n_observations: int,
        n_elites: int,
        n_actions: int,
        optim_type: str = "optax.adam",
        optim_settings: dict = {"learning_rate": 3e-4},
        logvar_loss_coef: float = 0.01
    ):
        super().__init__()

        self.done_fn = done_fn
        self.n_networks = n_networks
        self.n_elites = n_elites
        self.logvar_loss_coef = logvar_loss_coef
        self.scaler = ObservationActionMeanStdScaler(shape=(n_observations + n_actions,))
        optim = eval(optim_type)(**optim_settings)

        @chex.dataclass
        class EliteScalerTrainState(ScalerTrainState):
            elite_indexes: eqx.Module

        self.train_state = EliteScalerTrainState(
            module=ensemble, scaler=self.scaler, optim=optim, optim_state=optim.init(eqx.filter(ensemble, eqx.is_array)), elite_indexes=jnp.arange(n_elites))

    def initial_state(self):
        return self.train_state

    @classmethod
    def from_cfg(cls, cfg, n_observations, n_actions, env_name, **_ignored):

        return cls(
            ensemble=PETSGaussMlpEnsemble(cfg=cfg, n_observations=n_observations, n_actions=n_actions),
            n_networks=cfg.get("n_networks", 2),
            done_fn=get_done_fn(env_name),
            optim_type=cfg.optim.type,
            optim_settings=cfg.optim.settings,
            n_observations=n_observations,
            n_actions=n_actions,
            n_elites=cfg.n_elites
        )

    def predict_next_observation_and_reward(self, model: Module, observation: jnp.array, action: jnp.array, scaler: Module):
        x = jnp.concatenate((observation, action), -1)
        x = self.scaler.transform(x=x, scaler=scaler)
        return model(x)

    def __call__(self, model_state: ScalerTrainState, observation: jnp.array, action: jnp.array, key: jax.random.PRNGKey, sample_shape=(), deterministic=False, return_dist=False):
        output = self.predict_next_observation_and_reward(model=model_state.module, observation=observation, action=action, scaler=model_state.scaler)

        if return_dist:
            return output

        if deterministic:
            output = output.mean()
        else:
            output = output.sample(key, sample_shape=sample_shape)

        next_observation = output[..., :-1] + observation
        reward = jnp.expand_dims(output[..., -1], -1)
        done = self.done_fn(next_observation)

        return next_observation, reward, done

    @staticmethod
    def weight_decay_loss(model: PETSGaussMlpEnsemble):
        loss = 0
        decays = [0.000025, 0.00005, 0.000075, 0.000075, 0.0001, 0.0001]
        i = 0
        for x in model.ensemble.mlp:
            if isinstance(x, eqx.nn.Linear):

                loss += decays[i] * (0.5 * ((x.weight ** 2).sum()))
                i += 1

        loss += decays[-1] * (0.5 * ((model.ensemble.mean_layer.weight ** 2).sum()))
        loss += decays[-1] * (0.5 * ((model.ensemble.std_layer.weight ** 2).sum()))

        return loss

    def loss(self, model: PETSGaussMlp, scaler: Module, observation: jnp.array, action: jnp.array, target: jnp.array):
        next_observation_dist = self.predict_next_observation_and_reward(model=model, observation=observation, action=action, scaler=scaler)

        logvar = jnp.log(next_observation_dist.std()**2)

        mean_loss = (jnp.exp(-logvar) * (next_observation_dist.mean() - target) ** 2).mean((1, 2))
        var_loss = logvar.mean((1, 2))

        indv_loss = mean_loss.sum() + var_loss.sum()

        loss = indv_loss.sum() + self.weight_decay_loss(model) + self.logvar_loss_coef * model.max_logvar.sum() - self.logvar_loss_coef * model.min_logvar.sum()
        return loss, {"ensemble/loss": loss, "ensemble/indv_loss": indv_loss}

    def validation_loss(self, model: PETSGaussMlp, scaler: Module, observation: jnp.array, action: jnp.array, target: jnp.array):
        next_observation_dist = self.predict_next_observation_and_reward(model=model, observation=observation, action=action, scaler=scaler)
        return ((next_observation_dist.mean() - target) ** 2).mean((1, 2))

    def validation_step(
            self,
            model_state: ScalerTrainState,
            batch: Transition,
            key: jax.random.PRNGKey,
    ):
        validation_loss = self.validation_loss(
            model=model_state.module,
            scaler=model_state.scaler,
            observation=batch.observation,
            action=batch.action,
            target=jnp.concatenate((batch.next_observation - batch.observation, batch.reward), axis=-1),
        )

        return {"ensemble/ind_validation_loss": validation_loss}

    def training_step(
        self,
        model_state: ScalerTrainState,
        batch: Transition,
        key: jax.random.PRNGKey,
    ):
        (_, log), grads = eqx.filter_value_and_grad(self.loss, has_aux=True)(
            model_state.module,
            scaler=model_state.scaler,
            observation=batch.observation,
            action=batch.action,
            target=jnp.concatenate((batch.next_observation - batch.observation, batch.reward), axis=-1),
        )

        model_state = model_state.apply_updates(grads)

        return model_state, log


class MBPODreamRollout:
    def __init__(self, model, act_fn, steps_per_rollout, memory):
        self.model = model
        self.act_fn = act_fn
        self.steps_per_rollout = steps_per_rollout
        self.memory = memory

    @eqx.filter_jit
    def generate(self, agent_state, model_state, initial_observation, key: jax.random.PRNGKey, memory_state):
        keys = jax.random.split(key, 2)

        observation = initial_observation

        observations, next_observations, actions, rewards, dones = [], [], [], [], []

        def select_random_network(x, key_):
            network_i = jax.random.choice(key=key_, a=model_state.elite_indexes, shape=(initial_observation.shape[0],))

            return eqx.filter_vmap(lambda i: x[network_i[i], i])(jnp.arange(initial_observation.shape[0]))

        for step in range(self.steps_per_rollout):
            keys = jax.random.split(keys[0], 4)
            action, _ = self.act_fn(agent_state=agent_state, observation=observation, key=keys[1])
            action = action.sample

            next_observation, reward, done = self.model(model_state=model_state,
                                                        observation=observation,
                                                        action=action,
                                                        key=keys[2])

            next_observation = select_random_network(next_observation, keys[3])
            reward = select_random_network(reward, keys[3])
            done = select_random_network(done, keys[3])

            observations.append(observation)
            next_observations.append(next_observation)
            actions.append(action)
            rewards.append(reward)
            dones.append(done)

            observation = next_observation

        observations = jnp.stack(observations, axis=1)
        next_observations = jnp.stack(next_observations, axis=1)
        actions = jnp.stack(actions, axis=1)
        rewards = jnp.stack(rewards, axis=1)
        dones = jnp.stack(dones, axis=1)

        memory_state = self.memory.push(
            memory_state=memory_state,
            observation=observations,
            next_observation=next_observations,
            action=actions,
            reward=rewards,
            done=dones,
            mask=jnp.ones_like(rewards),
        )

        return memory_state

