import equinox as eqx
import jax
import jax.numpy as jnp
import omegaconf
import optax

from rlab_jax.distributions import EqxDistribution
from rlab_jax.modules.sequence_models.rssm import (
    RSSM, DeterministicStochasticRandomVariable,
    RSSMModel)
from rlab_jax.modules.sequence_models.sequence_model import SequenceModel
from rlab_jax.rlab_dataclasses import (Prediction, StateActionRandomVariable,
                                       StateActionRandomVariableList, ScalerAnnealingTrainState,
                                       random_variable, random_variable_list, RandomVariable, Transition)
from rlab_jax_projects.latent_actions.modules.latent_action_vae import (
    LatentActionVAE, LatentActionVAEModel)
from rlab_jax_projects.latent_actions.modules.mean_std_scaler import MeanStdScaler, NoScaler


class LatentActionRSSMModel(eqx.Module):
    sequence_model: RSSMModel
    latent_action_model: LatentActionVAEModel

    @property
    def reward(self):
        return self.sequence_model.reward

    @property
    def done(self):
        return self.sequence_model.done


class LatentActionRSSM(SequenceModel):
    _SEQUENCE_MODEL_TYPE = "LatentActionRSSM"

    def __init__(
        self,
        sequence_model,
        latent_action_model,
        transition_action,
        n_observations,
        beta=1.0,
        temperature=1.0,
        optim_type: str = "optax.adam",
        optim_settings: dict = {"lr": 3e-4},
        max_gradient_norm: float = 100,
        state_loss_scaling: float = 1.0,
        action_nll_scaling: float = 1.0,
        n_initial_filtered_steps: int = 1,
        standardize_observations: bool = False,
        *args,
        **kwargs
    ):
        super().__init__()

        self.sequence_model = sequence_model
        self.latent_action_model = latent_action_model
        self.transition_action = transition_action
        self.n_latent = self.sequence_model.n_latent
        self.n_actions = self.latent_action_model.n_actions

        self.beta = beta
        self.temperature = jnp.array(temperature)
        self.max_gradient_norm = max_gradient_norm

        self.state_loss_scaling = state_loss_scaling
        self.action_nll_scaling = action_nll_scaling

        self.n_initial_filtered_steps = n_initial_filtered_steps

        if standardize_observations:
            self.scaler = MeanStdScaler(shape=(n_observations,))
        else:
            self.scaler = NoScaler()

        optim_settings = dict(optim_settings)
        if "lr" in optim_settings.keys():
            optim_settings["learning_rate"] = optim_settings["lr"]
            del optim_settings["lr"]

        optim_settings["learning_rate"] = float(optim_settings["learning_rate"])

        if optim_type == "sam_adam":
            from optax import contrib

            def sam_adam(rho=0.01, sync_period=2):
                """A SAM optimizer using Adam for the outer optimizer."""
                opt = optax.chain(optax.clip_by_global_norm(self.max_gradient_norm), optax.adam(**optim_settings))
                adv_opt = optax.chain(contrib.normalize(), optax.adam(rho))
                return contrib.sam(opt, adv_opt, sync_period=sync_period)

            optim = sam_adam()

        elif optim_type == "sam_sgd":
            from optax import contrib

            def sam_sgd(rho=0.01, sync_period=2):
                """A SAM optimizer using Adam for the outer optimizer."""
                opt = optax.sgd(**optim_settings)
                adv_opt = optax.chain(contrib.normalize(), optax.sgd(rho))
                return contrib.sam(opt, adv_opt, sync_period=sync_period)

            optim = sam_sgd()

        else:
            optim = optax.chain(
                optax.clip_by_global_norm(self.max_gradient_norm), eval(optim_type)(**optim_settings)
            )

        model_state = LatentActionRSSMModel(
            sequence_model=sequence_model.initial_state().module,
            latent_action_model=latent_action_model.initial_state().module,
        )

        self.train_state = ScalerAnnealingTrainState(
            module=model_state,
            optim=optim,
            optim_state=optim.init(eqx.filter(model_state, eqx.is_array)),
            annealing=jnp.array(1 / temperature),
            scaler=self.scaler,
        )

    def initial_state(self):
        return self.train_state

    @classmethod
    def from_cfg(cls, cfg: omegaconf.DictConfig, n_observations, n_actions, env_name, **_ignored):
        latent_action_model = LatentActionVAE.from_cfg(
            cfg.latent_action_model,
            n_states=cfg.n_stochastic_states + cfg.n_deterministic_states,
            n_actions=n_actions,
        )

        if cfg.latent_action_model.transition_action == "latent_action":
            n_transition_actions=cfg.latent_action_model.n_latent_actions
        elif cfg.latent_action_model.transition_action == "both":
            n_transition_actions = cfg.latent_action_model.n_latent_actions + n_actions
        else:
            n_transition_actions = n_actions

        sequence_model = RSSM.from_cfg(
            cfg,
            n_observations,
            n_actions=n_actions,
            n_transition_actions=n_transition_actions,
            env_name=env_name
        )

        return cls(
            sequence_model=sequence_model,
            latent_action_model=latent_action_model,
            transition_action=cfg.latent_action_model.transition_action,
            beta=cfg.beta,
            temperature=cfg.temperature,
            optim_type=cfg.optim.type,
            optim_settings=cfg.optim.settings,
            state_loss_scaling=cfg.state_loss_scaling,
            action_nll_scaling=cfg.action_nll_scaling,
            n_initial_filtered_steps=cfg.n_initial_filtered_steps,
            standardize_observations=cfg.standardize_observations,
            n_observations=n_observations,
        )

    def __call__(
        self,
        model: LatentActionRSSMModel,
        observation_trajectory: jnp.array,
        action_trajectory: jnp.array,
        key: jax.random.PRNGKey,
        scaler,
        training=False,
    ) -> Prediction:

        keys = jax.random.split(key, 2)

        observation_trajectory = self.scaler.transform(observation_trajectory, scaler)

        prior, posterior = self._encode(
            model=model, observation_trajectory=observation_trajectory, action_trajectory=action_trajectory, training=training, key=keys[0]
        )

        reconstruction = self._decode(model=model, latent_states=posterior, training=training, key=keys[1])

        if not training:
            reconstruction.state.sample = self.scaler.inverse_transform(x=reconstruction.state.sample, scaler=scaler)

        return Prediction(posterior=posterior, prior=prior, reconstruction=reconstruction)

    def _encode(
        self,
        model: LatentActionRSSMModel,
        observation_trajectory: jnp.array,
        action_trajectory: jnp.array,
        key: jax.random.PRNGKey,
        training=False
    ):
        keys = jax.random.split(key, 2)

        observation_embedding = model.sequence_model.encoder(observation_trajectory)

        prior_states = StateActionRandomVariableList(
            self.initial_latent_state(model=model, batch_size=observation_trajectory.shape[1])
        )
        posterior_state = self.initial_latent_state(model=model, batch_size=observation_trajectory.shape[1])
        posterior_states = StateActionRandomVariableList(posterior_state)

        for i in range(action_trajectory.shape[0] - 1):
            keys = jax.random.split(keys[0], 3)
            prior_state = self.one_step(
                model=model,
                latent_state=posterior_state,
                action=action_trajectory[i + 1],
                key=keys[1],
                action_noise=training
            )

            posterior_state = self.filter_after_one_step(
                model=model,
                latent_state=prior_state,
                observation_embedding=observation_embedding[i + 1],
                key=keys[2],
            )

            prior_states.append(prior_state)
            posterior_states.append(posterior_state)

        prior_states = prior_states.stack()
        posterior_states = posterior_states.stack()

        prior_actions = self.latent_action_model._prior(
            model=model.latent_action_model, state=jnp.concatenate(
                (jnp.expand_dims(posterior_states.state.sample[0], 0), posterior_states.state.sample[:-1])
            ), key=keys[0]
        )

        prior_states = StateActionRandomVariable(state=prior_states.state, action=prior_actions)

        return prior_states, posterior_states

    def _decode(
        self,
        model: LatentActionRSSMModel,
        latent_states: StateActionRandomVariable,
        key: jax.random.PRNGKey,
        training=False
    ):
        # state decoder
        key, subkey = jax.random.split(key, 2)
        state_reconstruction_dist = model.sequence_model.decoder(latent_states.state.sample_deterministic_stochastic)
        state_reconstruction = state_reconstruction_dist.sample(key)

        # action decoder, add latent state for first time step which is ignored in loss
        action_reconstruction = self.latent_action_model._decode(
            model=model.latent_action_model,
            latent_action=latent_states.action.sample,
            state=jnp.concatenate(
                (jnp.expand_dims(latent_states.state.sample[0], 0), latent_states.state.sample[:-1])
            ),
            key=subkey,
            dropout=training,
            deterministic=False
        )

        return StateActionRandomVariable(
            state=random_variable(sample=state_reconstruction, **state_reconstruction_dist.params),
            action=action_reconstruction,
        )

    def one_step(
        self,
        model: LatentActionRSSMModel,
        latent_state: StateActionRandomVariable,
        action: RandomVariable,
        key: jax.random.PRNGKey,
        latent_action: RandomVariable = None,
        action_noise=False
    ):
        keys = jax.random.split(key, 4)

        if latent_action is None:
            latent_action = self.latent_action_model._encode(
                model=model.latent_action_model,
                action=action,
                state=latent_state.state.sample,
                key=keys[1],
            )

        if self.transition_action == "latent_action":
            transition_action = latent_action.sample
        elif self.transition_action == "both":
            if isinstance(action, jax.Array):
                transition_action = jnp.concatenate((latent_action.sample, action), axis=-1)
            else:
                transition_action = jnp.concatenate((latent_action.sample, action.sample), axis=-1)
        else:
            if isinstance(action, jax.Array):
                transition_action = action
            else:
                transition_action = action.sample

        next_deterministic_state, next_stochastic_dist = model.sequence_model.transition(
            deterministic_state=latent_state.state.deterministic,
            stochastic_state=latent_state.state.stochastic.sample,
            action=transition_action,
            key=keys[2]
        )

        return StateActionRandomVariable(
            state=DeterministicStochasticRandomVariable(
                deterministic=next_deterministic_state,
                stochastic=random_variable(
                    sample=next_stochastic_dist.sample(keys[3]), **next_stochastic_dist.params
                ),
            ),
            action=latent_action,
        )

    def filter_after_one_step(
        self,
        model: LatentActionRSSMModel,
        latent_state: StateActionRandomVariable,
        observation_embedding: jnp.array,
        key: jax.random.PRNGKey,
    ):
        inputs = jnp.concatenate([latent_state.state.deterministic, observation_embedding], -1)
        stochastic_state_dist = model.sequence_model.stochastic_state_inference(inputs)

        return StateActionRandomVariable(
            state=DeterministicStochasticRandomVariable(
                deterministic=latent_state.state.deterministic,
                stochastic=random_variable(
                    sample=stochastic_state_dist.sample(key), **stochastic_state_dist.params
                ),
            ),
            action=latent_state.action,
        )

    def initial_latent_state(self, model: LatentActionRSSMModel, batch_size: int):
        return StateActionRandomVariable(
            state=DeterministicStochasticRandomVariable(
                deterministic=jnp.zeros((batch_size, self.sequence_model.n_deterministic_states)),
                stochastic=random_variable(
                    sample=jnp.zeros((batch_size, self.sequence_model.n_stochastic_states)),
                    mean=jnp.zeros((batch_size, self.sequence_model.n_stochastic_states)),
                    std=jnp.ones((batch_size, self.sequence_model.n_stochastic_states)),
                    dist_type=model.sequence_model.stochastic_state_inference._dist_type,
                ),
            ),
            action=random_variable(
                sample=jnp.zeros((batch_size, self.latent_action_model.n_latent_actions)),
                mean=jnp.zeros((batch_size, self.latent_action_model.n_latent_actions)),
                std=0.5 * jnp.ones((batch_size, self.latent_action_model.n_latent_actions)),
                dist_type=model.latent_action_model.prior._dist_type,
            ),
        )

    def loss(
        self,
        model: LatentActionRSSMModel,
        batch: Transition,
        key: jax.random.PRNGKey,
        scaler: MeanStdScaler,
        annealing: jnp.array,
    ):
        predictions = self.__call__(model, action_trajectory=batch.action_trajectory, observation_trajectory=batch.observation_trajectory, training=True, scaler=scaler, key=key)
        observation_trajectory = self.scaler.transform(batch.observation_trajectory, scaler)

        # ignore s_{0} and z_{0}

        # State NLL
        nll_sample_wise_state = -predictions.reconstruction.state.dist.log_prob(observation_trajectory)
        nll_sample_wise_state = nll_sample_wise_state.at[0].set(0)
        nll_state = jnp.mean(nll_sample_wise_state)

        # State KL
        if self.sequence_model.kl_balance is not None:
            kl_prior = jax.lax.stop_gradient(predictions.posterior.state.stochastic.dist).kl_divergence(predictions.prior.state.stochastic.dist)
            kl_posterior = predictions.posterior.state.stochastic.dist.kl_divergence(jax.lax.stop_gradient(predictions.prior.state.stochastic.dist))
            kl_prior = kl_prior.at[0].set(0)
            kl_posterior = kl_posterior.at[0].set(0)

            if self.sequence_model.free_nats is not None:
                kl_prior = jnp.maximum(kl_prior.mean(), self.sequence_model.free_nats)
                kl_posterior = jnp.maximum(kl_posterior.mean(), self.sequence_model.free_nats)
            else:
                kl_prior = kl_prior.mean()
                kl_posterior = kl_posterior.mean()

            kl_state = self.sequence_model.kl_balance * kl_prior + (1 - self.sequence_model.kl_balance) * kl_posterior

        else:
            kl_sample_wise_state = predictions.posterior.state.stochastic.dist.kl_divergence(predictions.prior.state.stochastic.dist)
            kl_sample_wise_state = kl_sample_wise_state.at[0].set(0)

            if self.sequence_model.free_nats is not None:
                kl_sample_wise_state = jnp.maximum(kl_sample_wise_state, self.sequence_model.free_nats)

            kl_state = jnp.mean(kl_sample_wise_state)

        kl_state = kl_state * self.sequence_model.beta * annealing

        # ignore a_{-1} and v_{-1}, ignore a_{0} and v_{0}

        # Action NLL
        nll_sample_wise_action = -predictions.reconstruction.action.dist.log_prob(batch.action_trajectory)
        nll_sample_wise_action = nll_sample_wise_action.at[0].set(0)
        nll_sample_wise_action = nll_sample_wise_action.at[1].set(0)
        nll_action = jnp.mean(nll_sample_wise_action)

        # Action KL
        if self.latent_action_model.kl_balance is not None:
            kl_action_prior = jax.lax.stop_gradient(predictions.posterior.action.dist).kl_divergence(predictions.prior.action.dist)
            kl_action_posterior = predictions.posterior.action.dist.kl_divergence(jax.lax.stop_gradient(predictions.prior.action.dist))
            kl_action_prior = kl_action_prior.at[0].set(0)
            kl_action_prior = kl_action_prior.at[1].set(0)
            kl_action_posterior = kl_action_posterior.at[0].set(0)
            kl_action_posterior = kl_action_posterior.at[1].set(0)

            if self.latent_action_model.free_nats is not None:
                kl_action_prior = jnp.maximum(kl_action_prior.mean(), self.latent_action_model.free_nats)
                kl_action_posterior = jnp.maximum(kl_action_posterior.mean(), self.latent_action_model.free_nats)
            else:
                kl_action_prior = kl_action_prior.mean()
                kl_action_posterior = kl_action_posterior.mean()

            kl_action = self.latent_action_model.kl_balance * kl_action_prior + (1 - self.latent_action_model.kl_balance) * kl_action_posterior

        else:
            kl_sample_wise_action = predictions.posterior.action.dist.kl_divergence(predictions.prior.action.dist)
            kl_sample_wise_action = kl_sample_wise_action.at[0].set(0)
            kl_sample_wise_action = kl_sample_wise_action.at[1].set(0)

            if self.latent_action_model.free_nats is not None:
                kl_sample_wise_action = jnp.maximum(kl_sample_wise_action, self.latent_action_model.free_nats)

            kl_action = jnp.mean(kl_sample_wise_action)

        kl_action = kl_action * self.latent_action_model.beta * annealing

        # Loss
        state_loss = nll_state + kl_state
        action_loss = nll_action * self.action_nll_scaling + kl_action
        loss = action_loss + state_loss * self.state_loss_scaling

        reward_loss, reward_log = model.reward.loss(next_state=predictions.posterior.item(jnp.array(range(1, predictions.posterior.sample.shape[0]))),
                                                    action=batch.action,
                                                    target=batch.reward)

        done_loss, done_log = model.done.loss(next_state=predictions.posterior.item(jnp.array(range(1, predictions.posterior.sample.shape[0]))),
                                              action=batch.action,
                                              target=batch.done)

        loss += reward_loss
        loss += done_loss

        annealing = jnp.minimum(annealing + (1.0 / self.temperature), 1.0)

        log = {
            "sequence_model/loss": loss,
            "sequence_model/nll/s": nll_state,
            "sequence_model/kl/s": kl_state,
            "sequence_model/nll/v": nll_action,
            "sequence_model/kl/v": kl_action,
            "sequence_model/warmup_kl/s": annealing,
            **reward_log,
            **done_log
        }

        return loss, (log, predictions, annealing)

    def training_step(
        self, model_state: ScalerAnnealingTrainState, batch: Transition, key: jax.random.PRNGKey
    ):
        (_, (log, predictions, annealing)), grads = eqx.filter_value_and_grad(self.loss, has_aux=True)(
            model_state.module,
            batch=batch,
            annealing=model_state.annealing,
            scaler=model_state.scaler,
            key=key,
        )

        model_state = model_state.apply_updates(grads)
        model_state.annealing = annealing

        return model_state, predictions, log

    @eqx.filter_jit
    def filter(
        self,
        model_state: ScalerAnnealingTrainState,
        next_observation: jnp.array,
        latent_state: StateActionRandomVariable,
        latent_action: jnp.array,
        action: jnp.array,
        key: jax.random.PRNGKey,
        reset=False,
        **ignored
    ):
        if reset:
            return self.initial_latent_state(
                model=model_state.module, batch_size=next_observation.shape[0]
            )

        keys = jax.random.split(key, 2)

        next_observation = self.scaler.transform(next_observation, model_state.scaler)
        observation_embedding = model_state.module.sequence_model.encoder(next_observation)

        prior_state = self.one_step(model=model_state.module, latent_state=latent_state, latent_action=latent_action,
                                    action=action,  key=keys[0])

        prior_state.state.stochastic.sample = prior_state.state.stochastic.mean

        posterior_state = self.filter_after_one_step(
            model=model_state.module,
            latent_state=prior_state,
            observation_embedding=observation_embedding,
            key=keys[1],
        )

        posterior_state.state.stochastic.sample = posterior_state.state.stochastic.mean

        return posterior_state

    def generate(
        self,
        model_state: ScalerAnnealingTrainState,
        next_observations: jnp.array,
        actions: jnp.array,
        key: jax.random.PRNGKey,
    ) -> Prediction:
        latent_state = self.initial_latent_state(model=model_state.module, batch_size=next_observations.shape[1])

        next_observations = self.scaler.transform(next_observations, model_state.scaler)
        observation_embedding = model_state.module.sequence_model.encoder(next_observations)

        for i in range(actions.shape[0]):
            key, subkey = jax.random.split(key)
            latent_state = self.one_step(
                model=model_state.module,
                latent_state=latent_state,
                action=actions[i],
                key=subkey,
            )

            if i < self.n_initial_filtered_steps:
                key, subkey = jax.random.split(key)
                latent_state = self.filter_after_one_step(
                    model=model_state.module,
                    latent_state=latent_state,
                    observation_embedding=observation_embedding[i],
                    key=subkey,
                )

            if i == 0:
                latent_states = random_variable_list(latent_state)
            else:
                latent_states.append(latent_state)

        latent_states = latent_states.stack()

        key, subkey = jax.random.split(key, 2)
        reconstruction = self._decode(model=model_state.module, latent_states=latent_states, key=subkey)

        reconstruction.state.sample = self.scaler.inverse_transform(x=reconstruction.state.sample, scaler=model_state.scaler)
        reconstruction.state.mean = self.scaler.inverse_transform(x=reconstruction.state.mean, scaler=model_state.scaler)

        return Prediction(reconstruction=reconstruction, posterior=None, prior=latent_states)
