import equinox as eqx
import jax
import jax.numpy as jnp
import optax

from rlab_jax.modules.networks import GaussMlp, Module
from rlab_jax.rlab_dataclasses import Prediction, TrainState, Transition, DeterministicStochasticRandomVariable, RandomVariable
from rlab_jax.utils.keygen import keygen


class RSSMTransitionEnsemble(Module):
    _MODULE_TYPE = "RSSMPlanning2ExploreTransitionEnsemble"

    disagreement_weighting: float
    train_state: TrainState

    def __init__(
        self,
        ensemble: Module,
        disagreement_weighting: float,
        optim_type: str = "optax.adam",
        optim_settings: dict = {"learning_rate": 3e-4},
    ):
        super().__init__()
        self.disagreement_weighting = disagreement_weighting

        optim = eval(optim_type)(**optim_settings)
        self.train_state = TrainState(
            module=ensemble, optim=optim, optim_state=optim.init(eqx.filter(ensemble, eqx.is_array))
        )

    def initial_state(self):
        return self.train_state

    @classmethod
    def from_cfg(cls, cfg, n_states, n_actions, n_targets, **_ignored):
        @eqx.filter_vmap
        def ensemble(key):
            return GaussMlp(
                n_inputs=n_states + n_actions,
                n_outputs=n_targets,
                key=key,
                layers=cfg.layers,
                units=cfg.units,
                hidden_activation=cfg.hidden_activation,
            )

        return cls(
            ensemble=ensemble(jax.random.split(keygen(), cfg.get("n_networks", 2))),
            disagreement_weighting=cfg.disagreement_weighting,
            optim_type=cfg.optim.type,
            optim_settings=cfg.optim.settings,
        )

    @staticmethod
    def predict_next_state(ensemble: Module, state: jnp.array, action: jnp.array):
        input = jnp.concatenate((state, action), -1)

        @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
        def evaluate_ensemble(model, x):
            return model(x)

        return evaluate_ensemble(ensemble, input)

    def disagreement(self, ensemble: Module, state: jnp.array, action: jnp.array):
        next_state_dist = self.predict_next_state(ensemble, state.deterministic, action)
        ensemble_mean = jnp.mean(next_state_dist.mean(), axis=0)
        ensemble_variance = jnp.mean(
            next_state_dist.var() + jnp.square(next_state_dist.mean()), axis=0
        ) - jnp.square(ensemble_mean)
        ensemble_std = jnp.sqrt(ensemble_variance)

        return ensemble_std.mean(-1) * self.disagreement_weighting

    def loss(self, ensemble: Module, state: jnp.array, action: jnp.array, target: jnp.array):
        next_state_dist = self.predict_next_state(ensemble, state, action)
        loss = -next_state_dist.dist().log_prob(target).mean()
        return loss, {"ensemble/loss": loss}

    def training_step(
        self,
        ensemble_state: TrainState,
        model_state: TrainState,
        predictions: Prediction,
        batch: Transition,
        key: jax.random.PRNGKey,
    ):
        input = predictions.posterior.deterministic[:-1]

        observation_embedding_dist = model_state.module.encoder(batch.observation[1:])
        target = observation_embedding_dist.sample(key)

        (_, log), grads = eqx.filter_value_and_grad(self.loss, has_aux=True)(
            ensemble_state.module,
            state=input,
            action=batch.action[1:],
            target=target,
        )

        ensemble_state = ensemble_state.apply_updates(grads)

        return ensemble_state, log

