import os

import h5py
import jax.numpy as jnp
import tqdm
import matplotlib.pyplot as plt
import jax


def get_keys(h5file):
    keys = []

    def visitor(name, item):
        if isinstance(item, h5py.Dataset):
            keys.append(name)

    h5file.visititems(visitor)
    return keys


def load(h5path, env):
    data_dict = {}
    with h5py.File(h5path, "r") as dataset_file:
        for k in tqdm.tqdm(get_keys(dataset_file), desc="load datafile"):
            try:  # first try loading as an array
                data_dict[k] = dataset_file[k][:]
            except ValueError as e:  # try loading as a scalar
                data_dict[k] = dataset_file[k][()]

    # Run a few quick sanity checks
    for key in ["observation", "action", "reward", "step_type"]:
        assert key in data_dict, "Dataset is missing key %s" % key
    N_samples = data_dict["observation"].shape[0]
    if env.observation_space.shape is not None:
        assert (
            data_dict["observation"].shape[2:] == env.observation_space.shape
        ), "Observation shape does not match env: %s vs %s" % (
            str(data_dict["observation"].shape[1:]),
            str(env.observation_space.shape),
        )
    assert (
        data_dict["action"].shape[2:] == env.action_space.shape
    ), "Action shape does not match env: %s vs %s" % (
        str(data_dict["action"].shape[1:]),
        str(env.action_space.shape),
    )
    assert data_dict["reward"].shape == (N_samples, data_dict["observation"].shape[1],), "Reward has wrong shape: %s" % (
        str(data_dict["reward"].shape)
    )
    assert data_dict["step_type"].shape == (
    N_samples, data_dict["observation"].shape[1],), "Reward has wrong shape: %s" % (
        str(data_dict["step_type"].shape)
    )
    return data_dict


def get_offline_dataset(dataset_path, env, memory, prng_key, writer=None, validation_ratio=None, *args, **kwargs):
    original_datset = load(dataset_path, env)
    original_datset["step_type"][original_datset["step_type"] == 1] = 0.0
    original_datset["step_type"][original_datset["step_type"] == 2] = 0.0

    dataset = {}
    dataset["observation"] = original_datset["observation"][:, :-1]
    dataset["next_observation"] = original_datset["observation"][:, 1:]
    dataset["action"] = original_datset["action"][:, 1:]
    dataset["reward"] = jnp.expand_dims(original_datset["reward"][:, 1:], -1)
    dataset["done"] = jnp.expand_dims(original_datset["step_type"][:, 1:], -1)
    dataset["mask"] = jnp.ones_like(dataset["done"])

    memory.episode_length = dataset["observation"].shape[1]
    memory_state = memory.initial_state(size=dataset["observation"].shape[0], validation_ratio=validation_ratio, key=prng_key)

    memory_state = memory.push(memory_state=memory_state, **dataset)

    return memory, memory_state