import pathlib
import numpy as np
import h5py


def load_episodes(directory, capacity=None):
    # The returned directory from filenames to episodes is guaranteed to be in
    # temporally sorted order.
    filenames = sorted(directory.glob('*.npz'))
    if capacity:
        num_steps = 0
        num_episodes = 0
        for filename in reversed(filenames):
            length = int(str(filename).split('-')[-1][:-4])
            num_steps += length
            num_episodes += 1
            if num_steps >= capacity:
                break
        filenames = filenames[-num_episodes:]
    episodes = {}
    for filename in filenames:
        try:
            with filename.open('rb') as f:
                episode = np.load(f)
                episode = {k: episode[k] for k in episode.keys()}
                # Conversion for older versions of npz files.
                if 'is_terminal' not in episode:
                    episode['is_terminal'] = episode['discount'] == 0.
        except Exception as e:
            print(f'Could not load episode {str(filename)}: {e}')
            continue
        episodes[str(filename)] = episode
    return episodes


def main():
    input_dir = "/vd4rl/gdrive/cheetah_run/expert/64px/"
    output_dir = "/vd4rl/gdrive/hdf5/cheetah_run/expert/"
    output_name = "cheeath-run-expert.hdf5"

    step_type = np.ones(501)
    step_type[0] = 0
    step_type[500] = 2

    output = {}
    episodes = load_episodes(pathlib.Path(input_dir))
    episodes = list(episodes.values())

    actions = [e['action'] for e in episodes]
    discounts = [e['discount'] for e in episodes]
    observations = [e['image'] for e in episodes]
    rewards = [e['reward'] for e in episodes]
    step_types = [step_type for _ in episodes]

    output['action'] = np.stack(actions)
    output['discount'] = np.stack(discounts)
    output['observation'] = np.stack(observations).transpose(0, 1, 4, 2, 3)
    output['reward'] = np.stack(rewards)
    output['step_type'] = np.stack(step_types)

    out_dir = pathlib.Path(output_dir)
    out_dir.mkdir(parents=True, exist_ok=True)

    with h5py.File(out_dir / output_name, 'w') as shard_file:
        for k, v in output.items():
            shard_file.create_dataset(k, data=v, compression='gzip')


if __name__ == '__main__':
    main()
